#!/usr/bin/env python
# -*- coding: utf-8 -*-
assert '…' == '\xe2\x80\xa6', "file encoding error"

from lib.localize import (_, gettext)

# Copyright 2013 Axel Walthelm
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

# Outline of the algorithm:
# * read annotation information from kindle;
#   ignore annotations from other books,
#   merge edits into a single annotation entry,
#   ignore deleted annotations;
#   TBD: allow user to specify a time range of interesting annotations?
#   TBD: merge matching highlight and note into single entry?
# * determine which locations are relevant for kindle annotations
# * read raw kindle byte stream 'kindleHtml'
#   reason: that is what the kindle annotations locatios refer to
#   note: it would be nice to use unicode for processing, but I tried (python 2.7, March 2013)
#         and did run into a lot of trouble, like HTMLParser forcing type back to str;
#         since using KindleUnpack is essential, and it is not yet ported to python 3,
#         all processing is now done in utf-8 (which is a nice transfer format, but a not so nice processing format)
# * if needed convert kindleBytes to utf-8 as kindleHtml; maintain a map mapKindlePosToHtml
#   that maps from all relevant kindle locations to positions in kindleHtml;
#   make sure all positions in kindleHtml refer to valid utf-8 character start bytes
#   reason: usually kindleBytes is already in utf-8, but if not we can support conversion (although not easily)
# * convert kindleHtml to plain text 'kindleText' with location mapping 'mapKindleHtmlToText';
#   also normalize ASCII and unicode white spaces; special handling for zero width space (e.g. soft hyphen)
#   reason: when epub was converted to kindle format, html was processed and changed,
#           so we ignore html for mapping kindle to epub
#   reason: when epub was converted to kindle format, unicode may have been transformed e.g.
#           by normalization, inserting soft hyphens etc.
#   note: we deliberatley do not normalize the text appearance, like punctuation (e.g. '…' to '...'),
#         hyphenation (e.g. '—' to '-'), quoting (e.g. '“' to '"'), etc.;
#         nor do we do unicode normalization
#   reason: text appearance is usually _not_ "improved" in such ways when converting epub to kindle;
#           if it is done, most cases should be handled by the fuzzy diff text mapping later.
# * read epub html as utf-8 string to 'epubHtml' from multiple files
# * analyze epubHtml for annotations already put in there by this algorithm;
#   annotations already in epubHtml must not be set again;
#   extract/remove annotation-listing from epubHtml;
#   TBD: get a time of last annotation insertion and insert only newer annotations?
# * convert epubHtml to plain utf-8 text 'epubText'
#   using the same rules as for kindleHtml to plain text
# * do a diff analysis on 'kindleText' vs. 'epubText' resulting in 'matchKindleToEpub'
#   reason: when epub was converted to kindle format, text may have been added;
#           e.g. a table of content may have been added at the beginning  
#   reason: when epub was converted to kindle format, text may have been changed;
#           e.g. text appearance was "improved" (see above)
# * try to locate annotation in epub using diff analysis; allow for some changes
#   TBD: fallback strategies if that fails? E.g.:
#        - if marked text is unique in kindText and unique in epubText, consider it moved
#        - if a quite similar text is found near the expected location, consider it changed
# * determine which locations in epubText are relevant for annotations
#   and create map 'mapEpubTextToHtml to convert locations of epubText to index in epubHtml
#
# * mark annotations in epubHtml
#   - use three levels of highlighing, depending on how many highlights overlap
#   - notes displayed as tooltips
#   - ...
# * split epubHtml into separate file chunks and write them back
# * add or re-use file with annotation-listing
#
# Summary of the mapping:
#    mapKindlePosToHtml -> mapKindleHtmlToText -> matchKindleToEpub -> mapEpubTextToHtml

__license__   = 'GPL v3'
__copyright__ = '2013, Axel Walthelm'

import sys
import re

from KindleUnpack.lib import kindleunpack

import codecs
import tempfile
import shutil
import os
import glob
import xml.etree.ElementTree as ElementTree
import zipfile
import string
import random
import time
import difflib
import heapq
from cStringIO import StringIO

# import our stuff
import lib.ParseHtmlToPlainText as ParseHtmlToPlainText
import lib.ParseUserAnnotLog as ParseUserAnnotLog
import lib.TextMatcher as TextMatcher
import lib.MarkupEpub as MarkupEpub
import lib.Utf8Tools as utf8
import lib.XmlTools as xml

def default_progress_callback(progress, message):
    print "Progress: %.2f %s" % (progress, message)
global progress_callback
progress_callback = default_progress_callback

def findUniqueFile(root, folder, pattern):
    globpath = os.path.join(root, folder, pattern)
    files = glob.glob(globpath)
    if len(files) > 1:
        raise Exception("file ambiguous: %s (%d matches)" % (globpath, len(files)) )
    if len(files) == 0:
        raise Exception("pattern not found: "+globpath)
    return files[0]
    
def copyModifyZip(zipread, newFileName, replaceDict):
    date_time=time.localtime(time.time())[:6]
    with zipfile.ZipFile(newFileName, 'w', zipfile.ZIP_DEFLATED) as zipwrite:
        for item in zipread.infolist():
            if item.filename not in replaceDict:
                data = zipread.read(item.filename)
                zipwrite.writestr(item, data, zipfile.ZIP_DEFLATED)
            else:
                item.date_time = date_time
                zipwrite.writestr(item, replaceDict[item.filename], zipfile.ZIP_DEFLATED)

def processBook(kindleEbook, userannotlog, epubEbook, **opts):
    progress_callback(0.00, _('Starting'))
    print 'epubEbook = %r' % epubEbook
    print 'kindleEbook = %r' % kindleEbook
    print 'userannotlog = %r' % userannotlog
    print 'opts = %r' % opts
    assert isinstance(epubEbook, unicode)
    assert isinstance(kindleEbook, unicode)
    assert isinstance(userannotlog, unicode)

    def getopt(name, default):
        value = opts[name] if opts.has_key(name) else None
        return value if value is not None else default

    backup = max(0, int(getopt('backup', 99)))
    postfix = unicode(getopt('postfix', u" + KiNotes"))

    newEpub = tmpEpub = re.sub(ur"(\.epub|)$", postfix + u".epub", epubEbook, re.IGNORECASE)
    
    rand, tries = random.Random(), 1000
    while os.path.isfile(tmpEpub):
        tries -= 1
        if tries < 0:
            raise Exception("tmpEpub")
        tmpEpub = newEpub + u"(%06d)" % rand.randint(tries,999999)
    
    unpackDir = getopt('unpackDir', None)
    tmpdir = unpackDir if unpackDir else tempfile.mkdtemp('', prefix='KiNotes_tmp')
    if unpackDir:
        shutil.rmtree(tmpdir, ignore_errors=True)
    print "tmpdir = ", tmpdir
    try:
        progress_callback(0.01, _('KindleUnpack'))
        print "BEGIN kindleunpack.unpackBook"
        
        old_stdout = sys.stdout
        sys.stdout = mystdout = StringIO()
        try:
            kindleunpack.unpackBook(kindleEbook,
                                    tmpdir,
                                    dodump=False,
                                    dowriteraw=True,
                                    dosplitcombos=True)
        finally:
            sys.stdout = old_stdout
            unpackMessages = mystdout.getvalue()
            if unpackMessages:
                indent = 4
                print ' '*indent + unpackMessages.strip().replace('\n', '\n'+' '*indent)
        print "END kindleunpack.unpackBook"
        #print unpackMessages
        
        progress_callback(0.10, _('Open EPUB'))
        with zipfile.ZipFile(epubEbook) as epubZip:
            processOpenedBook(tmpdir, userannotlog, epubZip, tmpEpub, **opts)

    finally:
        if not unpackDir:
            shutil.rmtree(tmpdir, ignore_errors=True)
            print "tmpdir removed"

    if newEpub != tmpEpub:
        # backup all previous versions and move tmpEpub to newEpub
        if os.path.isfile(newEpub):
            def recurse_backup(pathEpub, nr):
                if nr > backup:
                    os.remove(pathEpub)
                    return
                bckEpub = re.sub(ur"\.epub$", u" (%d).epub" % nr, newEpub)
                if os.path.isfile(bckEpub):
                    recurse_backup(bckEpub, nr+1)
                os.rename(pathEpub, bckEpub)
            recurse_backup(newEpub, 1)
        os.rename(tmpEpub, newEpub)

    print "Result is:", newEpub
    progress_callback(1.00, _('Done'))

def processOpenedBook(kindleEbookUnpackDir, userannotlog, epubZip, newEpub, **opts):
    
    #################################################
    # * read annotation information from kindle;
    #################################################
    
    progress_callback(0.15, _('Read Kindle properties'))
    # get book identifier from OPF
    opf = ElementTree.parse(findUniqueFile(kindleEbookUnpackDir, "mobi7", "*.opf")).getroot()
    ns = { 'opf': "http://www.idpf.org/2007/opf" }
    asin = opf.find("opf:metadata/opf:meta[@name='ASIN']", ns)
    if asin is not None:
        asin = asin.get('content')
    else:
        # note: without an ASIN Kindle does log to "My Clippings.txt", but not to userannotlog;
        # todo: get matching annotations from MyClippings.txt as good as possible
        raise Exception("Book without ASIN not supported")
    print "ASIN =", asin
    
    # get encoding from OPF (expected: utf-8)
    kindleEnc = opf.find("opf:metadata/opf:meta[@name='output encoding']", ns).get('content')
    kindleEnc = codecs.lookup(kindleEnc).name # use standard name
    #print "kindleEnc=", kindleEnc
    if kindleEnc == 'utf-8-sig':
        kindleEnc = 'utf-8' # just in case
    if kindleEnc != 'utf-8':
        print "NOTE: kindle encoding is not utf-8 but", kindleEnc
    
    # read annotations
    progress_callback(0.20, _('Read annotations'))
    annoList = ParseUserAnnotLog.FromFileName(userannotlog, asin)
    if not annoList:
        raise Exception("No annotations found")
    print "Read %d annotations from userannotlog" % len(annoList)

    ###################################################################
    # * determine which locations are relevant for kindle annotations
    ###################################################################

    progress_callback(0.30, _('Read Kindle HTML'))
    mapKindleLocToHtml = {}
    for ka in annoList:
        mapKindleLocToHtml[ka.begin] = ka.begin
        # kindle loc for end is one char behind end,
        # but when tracking locations we better use location of
        # last character in range, if range is not empty
        if ka.end > ka.begin:
            mapKindleLocToHtml[ka.end] = max(0, ka.end-1)

    ################################################
    # * read raw kindle byte stream 'kindleBytes'
    ################################################        

    # read raw kindle markup as bytes
    with open(findUniqueFile(kindleEbookUnpackDir, "mobi7", "*.rawml"), 'rb') as file:
        kindleBytes = bytes( file.read() )
    assert isinstance(kindleBytes, bytes)
    
    #print "\n----------------\n"
    #for b in bytearray(kindleBytes):
    #    print "0x%x %s" % (b, chr(b))
    #print "\n----------------\n"
    #print unicode(kindleBytes, kindleEnc).encode('unicode_escape')
    #print "\n----------------\n"

    # sanity check for read kindleBytes:
    # - locations are in valid range
    # - locations are valid encoded characters (e.g. not on utf-8 continuation bytes)
    for loc in mapKindleLocToHtml.keys():
        if loc < 0 or loc >= len(kindleBytes):
            print "ERROR: invalid annotation location", loc
        elif kindleEnc == 'utf-8':
            if utf8.isCont(kindleBytes, loc):
                print "ERROR: misaligned annotation location", loc
        else:
            raise NotImplementedError()

    ##############################################################################################
    # * if needed convert kindleBytes to utf-8 as kindleHtml; maintain a map mapKindleLocToHtml
    #   that maps from all relevant kindle locations to positions in kindleHtml;
    #   make sure all positions in kindleHtml refer to valid utf-8 character start bytes
    ##############################################################################################
    
    if kindleEnc == 'utf-8':
        kindleHtml = kindleBytes
    else:
        # todo: implement support for utf-16, latin-1 etc. when needed
        # todo: update mapKindleLocToHtml accordingly
        # note: possible implementation may be to feed each character to codec
        #       maybe adding more bytes if codec raises an error
        raise NotImplementedError()

    #print "\n----------------\n"
    #for u in kindleHtml:
    #    print "0x%x %s" % (ord(u), u.encode('unicode_escape'))
    #print "\n----------------\n"
    #print kindleHtml.encode('unicode_escape')
    #print "\n----------------\n"

    # sanity check for read kindleBytes: locations are on word boundaries
    def mayBeSpace(html, loc):
        return loc < 0 or loc >= len(html) or re.match( r"[\s<>]", html[loc] )
    def mayBeWord(html, loc):
        return loc < 0 or loc >= len(html) or re.match( r"[\S<>]", html[loc] )
    for ka in annoList:
        beg = mapKindleLocToHtml[ka.begin]
        end = mapKindleLocToHtml[ka.end]
        if not mayBeWord(kindleHtml, beg) or not mayBeSpace(kindleHtml, beg-1):
            print "WARNING: annotation start location %d is not on word boundary '%s'" % (beg, kindleHtml[beg-1:beg+1].encode('string_escape'))
        if not mayBeWord(kindleHtml, end) or not mayBeSpace(kindleHtml, end+1):
            print "WARNING: annotation end location %d is not on word boundary '%s'" % (beg, kindleHtml[end:end+2].encode('string_escape'))

    #################################################################################################
    # * convert kindleHtml to plain text 'kindleText' with location mapping 'mapKindleHtmlToText';
    #################################################################################################
    
    # get plain text of kindle book with location mapping
    mapKindleHtmlToText = {}
    for loc in mapKindleLocToHtml.values():
        mapKindleHtmlToText[loc] = -1
    parser = ParseHtmlToPlainText.ParseHtmlToPlainText(mapKindleHtmlToText)
    parser.feed(kindleHtml)
    kindleText = parser.getPlainText()
    
    print "Extracting highlight text"
    progress_callback(0.40, _('Read annotations text'))
    mapKindleLocToText = {}
    for key in mapKindleLocToHtml.iterkeys():
        mapKindleLocToText[key] = mapKindleHtmlToText[mapKindleLocToHtml[key]]
    ParseUserAnnotLog.GetText(annoList, mapKindleLocToText, kindleText)
    for ka in annoList:
        print "#%d %d-%d %s %s" % (ka.ordernr, ka.begin, ka.end, ka.kind, '"%s"' % ka.note if ka.note else '')
        print ka.createtime, ka.modifytime
        print "'%s'" % ka.context[0]
        print "'%s'" % ka.text
        print "'%s'" % ka.context[1]

    ##########################################################################
    # * read epub html as utf-8 string to 'epubHtml' from multiple files
    ##########################################################################
    
    progress_callback(0.45, _('Read EPUB properties'))
    #print epubZip.namelist()    
    if epubZip.read("mimetype").lower() != "application/epub+zip":
        raise TypeError("invalid epub")
        
    metaCont = ElementTree.fromstring(epubZip.read("META-INF/container.xml"))
    ns = {'c': "urn:oasis:names:tc:opendocument:xmlns:container"}
    rootfile = metaCont.find("c:rootfiles/c:rootfile[@media-type='application/oebps-package+xml']",
                             ns).get('full-path')
    rootpath = rootfile[0:rootfile.rfind('/')+1]
    
    opf = ElementTree.fromstring(epubZip.read(rootfile))
    ns = {'p': "http://www.idpf.org/2007/opf"}
    idToFile = {}
    for item in opf.findall("p:manifest/p:item", ns):
        idToFile[item.get('id')] = rootpath + item.get('href')
    
    htmlFiles = []
    for itemref in opf.findall("p:spine/p:itemref", ns):
        htmlFiles.append(idToFile[itemref.get('idref')])

    progress_callback(0.45, _('Read EPUB HTML'))
    epubHtml = []
    for htmlFile in htmlFiles:
        epubHtml.append( xml.makeXmlUtf8(epubZip.read(htmlFile)) )
    isUnique = False
    while not isUnique:
        uniqueTag = 'filesep'+''.join(random.choice(string.digits+string.ascii_letters) for i in range(20))
        isUnique = True
        for h in epubHtml:
            if uniqueTag in h:
                isUnique = False
    filesSeparator = "<%s />" % uniqueTag
    epubHtml = filesSeparator.join(epubHtml)

    ###############################################################################
    # * analyze epubHtml for annotations already put in there by this algorithm;
    ###############################################################################

    # todo
    
    #########################################################
    # * convert epubHtml to plain utf-8 text 'epubText'
    #########################################################
    
    parser = ParseHtmlToPlainText.ParseHtmlToPlainText()
    parser.feed(epubHtml)

    epubText = parser.getPlainText()
    del parser # free resources
    
#    # test only: change some characters in text
#    print "ERROR: corrupting epub text"
#    def TestCorruption(html, text):
#        rand = random.Random(987654)
#        ops = 12000
#        textLocs  = [(rand.randint(1, len(epubText)-3)|1, 0) for _ in xrange(ops)]
#        textLocs += [(rand.randint(1, len(epubText)-3)|1, rand.randint(1,4)) for _ in xrange(ops)]
#        textLocs += [(rand.randint(1, len(epubText)-3)|1, -1) for _ in xrange(ops)]
#        textLocs = sorted({key: (key, value) for (key, value) in textLocs}.values())
#        def applyCorruption(locs, oldText):
#            newText = []
#            locCur = 0
#            for loc, dup in locs:
#                newText.append(oldText[locCur:loc])                
#                c = ord(oldText[loc])+1
#                if c == ord(' ') and dup > 0:
#                    c = ord('_') # don't create multi-space
#                elif ord('a') <= c <= ord('f') and dup == 0:
#                    c = ord(' ') # split a word
#                newText.append(chr(c)*(dup+1))
#                locCur = loc+1
#            newText.append(oldText[locCur:])
#            return ''.join(newText)
#        locMap = {loc: None for (loc, dup) in textLocs}
#        parser = ParseHtmlToPlainText.ParseHtmlToPlainText(locMap, doMapToPlainText=False)
#        parser.feed(html)
#        # spaces can only be changed when they map to a single space in both text and html
#        # also avoid changing non-ascii characters
#        textLocs = [(loc, dup) for (loc, dup) in textLocs
#                    if text[loc] == html[locMap[loc]] != '&' # don't change entities or such
#                    and ord(text[loc]) < 127 # don't change utf-8 characters
#                    and text[loc-1] != ' ' != text[loc+1] # never create multi-space in text
#                    and (text[loc] != ' ' # never change a space that is a multi-space in html
#                         or text[loc-1:loc+2] == html[locMap[loc]-1:locMap[loc]+2])]
#        #print "textLocs =", textLocs
#        htmlLocs = [(locMap[loc], dup) for (loc, dup) in textLocs]
#        return (applyCorruption(htmlLocs, html),
#                applyCorruption(textLocs, text),
#                len(textLocs))
#    epubHtml, epubText, corruptionMagnitude = TestCorruption(epubHtml, epubText)
#    assert not '  ' in epubText
#    print "ERROR: corrupted epub text at ca. %d locations" % corruptionMagnitude

    # dump input
    with open(r"C:\tmp\epub.html", "wb") as file:
        file.write(epubHtml)
    
    with open(r"C:\tmp\mobi.html", "wb") as file:
        file.write(kindleHtml)
    
    with open(r"C:\tmp\epub.txt", "wb") as file:
        file.write(epubText.replace(" ", "\n"))
        
    with open(r"C:\tmp\mobi.txt", "wb") as file:
        file.write(kindleText.replace(" ", "\n"))
    print "Cmp:", len(kindleText), len(epubText), kindleText==epubText

    ########################################################################################
    # * do a diff analysis on 'kindleText' vs. 'epubText' resulting in 'matchKindleToEpub'
    ########################################################################################
    
    print "matching text in books..."
    progress_callback(0.50, _('Matching texts between books'))
    matcher = TextMatcher.TextMatcher(a=kindleText, b=epubText)
    matcher.get_matching_blocks() # this may take some time
    mapper = TextMatcher.MatchMapper(matcher)
    print "len(matchKindleToEpub) =", len(matcher.get_matching_blocks())

    #print matchKindleToEpub
    
    ####################################################################################
    # * try to locate annotation in epub using diff analysis; allow for some changes
    ####################################################################################
    
    
    progress_callback(0.70, _('Locate annotations in EPUB'))
    for ka in annoList:
        kiBegin = mapKindleLocToText[ka.begin]
        kiEnd = mapKindleLocToText[ka.end]
        epubBegin, epubEnd, uncertain = mapper.mapBeginEnd(kiBegin, kiEnd)
        
        # redefine empty ranges from "before begin" to "after begin",
        # so they are mapped to the earlier location in case of ambiguity
        if epubEnd is None:
            epubBegin = utf8.adjustL(epubText, epubBegin-1)
        if uncertain and uncertain[1] is None:
            uncertain = (utf8.adjustL(epubText, uncertain[0]-1), None)

        ka.epubTextBegin = epubBegin
        ka.epubTextEnd = epubEnd
        ka.epubTextUncertain = uncertain
        
    del matcher
    del mapper

    ##########################################################################
    # * determine which locations in epubText are relevant for annotations
    ##########################################################################
    
    mapEpubTextToHtml = {}
    for ka in annoList:
        mapEpubTextToHtml[ka.epubTextBegin] = None
        mapEpubTextToHtml[ka.epubTextEnd] = None
        if ka.epubTextUncertain:
            mapEpubTextToHtml[ka.epubTextUncertain[0]] = None
            mapEpubTextToHtml[ka.epubTextUncertain[1]] = None
    
    ##############################################################################################
    #   and create map 'mapEpubTextToHtml to convert locations of epubText to index in epubHtml
    ##############################################################################################
    
    # print "locations =", sorted(mapEpubTextToHtml.keys())
    parser = ParseHtmlToPlainText.ParseHtmlToPlainText(mapEpubTextToHtml, doMapToPlainText=False)
    parser.feed(epubHtml)
    del parser # free resources
    
    ####################################
    # * mark annotations in epubHtml
    ####################################
    
    progress_callback(0.80, _('Markup annotations in EPUB'))
    # get css style rules
    cssOrig = epubZip.read(idToFile['css'])

    newEpubHtml, cssNew = MarkupEpub.mergeAnnotationsIntoHtml(annoList, epubHtml, mapEpubTextToHtml, cssOrig, epubText, **opts)

    ####################################################################
    # * split epubHtml into separate file chunks and write them back
    ####################################################################

    progress_callback(0.90, _('Create new EPUB'))
    newEpubHtml = newEpubHtml.split(filesSeparator)
    assert len(htmlFiles) == len(newEpubHtml)
    replaceDict = dict(zip(htmlFiles, newEpubHtml))
    
    if cssNew != cssOrig:
        replaceDict[idToFile['css']] = cssNew

    copyModifyZip(epubZip, newEpub, replaceDict)

    ##################################################
    # * add or re-use file with annotation-listing
    ##################################################
    
    ##############

# Catch any topaz books lurking as AZWs
def topaz(path, name, ext):
    if "azw" not in ext.lower():
        return False
    with file(os.path.join(path, name)+ext, 'rb') as f:
        return f.read(3) == 'TPZ'
        
# travers kindle documents and look for a file with similar name and suitable extension
def findMatchingBook(epub_path, kindle_root):
    
    # Kindle ebook extension in order of preference
    extensions = ('.mobi', '.prc', '.azw', '.azw3', '.azw4')
    
    target = os.path.splitext( os.path.split(epub_path)[1] )[0]
    matcher = difflib.SequenceMatcher(b=target, autojunk=False)
    
    heap = []
    
    for root, subFolders, files in os.walk(os.path.join(kindle_root, 'documents')):
        for candidate in files:
            candidate, ext = os.path.splitext( candidate )
            if not ext.lower() in extensions:
                continue
            if topaz(root, candidate, ext):
                continue
            matcher.set_seq1(candidate)
            score = matcher.ratio()
            if score < 0.5:
                continue
            score -= 0.001 * extensions.index(ext.lower())
            match = (score, root, candidate, ext)
            if len(heap) < 10:
                heapq.heappush(heap, match)
            else:
                heapq.heappushpop(heap, match)
    if not heap:
        return None
    score, path, name, ext = heapq.nlargest(1, heap)[0]
    return os.path.join(path,name)+ext

def run_calibre_job(*args, **kwargs):
    print "KiNotes: do_calibre_parallel_work(%r, %r)" % (args,kwargs)
    if kwargs.has_key('notification'):
        global progress_callback
        progress_callback = kwargs['notification']
    epub_path, kindle_path, kindle_root, opts = args
    if not kindle_path:
        kindle_path = findMatchingBook(epub_path, kindle_root)
        if not kindle_path:
            raise Exception("book not found on Kindle")
    userannotlog = os.path.join(kindle_root, 'system', 'userannotlog')
    processBook(kindle_path, userannotlog, epub_path, **opts)

def main():
    # check argmuments
    # todo: use sys.argv[1] etc.
    
    # disable all automatic conversion between str and unicode!
    #reload(sys)
    #sys.setdefaultencoding('undefined')
    # detect implicit conversions between str and unicode!
    reload(sys)
    sys.setdefaultencoding('ascii')
        
    epubEbook = unicode(r"TestData\Calibre\Calibre Library\Zahn, Timothy\Night Train to Rigel (1223)\Night Train to Rigel - Zahn, Timothy.epub")
    kindleEbook = unicode(r"TestData\Kindle\documents\Zahn, Timothy\Night Train to Rigel - Zahn, Timothy.mobi")
    userannotlog = unicode(r"TestData\Kindle\system\userannotlog")
    
#    epubEbook = u"K:/Programme/Calibre Portable/Calibre Library/Isaac Asimov/Lucky Starr 01 - David Starr, Space Rang (288)/Lucky Starr 01 - David Starr, Space Rang - Isaac Asimov.epub"
#    kindleEbook = u"D:/PythonWorkspace/KiNotes/TestData/Kindle/documents/Asimov, Isaac/Lucky Starr 01 - David Starr, Space Ranger - Isaac Asimov.mobi"
#    userannotlog = u"D:/PythonWorkspace/KiNotes/TestData/Kindle/system/userannotlog"
#    
#    epubEbook = u'K:\\Programme\\Calibre Portable\\Calibre Library\\Isaac Asimov\\Lucky Starr 02 - Lucky Starr and the Pir (341)\\Lucky Starr 02 - Lucky Starr and the Pir - Isaac Asimov.epub'
#    kindleEbook = u'D:\\PythonWorkspace\\KiNotes\\TestData\\Kindle\\documents\\Asimov, Isaac\\Lucky Starr 02 - Lucky Starr and the Pirates of the Asteroids - Isaac Asimov.mobi'
#    userannotlog = u'D:\\PythonWorkspace\\KiNotes\\TestData\\Kindle\\system\\userannotlog'
#    
#    epubEbook = u'K:\\Programme\\Calibre Portable\\Calibre Library\\Andre Norton\\Beast Master (815)\\Beast Master - Andre Norton.epub'
#    kindleEbook = u'D:\\PythonWorkspace\\KiNotes\\TestData\\Kindle\\documents\\Norton, Andre\\beast master, The - Andre Norton.mobi'
#    userannotlog = u'D:\\PythonWorkspace\\KiNotes\\TestData\\Kindle\\system\\userannotlog'    

#    # while testing make a copy of epub and modify that
#    newEpub = re.sub(r"(\.epub|)$", "_test.epub", epubEbook, re.IGNORECASE)
#    shutil.copyfile(epubEbook, newEpub)
#    epubEbook = newEpub

    processBook(kindleEbook, userannotlog, epubEbook,
                #backup = 4, postfix = u"_anno",
                style_tabstops = False,
                unpackDir = ur"C:\tmp\KiNotes_unpackDir_tmp")

    return 0

if __name__ == '__main__':
    sys.exit(main())

