# -*- coding: utf-8 -*-
assert '…' == '\xe2\x80\xa6', "file encoding error"

# Copyright 2013 Axel Walthelm
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import re
from HTMLParser import HTMLParser
import HtmlTools as html
import ParseUserAnnotLog

# Traverse changeList and html and insert appropriate highlighting spans into changesList.
# Uses three highlighting levels:
HighlightClasses = [ 'KiNoteMark0', 'KiNoteMark', 'KiNoteMarkX' ]
# KiNoteMark0: weakly highlight regions of uncertainty
# KiNoteMark: normal highlight
# KiNoteMarkX: strong highlight where multiple highlight regions overlap

class HighlightHtmlParser(HTMLParser):

    def __init__(self, changesList):
        HTMLParser.__init__(self)
        self.isClosed = False
        self.tagStack = [] # None as soon as XHTML is not well formed any more
        self.changesList = changesList
        self.changesIdx = 0
        self.isBodyTag = False
        self.isNonTextTag = 0
        self.isInBlock = 0
        self.highlightLevel = 0
        self.highlightLevel0 = 0

    # override of internal method HTMLParser.updatepos() to get access to string index location,
    # because self.getpos() returns line and column only
    def updatepos(self, i, j):
        #print "Location: %d-%d '%s'" % (i, j, utf8.forPrint(self.rawdata[i:j]))
        self.locCur = j; # only j is useful as start positon
        return HTMLParser.updatepos(self, i, j)
    
#    # use internal method HTMLParser.check_for_whole_start_tag()
#    # to determine end position of a start tag
#    def getEndOfStartTag(self):
#        locEnd = self.check_for_whole_start_tag(self.locCur)-1
#        assert self.rawdata[locEnd] == '>'
#        return locEnd

    def handle_starttag(self, tag, attrs):
        #print "start tag '%s'" % tag
        assert self.rawdata[self.locCur] == '<'
        if self.tagStack is not None:
            self.tagStack.insert(0, [tag, None, None])
        if tag == 'body':
            self.isBodyTag = True
            self.isNonTextTag = 0
            self.isInBlock = 0
        if tag in html.NOTEXT_TAGS:
            self.isNonTextTag += 1
        if tag in html.BLOCK_TAGS:
            self.isInBlock += 1
            
    def handle_endtag(self, tag):
        #print "end tag '%s'" % tag
        assert self.rawdata[self.locCur] == '<'
        if self.tagStack is not None:
            if not self.tagStack:
                print "WARNING: unexpected closing tag", tag, "in XHTML at", self.locCur
                self.tagStack = None
            elif self.tagStack[0][0] != tag:
                print "WARNING: end tag", self.tagStack[0][0], "missing in XHTML at", self.locCur
            else:
                self.tagStack.pop(0)
        if tag == 'body':
            self.isBodyTag = False
        if tag in html.NOTEXT_TAGS:
            self.isNonTextTag -= 1
        if tag in html.BLOCK_TAGS:
            self.isInBlock -= 1

    def handle_data(self, data):
        self.processData(self.locCur, self.locCur+len(data))
        
    def _highlightIndex(self):
        if self.highlightLevel > 0:
            return 1 + (self.highlightLevel - 1) % 2
        if self.highlightLevel0 > 0:
            return 0
        return -1
    
    def processData(self, htmlFrom, htmlTo):
        #print "RAW: '%s'" % self.rawdata[htmlFrom:htmlTo]

        # ignore text in non-text tags
        # note: no check for self.isInBlock > 0 here;
        # text outside of block is considered illegal, but is tolerated by most html viewers;
        # anchor outside of block seems to be ok with some html specs.
        # So we process the data, update the highlighting status, but insert highlighting spans
        # only when inside of a block or when content contains text (not spaces).
        # This way we avoid irritating span tags between paragraphs.
        if not self.isBodyTag or self.isNonTextTag > 0:
            return
            
        canInsert = self.isInBlock > 0 or re.search(r"\S", self.rawdata[htmlFrom:htmlTo])

        # open highlight span at start of data
        highlightIndex = self._highlightIndex()
        if canInsert and highlightIndex >= 0:
            # try to suppress useless combinations of closing and opening span
            if self.changesIdx > 0 and self.changesList[self.changesIdx-1] == [htmlFrom, (0,0,0), None, '</span>']:
                # suppress close/open with nothing in between
                self.changesList[self.changesIdx-1][3] = ''
            elif self.tagStack and self.tagStack[0][1] == highlightIndex:
                # suppress close/open with a sequence of subtrees in between
                self.changesList[self.tagStack[0][2]][3] = ''
            else:
                hiClass = HighlightClasses[ highlightIndex ]
                self.changesList.insert(self.changesIdx, [htmlFrom, (0,0,0), None, '<span class="%s">' % hiClass])
                self.changesIdx += 1
            
        # while next highlight anchor is within the processed data, update highlighting
        while self.changesIdx < len(self.changesList):
            change = self.changesList[self.changesIdx]
            nextLoc = change[0]
            if nextLoc >= htmlTo:
                break
            assert htmlFrom <= nextLoc # TBD: should we tolerate anchor in invalid region?
            insert = '</span>' if highlightIndex >= 0 else ''
            isStart = change[1][0] == 0
            self.highlightLevel += change[1][1]
            self.highlightLevel0 += change[1][2]
            oldHighlightIndex = highlightIndex
            highlightIndex = self._highlightIndex()
            if highlightIndex != oldHighlightIndex and canInsert:
                if highlightIndex >= 0:
                    hiClass = HighlightClasses[ highlightIndex ]
                    insert += '<span class="%s">' % hiClass
                if isStart:
                    change[3] += insert
                else:
                    change[3] = insert + change[3]
                    
            self.changesIdx += 1
            
        # close highlight span at end of data
        if canInsert and highlightIndex >= 0:
            hiClass = HighlightClasses[ highlightIndex ]
            self.changesList.insert(self.changesIdx, [htmlTo, (0,0,0), None, '</span>'])
            if self.tagStack: # register close for possible later close/open suppression optimization
                self.tagStack[0][1:2] = [highlightIndex, self.changesIdx]
            self.changesIdx += 1
        elif self.tagStack: # unregister last close (if any) for possible later close/open suppression optimization
            self.tagStack[0][1:2] = [None, None]
            
    def handle_entityref(self, name):
        #print "entity '%s'" % name
        self.processData(self.locCur, self.locCur+2+len(name))
        
    def handle_charref(self, number):
        #print "char '%s'" % number
        self.processData(self.locCur, self.locCur+3+len(number))

    def close(self):
        HTMLParser.close(self)
        self.isClosed = True
        if self.tagStack:
            print "WARNING: missing closing tag", self.tagStack[0][0], "in XHTML at end"
            self.tagStack = None

# some tests for developer analysis
if __debug__ and __name__ == '__main__':

    doPrint = False

    def printStartTest():
        import inspect
        stack = inspect.stack()
        print "TEST", " ".join(map(lambda i: stack[i][3], range(len(stack)-2, 0, -1)))
        
    def printEndTest():
        print "OK"

    # TEST: create plain text (no mapping)
    def devTest(html, changesList, expectedHtml, expectedChangesList = None):
        printStartTest()
        
        parser = HighlightHtmlParser(changesList)
        parser.feed(html)
        parser.close()
        
        if doPrint:
            print changesList
            print expectedChangesList
        
        import MarkupEpub
        newHtml = MarkupEpub.applyChanges(changesList, html)
        if doPrint:
            print "newHtml = '%s'" % newHtml.encode('string_escape')
            print "expectedHtml = '%s'" % expectedHtml.encode('string_escape')
            
        assert changesList == expectedChangesList or expectedChangesList is None
        assert newHtml == expectedHtml

        printEndTest()

def testSet1():
    html = "<body> <p>über den grünen &mdash Klee</p></body>"
    #       012345678902345678901245678901234567890123456789
    #       0         10       20       30        40
    ka = ParseUserAnnotLog.KindleAnnotation()
    ka.kind='highlight'
    ka.highlight=True
    changesList = [
        [16, (0,1,0), ka, '<a/>'],
        [18, (1,-1,0), ka, '<a/>'],
    ]
    expectedHtml = '<body> <p>über <a/><span class="KiNoteMark">den</span><a/> grünen &mdash Klee</p></body>'
    expectedChangesList = [
        [16, (0,1,0), ka, '<a/><span class="KiNoteMark">'],
        [18, (1,-1,0), ka, '</span><a/>'],
    ]
    devTest(html, changesList, expectedHtml, expectedChangesList)

def testSet2():
    html = "<body> <p>aber den</p> <p> grauen Klee</p></body>"
    #       01234567890123456789012345678901234567890123456789
    #       0         10        20        30        40
    ka = ParseUserAnnotLog.KindleAnnotation()
    ka.kind='highlight'
    ka.highlight=True
    changesList = [
        [15, (0,1,0), ka, '<a/>'],
        [32, (1,-1,0), ka, '<a/>'],
    ]
    expectedHtml = '<body> <p>aber <a/><span class="KiNoteMark">den</span></p> <p><span class="KiNoteMark"> grauen</span><a/> Klee</p></body>'
    expectedChangesList = [
        [15, (0,1,0), ka, '<a/><span class="KiNoteMark">'],
        [18, (0,0,0), None, '</span>'],
        [26, (0,0,0), None, '<span class="KiNoteMark">'],
        [32, (1,-1,0), ka, '</span><a/>'],
    ]
    devTest(html, changesList, expectedHtml, expectedChangesList)

def testSet3():
    html = "<body><p>aber den</p> <p><i>grauen</i> Klee</p></body>"
    #       01234567890123456789012345678901234567890123456789
    #       0         10        20        30        40
    ka = ParseUserAnnotLog.KindleAnnotation()
    ka.kind='highlight'
    ka.highlight=True
    changesList = [
        [14, (0,1,0), ka, '<a/>'],
        [31, (1,-1,0), ka, '<a/>'],
    ]
    expectedHtml = '<body><p>aber <a/><span class="KiNoteMark">den</span></p> <p><i><span class="KiNoteMark">grau</span><a/>en</i> Klee</p></body>'
    devTest(html, changesList, expectedHtml)
    
def testSet4():
    html = "<body><p>AAA BBB that just</p></body>"
    #       01234567890123456789012345678901234567890123456789
    #       0         10        20        30        40
    ka = ParseUserAnnotLog.KindleAnnotation()
    ka.kind='highlight'
    changesList = [
        [13, (0,1,0), ka, '<a1/>'],
        [17, (0,1,0), ka, '<a2/>'],
        [20, (1,-1,0), ka, '<a1/>'],
        [20, (1,-1,0), ka, '<a2/>'],
    ]
    expectedHtml = '<body><p>AAA <a1/><span class="KiNoteMark">BBB <a2/></span><span class="KiNoteMarkX">that</span><span class="KiNoteMark"><a1/></span><a2/> just</p></body>'
    devTest(html, changesList, expectedHtml)

def testSet5():
    html = "<body><p>X&nbsp;C &amp; A</p></body>"
    #       01234567890123456789012345678901234567890123456789
    #       0         10        20        30        40
    ka = ParseUserAnnotLog.KindleAnnotation()
    ka.kind='highlight'
    changesList = [
        [9, (0,1,0), ka, '<a b/>'],
        [24, (1,-1,0), ka, '<a e/>'],
    ]
    expectedHtml = '<body><p><a b/><span class="KiNoteMark">X&nbsp;C &amp; A</span><a e/></p></body>'
    devTest(html, changesList, expectedHtml)

def testSet6():
    html = "<body><p>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</p></body>"
    #       01234567890123456789012345678901234567890123456789
    #       0         10        20        30        40
    ka = ParseUserAnnotLog.KindleAnnotation()
    ka.kind='highlight'
    changesList = [
        [15, (0,1,0), ka, '<a b1/>'],
        [21, (0,1,0), ka, '<a b2/>'],
        [26, (1,-1,0), ka, '<a e2/>'],
        [32, (1,-1,0), ka, '<a e1/>'],
    ]
    expectedHtml = '<body><p>&nbsp;<a b1/><span class="KiNoteMark">&nbsp;<a b2/></span><span class="KiNoteMarkX">&nbsp;</span><span class="KiNoteMark"><a e2/>&nbsp;</span><a e1/>&nbsp;</p></body>'
    devTest(html, changesList, expectedHtml)

def testSet7():
    html = "<body>\n<p>No <i><b>I</b></i> in <i>team</i></p></body>"
    #       01234567890123456789012345678901234567890123456789
    #       0         10        20        30        40
    ka = ParseUserAnnotLog.KindleAnnotation()
    ka.kind='highlight'
    changesList = [
        [6, (0,1,0), ka, '<a dummy/>'],
    ]
    expectedHtml = '<body><a dummy/>\n<p><span class="KiNoteMark">No <i><b><span class="KiNoteMark">I</span></b></i> in </span><i><span class="KiNoteMark">team</span></i></p></body>'
    devTest(html, changesList, expectedHtml)

def testSet8():
    html = "<body>\n<p>No <i><b>I</b></i> in t<i>ea</i>m</p></body>"
    #       01234567890123456789012345678901234567890123456789
    #       0         10        20        30        40
    ka = ParseUserAnnotLog.KindleAnnotation()
    ka.kind='highlight'
    changesList = [
        [6, (0,1,0), ka, '<a dummy/>'],
    ]
    expectedHtml = '<body><a dummy/>\n<p><span class="KiNoteMark">No <i><b><span class="KiNoteMark">I</span></b></i> in t<i><span class="KiNoteMark">ea</span></i>m</span></p></body>'
    devTest(html, changesList, expectedHtml)

def testSet9():
    html = "<body><p>A<i>B</i><b>C</b><u>D</u>E</p></body>"
    #       01234567890123456789012345678901234567890123456789
    #       0         10        20        30        40
    ka = ParseUserAnnotLog.KindleAnnotation()
    ka.kind='highlight'
    changesList = [
        [9, (0,1,0), ka, '<a dummy/>'],
    ]
    expectedHtml = '<body><p><a dummy/><span class="KiNoteMark">A<i><span class="KiNoteMark">B</span></i><b><span class="KiNoteMark">C</span></b><u><span class="KiNoteMark">D</span></u>E</span></p></body>'
    devTest(html, changesList, expectedHtml)

def testSet10():
    html = "<body><p>A<i>B</i><b>C</b><u>D</u>E</p></body>"
    #       01234567890123456789012345678901234567890123456789
    #       0         10        20        30        40
    ka = ParseUserAnnotLog.KindleAnnotation()
    ka.kind='highlight'
    changesList = [
        [9, (0,1,0), ka, '<a 1/>'],
        [21, (0,1,0), ka, '<a 2/>'],
    ]
    expectedHtml = '<body><p><a 1/><span class="KiNoteMark">A</span><i><span class="KiNoteMark">B</span></i><b><span class="KiNoteMark"><a 2/></span><span class="KiNoteMarkX">C</span></b><u><span class="KiNoteMarkX">D</span></u><span class="KiNoteMarkX">E</span></p></body>'
    devTest(html, changesList, expectedHtml)

def testSet11():
    html = "<body><p>A<i>B</i></p> <p><b>C</b><u>D</u>E</p></body>"
    #       01234567890123456789012345678901234567890123456789
    #       0         10        20        30        40
    ka = ParseUserAnnotLog.KindleAnnotation()
    ka.kind='highlight'
    changesList = [
        [9, (0,1,0), ka, '<a 1/>'],
        [29, (0,1,0), ka, '<a 2/>'],
    ]
    expectedHtml = '<body><p><a 1/><span class="KiNoteMark">A</span><i><span class="KiNoteMark">B</span></i></p> <p><b><span class="KiNoteMark"><a 2/></span><span class="KiNoteMarkX">C</span></b><u><span class="KiNoteMarkX">D</span></u><span class="KiNoteMarkX">E</span></p></body>'
    devTest(html, changesList, expectedHtml)

def testSet20_Mark0():
    html = "<body> <p>aber den</p> <p> grauen Klee</p></body>"
    #       01234567890123456789012345678901234567890123456789
    #       0         10        20        30        40
    ka = ParseUserAnnotLog.KindleAnnotation()
    ka.kind='highlight'
    ka.highlight=True
    changesList = [
        [15, (0,0,1), ka, '<a/>'],
        [32, (1,0,-1), ka, '<a/>'],
    ]
    expectedHtml = '<body> <p>aber <a/><span class="KiNoteMark0">den</span></p> <p><span class="KiNoteMark0"> grauen</span><a/> Klee</p></body>'
    expectedChangesList = [
        [15, (0,0,1), ka, '<a/><span class="KiNoteMark0">'],
        [18, (0,0,0), None, '</span>'],
        [26, (0,0,0), None, '<span class="KiNoteMark0">'],
        [32, (1,0,-1), ka, '</span><a/>'],
    ]
    devTest(html, changesList, expectedHtml, expectedChangesList)

def testSet21_MarkAndMark0():
    html = "<body> <p>aber den</p> <p> grauen Klee</p></body>"
    #       01234567890123456789012345678901234567890123456789
    #       0         10        20        30        40
    ka = ParseUserAnnotLog.KindleAnnotation()
    ka.kind='highlight'
    ka.highlight=True
    changesList = [
        [10, (0,0,1), ka, '<a u/>'],
        [15, (0,1,0), ka, '<a/>'],
        [32, (1,-1,0), ka, '<a/>'],
        [37, (1,0,-1), ka, '<a u/>'],
    ]
    expectedHtml = '<body> <p><a u/><span class="KiNoteMark0">aber <a/></span><span class="KiNoteMark">den</span></p> <p><span class="KiNoteMark"> grauen</span><span class="KiNoteMark0"><a/> Klee</span><a u/></p></body>'
    devTest(html, changesList, expectedHtml)

def testSet22_MarkAndMark0():
    html = "<body> <p>aber den</p> <p> grauen Klee</p></body>"
    #       01234567890123456789012345678901234567890123456789
    #       0         10        20        30        40
    ka = ParseUserAnnotLog.KindleAnnotation()
    ka.kind='highlight'
    ka.highlight=True
    changesList = [
        [15, (0,0,1), ka, '<a u/>'],
        [15, (0,1,0), ka, '<a/>'],
        [32, (1,-1,0), ka, '<a/>'],
        [32, (1,0,-1), ka, '<a u/>'],
    ]
    expectedHtml = '<body> <p>aber <a u/><span class="KiNoteMark0"><a/></span><span class="KiNoteMark">den</span></p> <p><span class="KiNoteMark"> grauen</span><span class="KiNoteMark0"><a/></span><a u/> Klee</p></body>'
    devTest(html, changesList, expectedHtml)

def testSet23_MarkAndMark0():
    html = "<body> <p>aber den</p> <p> grauen Klee</p></body>"
    #       01234567890123456789012345678901234567890123456789
    #       0         10        20        30        40
    ka = ParseUserAnnotLog.KindleAnnotation()
    ka.kind='highlight'
    ka.highlight=True
    changesList = [
        [15, (0,0,1), ka, '<a u/>'],
        [15, (0,1,0), ka, '<a/>'],
        [32, (1,0,-1), ka, '<a/>'],
        [32, (1,-1,0), ka, '<a u/>'],
    ]
    expectedHtml = '<body> <p>aber <a u/><span class="KiNoteMark0"><a/></span><span class="KiNoteMark">den</span></p> <p><span class="KiNoteMark"> grauen<a/></span><a u/> Klee</p></body>'
    devTest(html, changesList, expectedHtml)

if __debug__ and __name__ == '__main__':
    # run tests
    testSet1()
    testSet2()
    testSet3()
    testSet4()
    testSet5()
    testSet6()
    testSet7()
    testSet8()
    testSet9()
    testSet10()
    testSet11()
    testSet20_Mark0()
    testSet21_MarkAndMark0()
    testSet22_MarkAndMark0()
    testSet23_MarkAndMark0()
