# -*- coding: utf-8 -*-
assert '…' == '\xe2\x80\xa6', "file encoding error"

# Copyright 2013 Axel Walthelm
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import re
import difflib
import collections

#import Utf8Tools as utf8
import HtmlTools as html
import CssStyles
import HighlightHtmlParser

# Convert an integer to a string of uppercase letters (A-Z, AA-ZZ, AAA-ZZZ, etc.)
def intToAlpha(i):
    if i < 0 or int(i) != i:
        raise ArithmeticError()
    a = ''
    while i > 0:
        a = chr(ord('A')+(i-1)%26) + a
        i = int((i-1)/26)
    return a

# Inverse of intToAlpha
def alphaToInt(a):
    if not re.match(r"^[A-Z]*$", a):
        raise ArithmeticError()
    i = 0;
    for p in range(len(a)):
        i = i*26 + (ord(a[p]) - ord('A')+1)
    return i

def updateStyleSheet(css, styleIdOrText):
    
    # stepwise reduce css until only class selectors are left
    classes = css
    classes = re.sub(r"[\s\r\n]+", " ", classes) # newlines and spaces
    classes = re.sub(r"/\*.*?\*/", "", classes) # comments
    classes = re.sub(r"'.*?[^\\]'", "''", classes) # string ''
    classes = re.sub(r'".*?[^\\]"', '""', classes) # string ""
    classes = re.sub(r"{.*?}", "{}", classes) # content in {...}
    classes = re.sub(r"[^._a-zA-Z0-9-]+", '\n', '\n'+classes+'\n')
    # todo: handle e.g. a[class~="KiNoteEnd"] => KiNoteEnd
    # todo: handle e.g. a[class="KiNoteEnd"] => KiNoteEnd
    # todo: handle e.g. a[class="KiNoteStart KiNoteEnd"] => KiNoteStart, KiNoteEnd
    # tbd: handle class^="KiNote"
    # tbd: handle class$="KiNote"
    # tbd: handle class*="KiNote"
    classes = re.findall(r"\n\.(-?[_a-zA-Z]+[_a-zA-Z0-9-]*)\n", classes)    
    
    if not set(classes).isdisjoint(CssStyles.KINOTE_CLASSES):
        return css # looks like style was already set; better not touch CSS

    # add style rules
    newline = "\r\n" if "\r\n" in css else "\n"
    styleTexts = [s for i, n, s in CssStyles.KINOTE_STYLES if styleIdOrText in (i, n)]
    styleText = styleTexts[0] if styleTexts else styleIdOrText
    css = re.sub(r"[\r\n]*$", newline, css) + re.sub(r"[\r\n]+", newline, styleText)
    return css
    
def formatUnicodeTextForHtml(text, replaceDoubleQuotes=None):
    assert isinstance( text, unicode )
    return formatUtf8TextForHtml(text.encode('utf-8'), replaceDoubleQuotes).decode('utf-8')

def formatUtf8TextForHtml(text, replaceDoubleQuotes=None):
    assert isinstance( text, bytes )
    #print "formatPlainTextForHtml => '%s'" % text
    # replace html/xhtml special characters by entities
    text = text.replace("&", "&amp;") # must be done first!
    text = text.replace("<", "&lt;")
    text = text.replace(">", "&gt;")
    if replaceDoubleQuotes:
        text = text.replace('"', "&quot;")
    # make all newlines to \n
    text = text.replace("\r\n", "\n").replace("\r", "\n")
    # convert weird ascii-spaces like tab, escape, etc. to ' ', but keep \n
    text = re.sub(r"[\0-\11\13-\37]", ' ', text)
    # display newlines in text by a special character (followed by a space when reasonable)
    # ↵ is &crarr;, but the named entity may not be valid for XHTML or some tools
    text = text.replace("\n ", "↵ ") # replace newline that don't need an extra space
    if text[-1:] == '\n': # replace newline at end (must not have an extra space)
        text = text[:-1] + "↵" # note: can't use re.sub(r"\n$",...) because $ also matches at \n before end of string
    text = text.replace("\n", "↵ ") # all remaining newline need an extra space
    # make sure spaces are preserved
    # note: XHTML and/or Sigil is ok with &#160; as a non-breaking space, but not with '\xc2\xa0' or &nbsp;
    # (&#32; seems to work well as a non-(re-)movable space for Sigil, but may be a removed by other tools)
    text = re.sub(r"^ ", "&#160;", text)
    text = re.sub(r" $", "&#160;", text)
    while "  " in text:
        text = text.replace("  ", "&#160; ")
    #print "formatPlainTextForHtml <= '%s'" % text
    return text

def markupHtml(annoList, epubHtml, mapEpubTextToHtml, epubText,
        hasHighlight = True,
        hasAnchors = True,
        hasStartTab = False,
        hasEndTab = True,
        hasHighlightTab = False,
        hasId = False,
        nameUncertainPostfix = " ?"):
    
    # note: the annotation markup format has undergone some changes;
    # the comments here will be better understood with this in mind.
    # for example for a long time a was trying to use <a class="..."></a> tag
    # to markup begin and end location of annotations, which I gave up on when I found
    # annotations may be on links in EPUB, but a-tag is theoretically not allowed
    # within a link - and some annoying tools feel obliged to enforce that sometimes;
    # then I fell back to usig <span> tags instead of <a>, which in turn made it
    # necessary to always have some text in the tag, because some stupid tools like Sigil
    # tend to optimize away empty span tags, even when they obviously have a special style...
    # similar try-and-error evolution happened to the sub-tags to mark up
    # text in the end end location tags...
    
    def makeId(ka):
        # todo?: check for collisions in html - and don't add notes twice!
        return ka.id # return ka.id+"_"+ka.kind+("_%d_%d" % (ka.begin, ka.end))

    inserts = {}
    # collection of locations to insert before(!) and corresponding insert text
    # for each location we keep a list of inserts; order is relevant;
    # each insert item is in the form of [ loc, highlight, ka, text ] where
    # loc: location before which to insert
    # type: tuple used for highlighting and to decide if insert is before or after loc
    #   type[0] in [0,1]: insert after if type[0] else insert before
    #   type[1] in [-1,0,1]: change normal highlighting by given increment
    #   type[2] in [-1,0,1]: change weak highlighting by given increment
    # ka: KindleAnnotation that caused this insert; may be None for inserts due to highlighting
    # text: text to insert; usually html tags like '<a ...></a>', '<span ...>', or '</span>'
    #
    def addInsert(inserts, item, doAppend):
        # since we are not yet working on the html text, we have
        # to expect inserts like '</p> <p>' anywhere within the text;
        # that means an insert after a character is not the same as
        # an insert before the next character.
        loc = (item[0], item[1][0])
        if inserts.has_key(loc):
            if doAppend:
                inserts[loc].append( item )
            else:
                inserts[loc].insert(0, item)
        else:
            inserts[loc] = [ item ]

    # remove old annotations from epubHtml; adjust mapEpubTextToHtml
    # analyze epubHtml for existing id (incl. a.name), tabindex, notes/highlights, bookmarks
    # todo: round trip concept
    
    # add start and end anchors to changeList
    nrs = [0, 0, 0, 0] # 0:note 1:highlight 2:bookmark 3:tabindex
    def count(nrs, i):
        nrs[i] += 1
        return nrs[i]
    mapEpubTextToHtml[None] = None
    for ka in annoList:
        begin = mapEpubTextToHtml[ka.epubTextBegin]
        end = mapEpubTextToHtml[ka.epubTextEnd]
        # note: end is None means to insert all text behind begin

        name = None
        if ka.kind in ('note', 'highlight'):
            isNote = ka.kind == 'note'
            
            # start anchor
            if isNote:
                clas = "KiNoteStarts"
                name = "%d" % count(nrs,0)
            else:
                clas = "KiNoteStart"
                name = "'%d" % count(nrs,1) # intToAlpha(count(nrs,1))
            if ka.epubTextUncertain:
                name += nameUncertainPostfix
            id_ = ' id="%s_s"' % makeId(ka) if hasId else ''
            tab = ' tabindex="%d"' % count(nrs,3) if hasStartTab else ''
            text = '<span class="%s"%s%s title="%s"><b>[</b></span>' % (clas, id_, tab, name) if hasAnchors else ''
            if end is not None:
                addInsert(inserts, [begin, (0,1,0), ka, text], 0)
            else:
                addInsert(inserts, [begin, (1,0,0), ka, text], 0)
            
            # end anchor
            id_ = ' id="%s_e"' % makeId(ka) if hasId else ''
            tab = ' tabindex="%d"' % count(nrs,3) if hasEndTab else ''
            if isNote:
                clas = "KiNoteEnds"
                note = "<u>%s</u>" % formatUtf8TextForHtml(ka.note)
            else:
                clas = "KiNoteEnd"
                note = ""
            text = '<span class="%s"%s%s title="%s"><b>]</b>%s</span>' % (clas, id_, tab, name, note) if hasAnchors else ''
            if end is not None:
                addInsert(inserts, [end, (1,-1,0), ka, text], 1)
            else:
                addInsert(inserts, [begin, (1,0,0), ka, text], 1)
            
        elif ka.kind == 'bookmark':
            id_ = ' id="%s"' % makeId(ka) if hasId else ''
            tab = ' tabindex="%d"' % count(nrs,3) if hasStartTab or hasEndTab else ''
            name = "#%d" % count(nrs,2)
            if ka.epubTextUncertain:
                name += nameUncertainPostfix
            text = '<span class="KiNoteBookmark%s%s title="%s"><b>[]</b></span>' % (id_, tab, name) if hasAnchors else ''
            addInsert(inserts, [begin, (0,0,0), ka, text], 0)
            
        if name is None:
            continue

        # uncertain begin
        if ka.epubTextUncertain:
            unc = ka.epubTextUncertain
            print "WARNING: uncertain annotation", name
            begin = mapEpubTextToHtml[unc[0]]
            end = mapEpubTextToHtml[unc[1]]

            text = '<span class="KiNoteFuzzyStart" title="%s"><b>{</b></span>' % (name) if hasAnchors else ''
            if end is not None:
                addInsert(inserts, [begin, (0,0,1), ka, text], 0)
            else:
                addInsert(inserts, [begin, (1,0,0), ka, text], 0)

            # yet another Sigil normalization quirk:
            # '<a><s>Cor<s>e-l</s>ine</s></a>' is normalized to
            # '<a><s>Cor</s>e-line</a>', i.e. Sigil is allergic to directly nested s tags in anchors
            # so instad of makeDiffTextHtml (using <s>/<u>) we need to use makeDelInsTextHtml (using <del>/<ins>)
            # note that we won't use <del> to mark fuzzy-end text, because
            # a) we don't know exactly how to disable the <del> tag using styles, because
            #    <del> is not necessarily of style 'text-decoration: line-through;'
            # b) Sigil is also allergic to nested <del>
            diffText = makeDelInsTextHtml(ka.text, epubText[unc[0]:unc[1]+1] if unc[1] is not None else '')
            text = '<span class="KiNoteFuzzyEnd" title="%s"><b>}</b><i>%s</i></span>' % (name, diffText) if hasAnchors else ''
            if end is not None:
                addInsert(inserts, [end, (1,0,-1), ka, text], 1)
            else:
                addInsert(inserts, [begin, (1,0,0), ka, text], 1)

    # sort inserts by location key and flatten into a single list of inserts
    changeList = [insert for (loc, insList) in sorted(inserts.items()) for insert in insList]

    # insert highlighting spans into changeList
    if hasHighlight:
        parser = HighlightHtmlParser.HighlightHtmlParser(changeList)
        parser.feed(epubHtml)
        parser.close()
        
    if hasHighlightTab:
        for i in xrange(len(changeList)):
            while re.search(r'<span class="KiNoteMark[^"]*">', changeList[i][-1]):
                changeList[i][-1] = re.sub(
                    r'(<span class="KiNoteMark[^"]*")>',
                    r'\1 tabindex="%d">' % count(nrs,3),
                    changeList[i][-1], 1)

    # apply all changes to epubHtml
    newEpubHtml = applyChanges(changeList, epubHtml)
    return newEpubHtml
    
def applyChanges(changesList, oldHtml):
    newHtml = []
    locCur = 0
    for loc, typ, ka, text in changesList:
        assert len(typ) == 3 and typ[0] in (0, 1)
        if typ[0]:
            loc = html.utf8HtmlNext( oldHtml, loc )
        assert loc >= locCur
        newHtml.append(oldHtml[locCur:loc])
        newHtml.append(text)
        locCur = loc
    newHtml.append(oldHtml[locCur:])
    return ''.join(newHtml)

def mergeAnnotationsIntoHtml(annoList, epubHtml, mapEpubTextToHtml, css, epubText, **opts):

    def getopt(name, default):
        value = opts[name] if opts.has_key(name) else None
        return value if value is not None else default
        
    # add styles to css if they do not yet exist
    css = updateStyleSheet(css, getopt('style', CssStyles.KINOTE_STYLES[0][0]))

    # check which annotations are already present in epubHtml and remove them from annoList
    # TODO

    # modify html
    hasTab = getopt('style_tabstops', True)
    hasAnchors = getopt('style_anchors', True)
    epubHtml = markupHtml(annoList, epubHtml, mapEpubTextToHtml, epubText,
        hasHighlight = getopt('style_highlighting', True),
        hasAnchors = hasAnchors,
        hasStartTab = False,
        hasEndTab = hasTab and hasAnchors,
        hasHighlightTab = hasTab and not hasAnchors,
        hasId = getopt('style_id_timestamp', False))
    return epubHtml, css

def makeDelInsTextHtml(textOld, textNew):
    return ( makeDiffTextHtml(textOld, textNew)
                .replace('<u>', '<ins>')
                .replace('</u>', '</ins>')
                .replace('<s>', '<del>')
                .replace('</s>', '</del>') )
    
def makeDiffTextHtml(textOld, textNew):
    if not textOld:
        return ''
    if max(len(textOld), len(textNew)) > 10000:
        return '<s>%s</s>' % formatUtf8TextForHtml(textOld)
    #return '<s>%s</s>%s' % (formatPlainTextForHtml(textOld), formatPlainTextForHtml(textNew))

    # to prevent difflib to match partial utf-8 characters, we better work in unicode
    textOld = textOld.decode('utf-8')
    textNew = textNew.decode('utf-8')
    
    matcher = difflib.SequenceMatcher(a=textOld, b=textNew)
    underlines = collections.defaultdict(int)
    strikeouts = collections.defaultdict(int)
    for op in matcher.get_opcodes():
        #print op
        if op[0] == 'insert':
            # underline locations where text was added
            assert op[1] == op[2]
            underlines[max(0, op[1]-1)] += 1
            underlines[max(0, op[2]+1)] -= 1
        elif op[0] in ('delete', 'replace'):
            # strike through location where text was removed or changed
            grow = 0 if op[2] - op[1] > 1 else 1
            strikeouts[max(0, op[1]-grow)] += 1
            strikeouts[max(0, op[2]+grow)] -= 1
    #print "underlines =", underlines.items()
    #print "strikeouts =", strikeouts.items()
    locs = sorted(set(strikeouts.iterkeys()).union(underlines.iterkeys()))
    #print locs
    
    htmlOld = []
    locCur = 0
    s_cnt = 0
    u_cnt = 0
    tos_is_u = False # top of stack is <u>?
    for loc in locs:
        htmlOld.append(formatUnicodeTextForHtml(textOld[locCur:loc]))
        if underlines[loc]:
            u_cnt += underlines[loc]
            if 0 < u_cnt == underlines[loc]:
                htmlOld.append('<u>')
                tos_is_u = True
            elif u_cnt == 0:
                if tos_is_u:
                    htmlOld.append('</u>')
                    tos_is_u = False
                else:
                    htmlOld.append('</s></u><s>')
        if strikeouts[loc]:
            s_cnt += strikeouts[loc]
            if 0 < s_cnt == strikeouts[loc]:
                htmlOld.append('<s>')
                tos_is_u = False
            elif s_cnt == 0:
                if tos_is_u:
                    htmlOld.append('</u></s><u>')
                else:
                    htmlOld.append('</s>')
                    tos_is_u = u_cnt > 0
        #print "###1 %d s:%d u:%d %s" % (loc, s_cnt, u_cnt, htmlOld)
        locCur = loc
    htmlOld.append(formatUnicodeTextForHtml(textOld[locCur:]))
    #print "###2 %d s:%d u:%d %s" % (locCur, s_cnt, u_cnt, htmlOld)
    htmlOld = ''.join(htmlOld).replace('<u></u>', '').replace('<s></s>', '')
    
    return htmlOld.encode('utf-8')

# some tests for developer analysis
if __debug__ and __name__ == '__main__':

    doPrint = True

    def printStartTest():
        import inspect
        stack = inspect.stack()
        print "TEST", " ".join(map(lambda i: stack[i][3], range(len(stack)-2, 0, -1)))
        
    def printEndTest():
        print "OK"

    # TEST: create plain text (no mapping)
    def devTestIntToAlpha():
        printStartTest()
    
        tests = {0: '', 1: 'A', 2: 'B', 3: 'C', 26: 'Z', 27: 'AA', 28: 'AB', 52: 'AZ', 53: 'BA', 702: 'ZZ', 703: 'AAA'}
        for i in tests.keys():
            a = intToAlpha(i)
            i2 = alphaToInt(a)
            if doPrint:
                print "'%s'" % a, i, i2
            assert tests[i] == a
            assert i == i2
    
        for i in range(0, 10000000, 100):
            a = intToAlpha(i)
            i2 = alphaToInt(a)
            #print "'%s'" % a, i
            assert i == i2
        
        try:
            intToAlpha(-1)
            assert False
        except ArithmeticError:
            pass
        
        try:
            intToAlpha(2.5)
            assert False
        except ArithmeticError:
            pass
    
        try:
            alphaToInt(" A")
            assert False
        except ArithmeticError:
            pass
        
        printEndTest()

    def devTest_formatPlainTextForHtml():
        printStartTest()
        def test(text, htmlExpected, quote=False):
            html = formatUtf8TextForHtml(text, quote)
            if doPrint:
                print ""
                print "text: '%s'" % text.encode('string_escape')
                print "html: '%s'" % html.encode('string_escape')
                if html != htmlExpected:
                    print "      '%s' was expected <- ERROR" % htmlExpected.encode('string_escape')
            assert html == htmlExpected
            
        test("", "")
        test(" ", "&#160;")
        test("\t", "&#160;")
        test("  ", "&#160;&#160;")
        test(" _ ", "&#160;_&#160;")
        test(" _ _ ", "&#160;_ _&#160;")
        test("\t_\0_\1", "&#160;_ _&#160;")
        test("  _  _  ", "&#160; _&#160; _ &#160;")
        
        test("-\0+", "- +")
        test("-\1+", "- +")
        test("-\10+", "- +")
        test("-\11+", "- +")
        test("-\12+", "-↵ +")
        test("-\14+", "- +")
        test("-\15+", "-↵ +")
        test("-\16+", "- +")
        test("-\x1f+", "- +")
        test("- +", "- +")
        test("-!+", "-!+")
        test("\n", "↵")
        test("\r\n", "↵")
        test("\r", "↵")
        test("\n\r", "↵ ↵")
        test("\n\n\n", "↵ ↵ ↵")
        test("X\n \r", "X↵ ↵")
        test("\n\r\r\n\n", "↵ ↵ ↵ ↵")
        test("↵", "↵")
        test("↵\n", "↵↵")
        test("\n↵", "↵ ↵")
        
        test('"', '"')
        test('"', '&quot;', True)
        test('<>&', '&lt;&gt;&amp;')
        test('!"#$%\'()*+,-./0123456789:;=?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~\x7f',
             '!"#$%\'()*+,-./0123456789:;=?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~\x7f')
        test('äöüßÄÖÜ', 'äöüßÄÖÜ')
        
        printEndTest()
        
    def devTest_makeDiffTextHtml():
        printStartTest()
        
        def test(textOld, textNew, htmlExpected):
            if doPrint:
                print ""
                print " old: '%s'" % textOld.encode('string_escape')
                print " new: '%s'" % textNew.encode('string_escape')
            html = makeDiffTextHtml(textOld, textNew)
            if doPrint:
                print ""
                print " old: '%s'" % textOld.encode('string_escape')
                print " new: '%s'" % textNew.encode('string_escape')
                print "html: '%s'" % html.encode('string_escape')
                if html != htmlExpected:
                    print "      '%s' was expected <- ERROR" % htmlExpected.encode('string_escape')
            html.decode('utf-8') # must be valid utf-8
            assert html == htmlExpected
            
        test('',
             '', '')
        test('a',
             '', '<s>a</s>')
        test('a',
             'a', 'a')
        test('a',
             'b', '<s>a</s>')
        test('ab',
             'a', '<s>ab</s>')
        test('abb',
             'a', 'a<s>bb</s>')
        test('ba',
             'a', '<s>ba</s>')
        test('bba',
             'a', '<s>bb</s>a')
        test('a',
             'bba', '<u>a</u>')
        test('a',
             'abb', '<u>a</u>')
        test('aa',
             'bbaa', '<u>a</u>a')
        test('aa',
             'aabb', 'a<u>a</u>')
        test('aa',
             'aabb', 'a<u>a</u>')
            
        test('abcdefg',
             'axbceyg', '<u>ab</u><s>cdefg</s>')
        test('aabbccddeeffgg',
             'aaxxbbcceeyygg', 'a<u>ab</u>bcc<s>dd</s>ee<s>ff</s>gg')

        test('cca',
             'cabb', '<s>cc</s><u>a</u>')
             
        test('ä',
             'ü', '<s>ä</s>')
        test('bäää',
             'büää', '<u>bä</u><s>ää</s>')

        import random
        rand = random.Random(87534)
        for _ in xrange(10000):
            old = (u''.join([rand.choice(u"abcäöü") for _ in xrange(rand.randint(1,15))])).encode('utf-8')
            new = (u''.join([rand.choice(u"abcäöü") for _ in xrange(rand.randint(1,15))])).encode('utf-8')
            html = makeDiffTextHtml(old, new)
            old2 = re.sub(r"<.*?>", '', html)
            assert old2 == old
            was = ''
            now = html
            #print "/html ", now
            while now != was:
                was = now
                now = re.sub(r"<(.*?)>.*?</\1>", '', was)
                #print "//// ", now
            #print "/////", now
            assert '<' not in now
            assert '>' not in now
            
        
        printEndTest()

if __debug__ and __name__ == '__main__':
    # run tests
    devTestIntToAlpha()
    devTest_formatPlainTextForHtml()
    devTest_makeDiffTextHtml()

