# -*- coding: utf-8 -*-
assert '…' == '\xe2\x80\xa6', "file encoding error"

# Copyright 2013 Axel Walthelm
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import re
from HTMLParser import HTMLParser
import Utf8Tools as utf8
import HtmlTools as html

class ParseHtmlToPlainText(HTMLParser):

    def __init__(self, locationsMap = None, doMapToPlainText = True):
        HTMLParser.__init__(self)
        self.isClosed = False
        self.nextLocationInit(locationsMap, doMapToPlainText)
        self.plainTextList = [' '] # plain text always starts with space
        self.plainTextLength = 1 # number of characters in plainTextList
        self.isBodyTag = False
        self.isNonTextTag = 0

    def nextLocationInit(self, locationsMap, doMapToPlainText):
        self.locationsMap = locationsMap
        self.doMapToPlainText = doMapToPlainText
        self.nextLocation = None
        if locationsMap is not None:
            self.sortedLocKeys = sorted(locationsMap.iterkeys(), reverse=True)
            for locKey in self.sortedLocKeys:
                locationsMap[locKey] = None
            if self.sortedLocKeys:
                self.nextLocation = self.sortedLocKeys.pop()
                if self.nextLocation is None: # None as a location value is legal (and maps to None)
                    self.nextLocation = self.sortedLocKeys.pop()
                if self.nextLocation < 0:
                    raise IndexError("invalid location", self.nextLocation)

    def nextLocationFound(self, plainTextLoc):
        assert self.nextLocation is not None
        # update locations map
        #print "MAP loc %d to %d" % (self.nextLocation, plainTextLoc)
        self.locationsMap[self.nextLocation] = plainTextLoc
        self.nextLocation = self.sortedLocKeys.pop() if self.sortedLocKeys else None
        
    # override of internal method to get access to string index location,
    # because self.getpos() returns line and column only
    def updatepos(self, i, j):
        #print "Location: %d-%d '%s'" % (i, j, utf8.forPrint(self.rawdata[i:j]))
        self.locCur = j;
        return HTMLParser.updatepos(self, i, j)
    
    # the trick to use updatepos to supply index into self.rawdata does only work for a single
    # processing of HTMLParser.goahead, so our html must be processed in a single feed() call;
    # therefore we also override methods feed() and close() here,
    # using internal knowledge about their implementation in HTMLParser.
    def feed(self, data):
        assert not self.isClosed
        self.rawdata = data
        self.goahead(1)
        self.isClosed = True
        
        # clean up mapping
        if self.locationsMap:
            if self.doMapToPlainText:
                for key, value in self.locationsMap.iteritems():
                    # value None indicates unprocessed locations
                    # value just behind end of text indicates space pointing to next char (which did not come)
                    # both cases mean last character
                    if value is None and key is not None or value >= self.plainTextLength:
                        self.locationsMap[key] = self.plainTextLength-1 # no utf8.adjust, plainText ends with a ' '
            else:
                for key, value in self.locationsMap.iteritems():
                    # value None indicates unprocessed locations
                    # map it to last valid data location in html
                    if value is None and key is not None:
                        self.locationsMap[key] = utf8.adjustL(data, self.lastHtmlTo-1)
        
    def close(self):
        assert False # no need to close
        assert self.isClosed
                        
    def handle_starttag(self, tag, attrs):
        #print "start tag '%s'" % tag
        if tag == 'body':
            self.isBodyTag = True
            self.isNonTextTag = 0
        if tag in html.NOTEXT_TAGS:
            self.isNonTextTag += 1
        self.checkWordBreak(tag)
            
    def handle_endtag(self, tag):
        #print "end tag '%s'" % tag
        if tag == 'body':
            self.isBodyTag = False
        if tag in html.NOTEXT_TAGS:
            self.isNonTextTag -= 1
        self.checkWordBreak(tag)
        
    # normalize spaces
    def cleanData(self, data):
        # remove all invisible spaces
        data = re.sub( utf8.RE_ZERO_WIDTH_SPACE, '', data )
        # collapse whitespaces to single ' '
        data = re.sub( utf8.RE_WHITESPACES, ' ', data )
        # collapse leading white space into whitespace of previous data, if any
        if data.startswith(' ') and self.plainTextList[-1].endswith(' '):
            data = data[1:]
        return data

    # calculates next plaint text location if the given cData (usually returned by cleanData())
    # is appended to the collected plain text, but not counting whitespace at the end
    def plainTextCursorLoc(self, cData = ''):
        assert not (self.plainTextList[-1].endswith(' ') and cData.startswith(' ')), "cData is not clean"
        loc = self.plainTextLength + len(cData)
        # plain text, cData => correction of loc, where X means r'\S+' and _ means r' '
        # X ,  X  => -0
        # X ,  X_ => -1
        # X , _X  => -0
        # X , _X_ => -1
        # X_,  X  => -0
        # X_,  X_ => -1
        # X_, _X  => -1 # can't happen
        # X_, _X_ => -2 # can't happen
        # X ,     => -0
        # X ,  _  => -1
        # X_,     => -1
        # X_,  _  => -1 # can't happen
        if cData.endswith(' ') or not cData and self.plainTextList[-1].endswith(' '):
            loc -= 1
        return loc

    # enforces a single space at the end of the collected plain text
    def checkWordBreak(self, tag):
        if tag not in html.NOSPACE_TAGS and not self.plainTextList[-1].endswith(' '):
            self.plainTextList.append(' ')
            self.plainTextLength += 1

    def handle_data(self, data):
        self.processData(data, None, self.locCur, self.locCur+len(data))
    
    def processData(self, data, origData, htmlFrom, htmlTo):
        #print type(data)
        assert isinstance(data, bytes)
        #print "@%d-%d data '%s'" % (htmlFrom, htmlTo, data)
        #print "RAW: '%s'" % self.rawdata[htmlFrom:htmlTo]
        # if data is already modified, its not safe to split it
        canSplit = origData is None
        if canSplit:
            origData = data
        assert origData == self.rawdata[htmlFrom:htmlTo], "parser out of sync"

        # ignore text in non-text tags
        if not self.isBodyTag or self.isNonTextTag > 0:
            return

        # remove some "garbage" from data, like double spaces or invisible spaces
        cData = self.cleanData(data)

        if self.doMapToPlainText:
            # if the next location is within this data chunk, but not at the beginning,
            # split off the first part (or else whitespace normalization would affect location)
            # if data is already transformed, we must not split; location is at start of chunk
            while self.nextLocation is not None and self.nextLocation < htmlTo:
                if canSplit and self.nextLocation > htmlFrom:
                    splitIdx = utf8.adjustL(data, self.nextLocation - htmlFrom)
                    if splitIdx > 0:
                        data1 = data[:splitIdx]
                        self.processData(data1, data1, htmlFrom, htmlFrom+splitIdx)
                        #self.processData(data[splitIdx:], None, htmlFrom+splitIdx, htmlTo)
                        #return;
                        data = data[splitIdx:]
                        cData = self.cleanData(data)
                        htmlFrom += splitIdx
                        continue
                # start of data corresponds to nextLocation
                plainTextLoc = self.plainTextLength
                if cData.startswith(' '):
                    plainTextLoc += 1 # consistently map to non-space location (assuming there will be one)
                self.nextLocationFound( plainTextLoc )
        elif cData:
            # if next location is within cData, figure out where exactly; repeat while more locations within
            plainFrom = self.plainTextCursorLoc()
            plainTo = self.plainTextCursorLoc(cData)
            while self.nextLocation is not None and self.nextLocation < plainTo:
                if not canSplit or self.nextLocation < plainFrom:
                    assert not canSplit or self.nextLocation == plainFrom # else nextLocation should have been found before
                    self.nextLocationFound(htmlFrom)
                    continue

                # figure out how many utf-8 bytes of data are needed to reach next location;
                # use nested intervals to find out fast for large data
                # (also avoids to recode self.cleanData() for iterative char-by-char use)
                # note: if you really believe utf-8 to be a good text processing format,
                # then try to prove that a) the nested intervals iteration always terminates and
                # b) when it terminates htmlFrom+iMin is always the first location in html that generates
                # self.nextLocation plain characters, not counting whitespace at the end.
                # Now, do you know what I mean? Sometimes a little more complexity is too much.
                iMin = 0
                iMax = len(data)
                while True:
                    assert iMin < iMax # invalid interval
                    # loop invariant: loc(iMin) <= nextLocation < loc(iMax)
                    # where: loc(i) = plainTextCursorLoc(cleanData(data[:i]))
                    # additional loop invariant: iMin and iMax never point to an utf-8 continuation byte in data
                    mid = int((iMin+iMax+1)/2) # round middle upwards
                    assert iMin < mid
                    iMid = utf8.adjustR( data, mid )
                    if iMid >= iMax:
                        assert iMid == iMax
                        if iMid == mid:
                            break # just one single-byte char in interval => done
                        # try to strip last multi-byte char, or else it may eclipse shorter char before it
                        iMid = utf8.adjustL( data, mid )
                        if iMid == iMin:
                            break # just one multi-byte char in interval => done
                    assert iMin < iMid
                    loc = self.plainTextCursorLoc( self.cleanData(data[:iMid]) )
                    if loc <= self.nextLocation:
                        iMin = iMid
                    else:
                        iMax = iMid

                # store location
                self.nextLocationFound(htmlFrom + iMin)

        # append to text list
        #print "Append: '%s'" % data
        assert not self.isClosed
        self.lastHtmlTo = htmlTo
        if cData != '':
            self.plainTextList.append(cData)
            self.plainTextLength += len(cData)
            
    def handle_entityref(self, name):
        #print "entity '%s'" % name
        # get original data from rawdata, because it may not end with a ;
        end = self.locCur+2+len(name)
        origData = self.rawdata[self.locCur:end]
        assert name == origData[1:-1], "parser out of sync"
        try:
            data = html.utf8Entities()[name]
        except KeyError:
            data = origData
        self.processData(data, origData, self.locCur, end)
        
    def handle_charref(self, number):
        #print "char '%s'" % number
        # get original data from rawdata, because it may not end with a ;
        end = self.locCur+3+len(number)
        origData = self.rawdata[self.locCur:end]
        assert number == origData[2:-1], "parser out of sync"
        if number[0] in ('x','X'):
            c = int(number[1:], 16)
        else:
            c = int(number)
        data = utf8.chr(c)
        self.processData(data, origData, self.locCur, self.locCur+3+len(number))
        
    def getPlainText(self):
        assert self.isClosed
        if len(self.plainTextList) > 1:
            self.plainTextList = [ ''.join(self.plainTextList) ]
        text = self.plainTextList[0]
        # plain text always ends with space
        return text if text.endswith(' ') else text+' '

# some tests for developer analysis
if __debug__ and __name__ == '__main__':

    doPrint = True

    def printStartTest():
        import inspect
        stack = inspect.stack()
        print "TEST", " ".join(map(lambda i: stack[i][3], range(len(stack)-2, 0, -1)))
        
    def printEndTest():
        print "OK"

    # convert a full html to plain location map into a more compact and readable form;
    # ... (10, 4), (15, 5) ... means html location 10-14 map to plain text location 4, html 15ff maps to 5
    # ... (20, 10, 5) ... means html 20-24 maps to plain 10-14
    def fullLocMapToMatchList(fullLocMap, keysLen, valuesLen):

        # keys are contiguous and increasing        
        keys = sorted( fullLocMap.keys() )
        assert keys == range(keysLen), "not a full map"
        
        # values are monotonous up
        values = fullLocMap.values()
        for i in range(1, len(values)):
            assert values[i-1] <= values[i]
            
        matchList = []
        i = 0
        while i < keysLen:
            j = i
            while j+1 < keysLen and values[j]+1 == values[j+1]:
                j += 1
            if i != j: # if consecutive matches
                # emit a matching range
                matchList.append( (i, values[i], j-i+1) )
            elif i == 0 or values[i] != values[i-1]: # else if new value at this position
                # emit a value change
                matchList.append( (i, values[i]) )
            i = j+1
            
        # self test
        for i in keys:
            loc = mapUsingMatchList(i, matchList, keysLen, valuesLen)
            assert loc == fullLocMap[i], "@%d %d != %d" % (i, loc, fullLocMap[i])
            
        return matchList

    # map a location using the matchList returned from fullLocMapToMatchList()
    def mapUsingMatchList(loc, matchList, keysLen, valuesLen):
        mapLoc = None
        for i in range(len(matchList)):
            l = 1
            if len(matchList[i]) == 3:
                off = loc - matchList[i][0]
                l = matchList[i][2]
                if 0 <= off < l:
                    mapLoc = matchList[i][1] + off
                    break
            nextChange = matchList[i+1][0] if i+1 < len(matchList) else keysLen
            if matchList[i][0] <= loc < nextChange:
                mapLoc = matchList[i][1] + l-1
                break
        return mapLoc

    def printMatchList(matchList, keysText, valuesText, listName):
        #print "MatchList =", matchList
        #print ("MatchList =%s" % matchList).replace("(", "\n    (")
        print "%s = [" % listName
        cOrd = (0, 1, '->') if len(keysText) < len(valuesText) else (1, 0, '<-')
        for i in range(len(matchList)):
            com = []
            # what is matched in key text?
            p = matchList[i][0]
            q = matchList[i+1][0] if i+1 < len(matchList) else len(keysText)
            com += [" '%s'" % utf8.forPrint(keysText, p, q).encode('string_escape')]
            # what is matched in values text?
            p = matchList[i][1]
            q = p + (matchList[i][2] if len(matchList[i]) == 3 else 1)
            com += [' "%s"' % utf8.forPrint(valuesText, p, q).encode('string_escape')]
            print "        %s%s #%s %s%s" % (matchList[i], "," if i+1 < len(matchList) else "]", com[cOrd[0]], cOrd[2], com[cOrd[1]])

    def htmlCharForPrint(html, loc):
        match = re.match(r"&[^;]+;", html[loc:])
        if match:
            return match.group(0)
        return utf8.charForPrint(html, loc)
            
    def printMap(locMap, html, plainText, doMapToPlainText=True):
        for loc in sorted(locMap.keys()):
            if doMapToPlainText:
                lHtml = loc
                lPlain = locMap[loc]
                lDir = "<-"
            else:
                lHtml = locMap[loc]
                lPlain = loc
                lDir = "->"
            hc = htmlCharForPrint(html, lHtml)
            pc = utf8.charForPrint(plainText, lPlain)
            print "[%d, %d] \"%s\" %s '%s'" % (loc, locMap[loc], pc, lDir, hc)
            
    # internal test if given match lists are inverse on a not bijective mapping,
    # i.e. there is a bijective core, but also a lot of "junk" keys which
    # will somehow map into the core (but if that 'somehow' is reasonable can't be checked here)
    def devTestMatchListSanityCheck(html, plainText, expectedMatchList, expectedInverseMatchList):
        printStartTest()
        
        # expand expectedMatchList as forward
        forward = {} # html -> plainText
        for locHtml in range(len(html)):
            locPlain = mapUsingMatchList(locHtml, expectedMatchList, len(html), len(plainText))
            forward[locHtml] = locPlain
        #print "forward =", forward
        
        
        # expand expectedInverseMatchList as backward
        backward = {} # plainText -> html
        assert plainText[0] == plainText[-1] == ' '
        for locPlain in range(len(plainText)):
            locHtml = mapUsingMatchList(locPlain, expectedInverseMatchList, len(plainText), len(html))
            backward[locPlain] = locHtml
        #print "backward =", backward
        
        # values of maps do not reference utf-8 continuation bytes
        for locPlain in forward.values():
            assert not utf8.isCont(plainText, locPlain)
        for locHtml in backward.values():
            assert not utf8.isCont(html, locHtml)
        
        # identify the bijective core: everything in plainText that is not a space must be in core
        plainCore = set()
        for locPlain in range(len(plainText)):
            if plainText[locPlain] != ' ' and not utf8.isCont(plainText, locPlain):
                plainCore.add(locPlain)
        #plainCore = sorted(plainCore)

        # print core mapping plain -> html -> plain
        if doPrint:
            for locPlain in plainCore:
                locHtml = backward[locPlain]
                locPlain2 = forward[locHtml]
                print locPlain, locHtml, locPlain2, \
                    "'%s'" % utf8.charForPrint(plainText, locPlain), \
                    "'%s'" % htmlCharForPrint(html, locHtml), \
                    "'%s'" % utf8.charForPrint(plainText, locPlain2), \
                    "<======== ???" if locPlain != locPlain2 else ""
        # check without print
        for locPlain in plainCore:
            assert locPlain == forward[backward[locPlain]], "@%d: != %d" % (locPlain, backward[forward[locPlain]])

        htmlCoreLocs = sorted(set([htmlLoc for plainLoc, htmlLoc in backward.items() if plainLoc < len(plainText)-1]))
        if doPrint:
            print "htmlCoreLocs =", map(lambda l: htmlCharForPrint(html, l), htmlCoreLocs)
        
        plainCoreLocs = sorted(set([plainLoc for plainLoc in forward.values() if plainLoc < len(plainText)-1]))
        if doPrint:
            print "plainCoreLocs =", map(lambda l: utf8.charForPrint(plainText, l), plainCoreLocs)
        
        # if the core is bijective, it must have same size in html and plainText
        assert len(htmlCoreLocs) == len(plainCoreLocs), ("core not bijective", len(htmlCoreLocs), len(plainCoreLocs))
        
        printEndTest()
        
    # TEST: create plain text (no mapping)
    def devTestHtmlToPlain(html, expectedPlainText):
        printStartTest()
        parser = ParseHtmlToPlainText()
        parser.feed(html)
        plainText = parser.getPlainText()

        try:
            assert plainText.startswith(' ')
            assert plainText.endswith(' ')
            assert plainText == expectedPlainText
        except:
            print "PlainText: \"%s\"" % utf8.forPrint(plainText)
            print "Expected:  \"%s\"" % utf8.forPrint(expectedPlainText)
            import difflib
            matcher = difflib.SequenceMatcher();
            matcher.set_seq1(plainText)
            matcher.set_seq2(expectedPlainText)
            print matcher.get_matching_blocks()
            raise
        printEndTest()

    # TEST: create plain text and do full mapping html->plain
    def devTestMapHtmlToPlainFull(html, expectedPlainText, expectedMatchList):
        printStartTest()
        locMap = {}
        for loc in range(len(html)):
            locMap[loc] = 0
        
        parser = ParseHtmlToPlainText(locMap)
        parser.feed(html)
        
        plainText = parser.getPlainText()
        if doPrint:
            print "PlainText: \"%s\"" % utf8.forPrint(plainText)
            printMap(locMap, html, plainText)
        #print locMap.values()
        assert plainText == expectedPlainText

        # compress the map for easier analysis and storage in test
        matchList = fullLocMapToMatchList(locMap, len(html), len(plainText))
        if doPrint:
            print "If OK the following may become a new value for expectedMatchList:"
            printMatchList(matchList, html, plainText, "MatchList")
        assert matchList == expectedMatchList

        printEndTest()

    # TEST:
    def devTestMapHtmlToPlainLocationFarBeyondEnd(html, expectedPlainText):
        printStartTest()
        
        locBeyondEnd = 1
        while locBeyondEnd < max(len(html), len(expectedPlainText)):
            locBeyondEnd *= 10
        locBeyondEnd = locBeyondEnd*10 - 1
        locMap = { locBeyondEnd: "does not matter what is here" }
        
        parser = ParseHtmlToPlainText(locMap)
        parser.feed(html)

        # location beyond end points to end of plain text
        assert len(locMap) == 1
        assert isinstance(locMap[locBeyondEnd], int)
        assert locMap[locBeyondEnd] == len(expectedPlainText)-1

        printEndTest()

     # TEST:
    def devTestMapPlainToHtmlLocationFarBeyondEnd(html, expectedPlainText):
        printStartTest()
        
        locBeyondEnd = 1
        while locBeyondEnd < max(len(html), len(expectedPlainText)):
            locBeyondEnd *= 10
        locBeyondEnd = locBeyondEnd*10 - 1
        locMap = { locBeyondEnd: "does not matter what is here" }
        
        parser = ParseHtmlToPlainText(locMap, doMapToPlainText=False)
        parser.feed(html)

        # location beyond end points to end of last processed data
        assert len(locMap) == 1
        assert isinstance(locMap[locBeyondEnd], int)
        assert not utf8.isCont(html, locMap[locBeyondEnd])
        assert html[utf8.adjustR(html, locMap[locBeyondEnd]+1)] == '<'

        printEndTest()

   # TEST: create plain text and do sparse mapping html->plain
    def devTestMapHtmlToPlainSparse(html, expectedPlainText, expectedMatchList):
        printStartTest()
        for step in range(1,30):
            if doPrint:
                print "step", step
            for start in range(step):
                if step < 2:
                    locMap = { 0: -1, 1: -1, 25: -1, 40: -1, 30: -1 }
                else:
                    locMap = {}
                    for loc in range(start, len(html), step):
                        locMap[loc] = -10 - loc

                parser = ParseHtmlToPlainText(locMap)
                parser.feed(html)
                
                plainText = parser.getPlainText()
                #print "PlainText: \"%s\"" % plainText
                assert plainText == expectedPlainText
            
                for key in locMap:
                    expectedValue = mapUsingMatchList(key, expectedMatchList, len(html), len(plainText))
                    assert locMap[key] == expectedValue, "@%d: %d != %d" % (key, locMap[key], expectedValue)
        printEndTest()

    # TEST: create plain text and do full inverse mapping, i.e. plain->html
    def devTestMapPlainToHtmlFull(html, expectedPlainText, expectedInverseMatchList):
        printStartTest()
        locMap = {}
        for loc in range(len(expectedPlainText)):
            locMap[loc] = -10 - loc

        parser = ParseHtmlToPlainText(locMap, doMapToPlainText=False)
        parser.feed(html)
        #plainText = parser.getPlainText()
        
        if doPrint:
            printMap(locMap, html, expectedPlainText, doMapToPlainText=False)
        
        # compress the map for easier analysis and storage in test
        matchList = fullLocMapToMatchList(locMap, len(expectedPlainText), len(html))
        if doPrint:
            print "If OK the following may become a new value for expectedInverseMatchList:"
            printMatchList(matchList, expectedPlainText, html, "InverseMatchList")
        assert matchList == expectedInverseMatchList
        printEndTest()
        
    # TEST: create plain text and do sparse inverse mapping plain->html
    def devTestMapPlainToHtmlSparse(html, expectedPlainText, expectedInverseMatchList):
        printStartTest()
        for step in range(1,30):
            if doPrint:
                print "step", step
            for start in range(step):
                if step < 2:
                    locMap = { 0: -1, 1: -1, 25: -1, 40: -1, 30: -1 }
                else:
                    locMap = {}
                    for loc in range(start, len(expectedPlainText), step):
                        locMap[loc] = -10 - loc

                parser = ParseHtmlToPlainText(locMap, doMapToPlainText=False)
                parser.feed(html)
                
                plainText = parser.getPlainText()
                #print "PlainText: \"%s\"" % plainText
                assert plainText == expectedPlainText
            
                for key in locMap:
                    expectedValue = mapUsingMatchList(key, expectedInverseMatchList, len(plainText), len(html))
                    assert locMap[key] == expectedValue, "@%d: %d != %d" % (key, locMap[key], expectedValue)
        printEndTest()
 
def testSet1():
    # focus is to test html problems; also covers basic utf-8 support
    html = """<head>ignore</head>
    <body> \xef\xbb\xbf<p>this &uuml  &lt;is&gt; an ü-<I>Um&shy;laut </i> for <a ref="xyz" />test</P>
    <p class="some">&mdash; &copy; &#169; <!-- comment --> €. <object><style>goo</style></object><script>bogus</script>
    </body>
    garbage
    <html><head /><BODY>New<br>HTML<p>fi\xc2\xad<!--=shy-->le,</p>truncateß<head>next file"""
    expectedPlainText = " this ü <is> an ü-Umlaut for test — © © €. New HTML file, truncateß "
    expectedMatchList = [
        (0, 1), # "t" <- '<head>ignore</head>\\n    <body> \\ufeff<p>'
        (37, 1, 4), # "this" <- 'this'
        (41, 6), # "\\xfc" <- ' &uuml;'
        (48, 9), # "<" <- ' &lt'
        (52, 9, 4), # "<is>" <- ';is&gt;'
        (59, 14), # "a" <- ' '
        (60, 14, 2), # "an" <- 'an'
        (62, 17), # "\\xfc" <- ' \\xfc'
        (65, 19, 2), # "-U" <- '-<I>'
        (69, 20, 3), # "Uml" <- 'Um&shy;'
        (76, 22, 4), # "laut" <- 'laut'
        (80, 27), # "f" <- ' </i> '
        (86, 27, 3), # "for" <- 'for'
        (89, 31), # "t" <- ' <a ref="xyz" />'
        (105, 31, 4), # "test" <- 'test'
        (109, 36), # "\\u2014" <- '</P>\\n    <p class="some">&mdash;'
        (141, 40), # "\\xa9" <- ' &copy;'
        (148, 43), # "\\xa9" <- ' &#169;'
        (155, 46), # "\\u20ac" <- ' <!-- comment --> \\u20ac'
        (176, 49), # "." <- '.'
        (177, 51), # "N" <- ' <object><style>goo</style></object><script>bogus</script>\\n    </body>\\n    garbage\\n    <html><head /><BODY>'
        (284, 51, 3), # "New" <- 'New'
        (287, 55), # "H" <- '<br>'
        (291, 55, 4), # "HTML" <- 'HTML'
        (295, 60), # "f" <- '<p>'
        (298, 60, 3), # "fil" <- 'fi\\xad<!--=shy-->'
        (313, 62, 3), # "le," <- 'le,'
        (316, 66), # "t" <- '</p>'
        (320, 66, 9), # "truncate\\xdf" <- 'truncate\\xdf'
        (330, 76)] # " " <- '<head>next file'
    expectedInverseMatchList = [
        (0, 37), # ' ' -> "t"
        (1, 37, 4), # 'this' -> "this"
        (5, 42), # ' \\xfc' -> "&"
        (8, 49), # ' <' -> "&"
        (10, 53, 3), # 'is>' -> "is&"
        (13, 60), # ' ' -> "a"
        (14, 60, 2), # 'an' -> "an"
        (16, 63), # ' \\xfc' -> "\\xfc"
        (19, 65), # '-' -> "-"
        (20, 69, 2), # 'Um' -> "Um"
        (22, 76, 4), # 'laut' -> "laut"
        (26, 86), # ' ' -> "f"
        (27, 86, 3), # 'for' -> "for"
        (30, 105), # ' ' -> "t"
        (31, 105, 4), # 'test' -> "test"
        (35, 134), # ' \\u2014' -> "&"
        (39, 142), # ' \\xa9' -> "&"
        (42, 149), # ' \\xa9' -> "&"
        (45, 173), # ' \\u20ac' -> "\\u20ac"
        (49, 176), # '.' -> "."
        (50, 284), # ' ' -> "N"
        (51, 284, 3), # 'New' -> "New"
        (54, 291), # ' ' -> "H"
        (55, 291, 4), # 'HTML' -> "HTML"
        (59, 298), # ' ' -> "f"
        (60, 298, 2), # 'fi' -> "fi"
        (62, 313, 3), # 'le,' -> "le,"
        (65, 320), # ' ' -> "t"
        (66, 320, 9)] # 'truncated' -> "truncated"
    devTestMatchListSanityCheck(html, expectedPlainText, expectedMatchList, expectedInverseMatchList)

    devTestHtmlToPlain(html, expectedPlainText)
    devTestMapHtmlToPlainLocationFarBeyondEnd(html, expectedPlainText)
    devTestMapHtmlToPlainFull(html, expectedPlainText, expectedMatchList)
    devTestMapHtmlToPlainSparse(html, expectedPlainText, expectedMatchList)
    devTestMapPlainToHtmlFull(html, expectedPlainText, expectedInverseMatchList)
    devTestMapPlainToHtmlLocationFarBeyondEnd(html, expectedPlainText)
    devTestMapPlainToHtmlSparse(html, expectedPlainText, expectedInverseMatchList)
   
def testSet2():
    # focus is to test for utf-8 problems
    utf8Chars = [
        # C0 Controls and Basic Latin: U+0000 – U+007F (0–127)
        0, 0x7f, '!', '5', 'A', 'z',
        # C1 Controls and Latin–1 Supplement: U+0080 – U+00FF (128–255)
        0x80, 0xff, 165, 188, 209, 241,
        # utf-8 2 to 3 bytes
        '\xdf\xbf', '\xe0\xa0\x80',
        #Cyrillic: U+0400 – U+04FF (1024–1279)
        1033, 1065, 1097, 1219,
        # Enclosed Alphanumeric Supplement: U+1F100 – U+1F1FF (127232–127487)
        127232, 127254, 127277, 127376,
        # Hiragana: U+3040 – U+309F (12352–12447)
        12354, 12368, 12427, 12446,
        # Alphabetic Presentation Forms: U+FB00 – U+FB4F (64256–64335)
        64256, 64257, 64279, 64335,
        # Specials: U+FEFF, U+FFF0 – U+FFFF (65279, 65520–65535)
        0xfeff, 65529, 65530, 65532, 65533,
        # Tai Xuan Jing Symbols: U+1D300 – U+1D35F (119552–119647)
        119552, 119555, 119569, 119626,
        # Mathematical Alphanumeric Symbols: U+1D400 – U+1D7FF (119808–120831)
        120016, 120172, 120643, 120829,
        # last valid unicode code point
        1114111]

    # convert code points in utf8Chars to utf-8 byte strings
    for i in range(len(utf8Chars)):
        if isinstance(utf8Chars[i], int):
            utf8Chars[i] = utf8.chr(utf8Chars[i])

    if doPrint:
        print
        print "utf8Chars:"
        for c in utf8Chars:
            v = utf8.ord(c)
            print v, "0x%x" % v, "'%s'" % c.encode('string_escape'), '"%s"' % utf8.forPrint(c)
        print

    assert "<" not in utf8Chars
    assert ">" not in utf8Chars
    assert "&" not in utf8Chars
    
    # assemble a pseudo-random string
    import random
    blocks = utf8Chars
    for l in range(10):
        blocks.append( ' '*int(l/2+1) )
    for _ in range(8):
        blocks.append( "<a />" )
    for _ in range(5):
        blocks.append( "&amp;" )
    random.seed(4711)
    html = "<body>%s</body>" % ''.join(random.choice(blocks) for i in range(100))

    # to prevent being dependent on implementation of pseudo random generator,
    # here is what I got, and what the following expected match lists refer to:
    html = html = "<body>\xc2\xbc\xef\xbf\xba\xf0\x9d\x8c\x83<a /><a />     \xf0\x9f\x84\xad\xef\xac\x97  \xe3\x81\x82     \xf0\x9f\x84\xad  !<a />\xef\xbf\xbc\xdf\xbf\xf0\x9d\x9d\x83\xd0\xa9<a />&amp;     \xc2\xa5\xf4\x8f\xbf\xbf&amp;\xd1\x89\xc3\x91\xc2\xa5&amp;<a />\xf0\x9d\x8c\x83&amp;\xc2\xa5\xd0\xa9   \xd0\x89\xf0\x9f\x86\x90\xe3\x82\x8b&amp;\xf0\x9f\x86\x90\xd3\x83\xe3\x82\x8b\xf0\x9d\x8c\x80\xf0\x9d\x9d\x83    \xf0\x9d\x8c\x91\xf0\x9f\x84\xad   \xf0\x9d\x95\xac\xf0\x9d\x8d\x8a\xef\xac\x81\xf0\x9d\x9d\x83\xe3\x82\x9e&amp;<a />\xf0\x9d\x8c\x83&amp;<a />\xf0\x9d\x9d\x83\xf0\x9d\x9d\x83<a />\x00\xd3\x835   \xc2\x80<a />\xd0\x89\xef\xbf\xba\xf0\x9d\x93\x90A&amp;    !&amp;A  \xe3\x81\x82   \xef\xbb\xbf\xd0\x89\xc3\xb1 \xd0\x89  \xd1\x89\xef\xac\x97 \xf0\x9d\x9d\x83\xc2\x80\xef\xac\x97<a />\xef\xbf\xbd \xf0\x9f\x84\x96  z    </body>"
    
    if doPrint:
        print
        print 'html = "%s"' % html.encode('string_escape')
        print

    expectedPlainText = html
    expectedPlainText = re.sub( r"<[^>]*>", "", expectedPlainText )
    expectedPlainText = re.sub( r"&amp;", "&", expectedPlainText )
    expectedPlainText = re.sub( utf8.RE_ZERO_WIDTH_SPACE, "", expectedPlainText )
    expectedPlainText = re.sub( utf8.RE_WHITESPACES, " ", expectedPlainText )
    expectedPlainText = re.sub( r"^ *", " ", expectedPlainText )
    expectedPlainText = re.sub( r" *$", " ", expectedPlainText )
    if doPrint:
        print
        print 'expectedPlainText = "%s"' % utf8.forPrint(expectedPlainText)
        print
    
    " this ü <is> an ü-Umlaut for test — © © €. New HTML file, truncated "
    expectedMatchList = [
        (0, 1), # "\\xbc" <- '<body>\\xbc'
        (8, 3), # "\\ufffa" <- '\\ufffa'
        (11, 6), # "\\U0001d303" <- '\\U0001d303'
        (15, 11), # "\\U0001f12d" <- '<a /><a />     \\U0001f12d'
        (34, 15), # "\\ufb17" <- '\\ufb17'
        (37, 19), # "\\u3042" <- '  \\u3042'
        (42, 23), # "\\U0001f12d" <- '     \\U0001f12d'
        (51, 28), # "!" <- '  '
        (53, 28, 2), # "!\\ufffc" <- '!<a />\\ufffc'
        (62, 32), # "\\u07ff" <- '\\u07ff'
        (64, 34), # "\\U0001d743" <- '\\U0001d743'
        (68, 38), # "\\u0429" <- '\\u0429'
        (70, 40), # "&" <- '<a />&amp;'
        (80, 42), # "\\xa5" <- '     \\xa5'
        (87, 44), # "\\U0010ffff" <- '\\U0010ffff'
        (91, 48), # "&" <- '&amp'
        (95, 48, 2), # "&\\u0449" <- ';\\u0449'
        (98, 51), # "\\xd1" <- '\\xd1'
        (100, 53), # "\\xa5" <- '\\xa5'
        (102, 55), # "&" <- '&amp'
        (106, 55, 2), # "&\\U0001d303" <- ';<a />\\U0001d303'
        (116, 60), # "&" <- '&amp'
        (120, 60, 2), # "&\\xa5" <- ';\\xa5'
        (123, 63), # "\\u0429" <- '\\u0429'
        (125, 66), # "\\u0409" <- '   \\u0409'
        (130, 68), # "\\U0001f190" <- '\\U0001f190'
        (134, 72), # "\\u308b" <- '\\u308b'
        (137, 75), # "&" <- '&amp'
        (141, 75, 2), # "&\\U0001f190" <- ';\\U0001f190'
        (146, 80), # "\\u04c3" <- '\\u04c3'
        (148, 82), # "\\u308b" <- '\\u308b'
        (151, 85), # "\\U0001d300" <- '\\U0001d300'
        (155, 89), # "\\U0001d743" <- '\\U0001d743'
        (159, 94), # "\\U0001d311" <- '    \\U0001d311'
        (167, 98), # "\\U0001f12d" <- '\\U0001f12d'
        (171, 103), # "\\U0001d56c" <- '   \\U0001d56c'
        (178, 107), # "\\U0001d34a" <- '\\U0001d34a'
        (182, 111), # "\\ufb01" <- '\\ufb01'
        (185, 114), # "\\U0001d743" <- '\\U0001d743'
        (189, 118), # "\\u309e" <- '\\u309e'
        (192, 121), # "&" <- '&amp'
        (196, 121, 2), # "&\\U0001d303" <- ';<a />\\U0001d303'
        (206, 126), # "&" <- '&amp'
        (210, 126, 2), # "&\\U0001d743" <- ';<a />\\U0001d743'
        (220, 131), # "\\U0001d743" <- '\\U0001d743'
        (224, 135), # "\\x00" <- '<a />'
        (229, 135, 2), # "\\x00\\u04c3" <- '\\x00\\u04c3'
        (232, 138), # "5" <- '5'
        (233, 140), # "\\x80" <- '   \\x80'
        (238, 142), # "\\u0409" <- '<a />\\u0409'
        (245, 144), # "\\ufffa" <- '\\ufffa'
        (248, 147), # "\\U0001d4d0" <- '\\U0001d4d0'
        (252, 151, 2), # "A&" <- 'A&amp;'
        (258, 154), # "!" <- '    '
        (262, 154, 2), # "!&" <- '!&amp'
        (267, 155, 2), # "&A" <- ';A'
        (269, 158), # "\\u3042" <- '  \\u3042'
        (274, 162), # "\\u0409" <- '   \\ufeff\\u0409'
        (282, 164), # "\\xf1" <- '\\xf1'
        (284, 167), # "\\u0409" <- ' \\u0409'
        (287, 170), # "\\u0449" <- '  \\u0449'
        (291, 172), # "\\ufb17" <- '\\ufb17'
        (294, 176), # "\\U0001d743" <- ' \\U0001d743'
        (299, 180), # "\\x80" <- '\\x80'
        (301, 182), # "\\ufb17" <- '\\ufb17'
        (304, 185), # "\\ufffd" <- '<a />\\ufffd'
        (312, 189), # "\\U0001f116" <- ' \\U0001f116'
        (317, 194), # "z" <- '  '
        (319, 194, 2)] # "z " <- 'z    </body>'
    expectedInverseMatchList = [
        (0, 6), # ' \\xbc' -> "\\xbc"
        (3, 8), # '\\ufffa' -> "\\ufffa"
        (6, 11), # '\\U0001d303' -> "\\U0001d303"
        (10, 30), # ' \\U0001f12d' -> "\\U0001f12d"
        (15, 34), # '\\ufb17' -> "\\ufb17"
        (18, 39), # ' \\u3042' -> "\\u3042"
        (22, 47), # ' \\U0001f12d' -> "\\U0001f12d"
        (27, 53), # ' !' -> "!"
        (29, 59), # '\\ufffc' -> "\\ufffc"
        (32, 62), # '\\u07ff' -> "\\u07ff"
        (34, 64), # '\\U0001d743' -> "\\U0001d743"
        (38, 68), # '\\u0429' -> "\\u0429"
        (40, 75), # '&' -> "&"
        (41, 85), # ' \\xa5' -> "\\xa5"
        (44, 87), # '\\U0010ffff' -> "\\U0010ffff"
        (48, 91), # '&' -> "&"
        (49, 96), # '\\u0449' -> "\\u0449"
        (51, 98), # '\\xd1' -> "\\xd1"
        (53, 100), # '\\xa5' -> "\\xa5"
        (55, 102), # '&' -> "&"
        (56, 112), # '\\U0001d303' -> "\\U0001d303"
        (60, 116), # '&' -> "&"
        (61, 121), # '\\xa5' -> "\\xa5"
        (63, 123), # '\\u0429' -> "\\u0429"
        (65, 128), # ' \\u0409' -> "\\u0409"
        (68, 130), # '\\U0001f190' -> "\\U0001f190"
        (72, 134), # '\\u308b' -> "\\u308b"
        (75, 137), # '&' -> "&"
        (76, 142), # '\\U0001f190' -> "\\U0001f190"
        (80, 146), # '\\u04c3' -> "\\u04c3"
        (82, 148), # '\\u308b' -> "\\u308b"
        (85, 151), # '\\U0001d300' -> "\\U0001d300"
        (89, 155), # '\\U0001d743' -> "\\U0001d743"
        (93, 163), # ' \\U0001d311' -> "\\U0001d311"
        (98, 167), # '\\U0001f12d' -> "\\U0001f12d"
        (102, 174), # ' \\U0001d56c' -> "\\U0001d56c"
        (107, 178), # '\\U0001d34a' -> "\\U0001d34a"
        (111, 182), # '\\ufb01' -> "\\ufb01"
        (114, 185), # '\\U0001d743' -> "\\U0001d743"
        (118, 189), # '\\u309e' -> "\\u309e"
        (121, 192), # '&' -> "&"
        (122, 202), # '\\U0001d303' -> "\\U0001d303"
        (126, 206), # '&' -> "&"
        (127, 216), # '\\U0001d743' -> "\\U0001d743"
        (131, 220), # '\\U0001d743' -> "\\U0001d743"
        (135, 229, 2), # '\\x00\\u04c3' -> "\\x00\\u04c3"
        (138, 232), # '5' -> "5"
        (139, 236), # ' \\x80' -> "\\x80"
        (142, 243), # '\\u0409' -> "\\u0409"
        (144, 245), # '\\ufffa' -> "\\ufffa"
        (147, 248), # '\\U0001d4d0' -> "\\U0001d4d0"
        (151, 252, 2), # 'A&' -> "A&"
        (153, 262), # ' ' -> "!"
        (154, 262, 2), # '!&' -> "!&"
        (156, 268), # 'A' -> "A"
        (157, 271), # ' \\u3042' -> "\\u3042"
        (161, 280), # ' \\u0409' -> "\\u0409"
        (164, 282), # '\\xf1' -> "\\xf1"
        (166, 285), # ' \\u0409' -> "\\u0409"
        (169, 289), # ' \\u0449' -> "\\u0449"
        (172, 291), # '\\ufb17' -> "\\ufb17"
        (175, 295), # ' \\U0001d743' -> "\\U0001d743"
        (180, 299), # '\\x80' -> "\\x80"
        (182, 301), # '\\ufb17' -> "\\ufb17"
        (185, 309), # '\\ufffd' -> "\\ufffd"
        (188, 313), # ' \\U0001f116' -> "\\U0001f116"
        (193, 319), # ' z' -> "z"
        (195, 323)] # ' ' -> " "
    devTestMatchListSanityCheck(html, expectedPlainText, expectedMatchList, expectedInverseMatchList)

    devTestHtmlToPlain(html, expectedPlainText)
    devTestMapHtmlToPlainLocationFarBeyondEnd(html, expectedPlainText)
    devTestMapHtmlToPlainFull(html, expectedPlainText, expectedMatchList)
    devTestMapHtmlToPlainSparse(html, expectedPlainText, expectedMatchList)
    devTestMapPlainToHtmlFull(html, expectedPlainText, expectedInverseMatchList)
    devTestMapPlainToHtmlLocationFarBeyondEnd(html, expectedPlainText)
    devTestMapPlainToHtmlSparse(html, expectedPlainText, expectedInverseMatchList)
    
if __debug__ and __name__ == '__main__':
    # run tests
    testSet1()
    testSet2()
