# -*- coding: utf-8 -*-
assert '…' == '\xe2\x80\xa6', "file encoding error"

# Copyright 2013 Axel Walthelm
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import re
import xml.etree.ElementTree as ElementTree
import Utf8Tools as utf8

# Kindle tends to generate a 'highlight' and a 'note' for
# every note set. It may happen that the highlight is deleted later,
# while the note remains. But usually a note has a corresponding highlight.
# To simplifiy things we join them both into a single KindleAnnotation note,
# which has note=None if it is a pure highlight
# and has highlight=None if it is a pure note.

class KindleAnnotation:
    # ordernr: consecutive number
    # id: unique identifier intended to be used in html
    # kind: one of ['highlight', 'note', 'bookmark']
    # createtime: creation time in format like "2013-04-25T23:58:00" (sortable)
    # modifytime: modification time in same format
    # begin: start location as index into rawml (full resolution, not divided by 150)
    # end: end location as index into rawml (full resolution, not divided by 150)
    #      end is included in range; if range is empty end is None
    # text: text that is in book from begin to end
    # context: tuple of some text before and after above text
    # note: note text if it is a note; else None
    # highlight: (only if kind=='note') do highlight text?
    #
    # --- later also:
    # epubTextBegin: matched start in epubText
    # epubTextEnd: matched end in epubText
    # ka.epubTextUncertain: if not None, an uncertainty range for the match
    pass

# get user annotations log and extract all annotations for the book
# todo/tbd: handle userannotlog.0 userannotlog.1 etc.?
def FromFileName(userAnnotLogFilePath, bookAsin):
    with file( userAnnotLogFilePath, 'rb' ) as f: # file is UTF-8 without BOM => read binary!
        # in python 2.x ElementTree.fromstring expects utf-8 input,
        # but but note that access methods later may return unicode strings
        anno = ElementTree.fromstring("<anno>%s</anno>" % f.read())
    annodict = {}
    nr = 0
    for book in anno:
        if book.tag != 'book':
            print "WARNING: non-book annotation entry", book.tag
            continue
        if book.get('key') != bookAsin:
            #print "INFO: entry for another book"
            continue
        if book.get('format') != 'mobi7':
            print "ERROR: unsupported book format", book.get('format')
            continue
        #print "ENTRY: ", elt
        for entry in book:
            kind = utf8.str(entry.tag)
            if kind not in ('highlight', 'note', 'bookmark'):
                if kind not in ('last_read'):
                    print "WARNING: unkown annotation entry", kind
                continue
            
            action = utf8.str(entry.get('action'))
            if action not in ('create', 'modify', 'delete'):
                print "WARNING: unknown annotation action", action
                continue
            
            begin = int(entry.get('begin'))
            end = int(entry.get('end'))
            key = "%s %d-%d" % (kind, begin, end)
            
            if end < begin:
                print "WARNING: end before start in annotation", key
                begin, end = end, begin
            
            if action == 'delete':
                if annodict.has_key(key):
                    del annodict[key]
                else:
                    print "WARNING: delete without create for annotation", key
                continue
            
            # action is 'create' or 'modify'
            if annodict.has_key(key):
                if action != 'modify':
                    print "WARNING: multiple create for annotation", key
                ka = annodict[key]
            else:
                if action != 'create':
                    print "WARNING: modify without create for annotation", key
                ka = KindleAnnotation()
                ka.createtime = utf8.str(entry.get('timestamp'))
                ka.ordernr = nr
                nr = nr+1
            ka.kind = kind
            ka.begin = begin
            ka.end = end
            ka.modifytime = utf8.str(entry.get('timestamp'))
            ka.note = utf8.str(entry.text) if kind == 'note' or entry.text else None
            annodict[key] = ka
            
    # merge matching pairs of 'highlight' and 'note'
    for note in annodict.keys():
        if note.startswith('note '):
            annodict[note].highlight = False
            high = 'highlight'+note[4:]
            if annodict.has_key(high):
                del annodict[high]
                annodict[note].highlight = True

    annolist = sorted(annodict.values(), key=lambda ka: ka.ordernr)
    #for ka in annolist:
    #    print "#%d %d-%d %s \"%s\"" % (ka.ordernr, ka.begin, ka.end, ka.kind, ka.note)
    
    # make identifier for annotations, ensuring they are unique (within this userannotlog)
    # basic idea is to use the time stamp, but Kindle seems to be a bit inaccurate with them,
    # so even it is almost impossible to do two annotation operations in a second, the same
    # time stamp can be used for two operations.
    usedIds = set()
    for ka in annolist:
        # html id must start match re "[A-Za-z][0-9A-Za-z:._-]*"
        # '.' and ':' may be difficult in CSS, but this id is for linking only, not for CSS
        baseId = "at_"+re.sub(r"[^0-9A-Za-z:._-]", "-", ka.modifytime)
        ka.id = baseId
        cnt = 0
        while ka.id in usedIds:
            cnt += 1
            ka.id = baseId+"."+str(cnt)
        usedIds.add(ka.id)
    
    return annolist

# extract information of ka.text and ka.context from given text and corresponding location mapping
def GetText(annolist, mapKindleLocToText, kindleText):
    def adjustWord(text, loc, inc):
        if loc < 0:
            loc = 0
        if loc >= len(text):
            loc = len(text)-1
        while loc+inc >= 0 and loc+inc < len(text) and text[loc] != ' ':
            loc += inc
        return loc - inc
    for ka in annolist:
        beg = mapKindleLocToText[ka.begin]
        end = utf8.adjustR(kindleText, mapKindleLocToText[ka.end]+1)
        ka.text = kindleText[beg:end]
        totalTagetTextLen = 120
        cLen = (totalTagetTextLen - len(ka.text)) / 2
        cBeg = adjustWord(kindleText, beg - cLen, -1)
        cEnd = adjustWord(kindleText, end + cLen, 1)
        ka.context = kindleText[cBeg:beg], kindleText[end:cEnd+1]
