# -*- coding: utf-8 -*-
assert '…' == '\xe2\x80\xa6', "file encoding error"

# Copyright 2013 Axel Walthelm
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

# difflib.SequenceMatcher is too slow to compare complete ebook texts as strings,
# even if the texts are identical it takes minutes.
# Since we are only interested in very similar texts, with maybe
# some extra text at the beginning and/or the end (e.g. table of contents)
# plus maybe some minor editing, a faster matching is possible.

from difflib import Match
from difflib import SequenceMatcher as _difflibMatcher
from collections import namedtuple as _namedtuple
import Utf8Tools as utf8
import SortedList

def _calculateRatio(matches, len_a, len_b):
    if not matches:
        return 0

    count = matches[0].size
    for i in xrange(1, len(matches)):
        m0 = matches[i-1]
        m1 = matches[i]
        overlap = max( m0.a+m0.size - m1.a, m0.b+m0.size - m1.b, 0 )
        assert overlap in (0, 1), overlap
        count += m1.size - overlap
    
    return 2.0 * count / (len_a + len_b)

_Split = _namedtuple('Split', 'a b isMatch')

# helper class to map an interval in a to the corresponding interval in b
# using initialized TextMatcher
class MatchMapper:
    def __init__(self, matcher):
        assert isinstance(matcher, TextMatcher)
        matches = matcher.get_matching_blocks()
        self.a = a = matcher.a
        self.b = b = matcher.b
        assert matches[-1] == (len(a), len(b), 0) # matches must have terminating dummy match at end
        posInf = float('inf')
        negInf = -posInf
        splits = { negInf: _Split(negInf, 0, False),
                   posInf: _Split(posInf, len(b)-1, False) }
        for m in matches:
            if m.size > 0:
                splits[m.a] = _Split(m.a, m.b, True)
                off = m.size-1
                splits[m.a+off] = _Split(m.a+off, m.b+off, False)
                
        self._splits = sorted(splits.itervalues())
        self._splitKeys = map(lambda s: s.a, self._splits)
      
    # end position is included in interval;
    # if the mapped to interval is empty, end is None;
    # returns (bBegin, bEnd, uncertain) where uncertain is (bBeginUncertain, bEndUncertain) or None
    # note that bEndUncertain may be None too to indicate empty uncertain interval
    def mapBeginEnd(self, aBegin, aEnd):
        splits = self._splits
        a = self.a
        b = self.b

        # invalid utf-8 locations can cause bizarre errors - better correct input here;
        if 0 <= aBegin < len(a):
            aBegin = utf8.adjustL(a, aBegin)
        if aEnd is not None:
            if 0 <= aEnd < len(a):
                aEnd = utf8.adjustL(a, aEnd)

        # quickly find relevant begin and end split regions
        iBegin = SortedList.largestIndexLessOrEqual(self._splitKeys, aBegin)
        sBegin = splits[iBegin]
        if aEnd is None or aEnd-aBegin <= 0:
            if sBegin.a == aBegin and iBegin > 0 and splits[iBegin-1].isMatch:
                # for single or zero character range on a split point prefer matching region
                iBegin -= 1
                sBegin = splits[iBegin]
            iEnd = iBegin
        else:
            iEnd = SortedList.largestIndexLess(self._splitKeys, aEnd)
        assert iBegin <= iEnd
        sEnd = splits[iEnd]

        isUncertain = iBegin != iEnd or not sBegin.isMatch

        # if range was completely deleted/edited, place empty match after edit;
        # if range is uncertain, move begin and end away from split point
        # to state clearly to which side of the split they belong;
        # usually that means to move begin to the right and end to the left;
        # if begin and end in b happen to be on the same split,
        # move insertion point to the left by setting end to None;
        # also: take care of utf-8
        # also: consider bEnd == None

        if sBegin.isMatch:
            bBegin = bBeginUncertain = aBegin - sBegin.a + sBegin.b
        else:
            bBegin = splits[iBegin+1].b
            bBeginUncertain = sBegin.b
            
        if sEnd.isMatch:
            bEnd = bEndUncertain = aEnd - sEnd.a + sEnd.b if aEnd is not None else None
        else:
            bEnd = sEnd.b if iBegin != iEnd else None
            bEndUncertain = splits[iEnd+1].b

        if isUncertain:
            if bEnd is None or bBegin == bEnd:
                # move split point to the left
                bEnd = None
            else:
                if not sBegin.isMatch:
                    # move begin truncation to the right
                    bBegin += 1
                if not sEnd.isMatch:
                    # move end truncation to the left
                    bEnd = utf8.adjustL(b, bEnd - 1)
                
            if bEndUncertain is None or bBeginUncertain == bEndUncertain:
                # move split point to the left
                bEndUncertain = None
            else:
                if not sBegin.isMatch:
                    # move begin truncation to the right
                    bBeginUncertain += 1
                if not sEnd.isMatch:
                    # move end truncation to the left
                    bEndUncertain = utf8.adjustL(b, bEndUncertain - 1)
                
        # prevent matches to be on start and end ' '
        if bBegin == 0:
            bBegin = 1
            if bBegin > bEnd:
                bEnd = None
        if bEnd == len(b)-1:
            bEnd = len(b)-2
            if bBegin > bEnd:
                bEnd = None
        if isUncertain:
            if bBeginUncertain == 0:
                bBeginUncertain = 1
                if bBeginUncertain > bEndUncertain:
                    bEndUncertain = None
            if bEndUncertain == len(b)-1:
                bEndUncertain = len(b)-2
                if bBeginUncertain > bEndUncertain:
                    bEndUncertain = None

            uncertain = (bBeginUncertain, bEndUncertain)
        else:
            uncertain = None

        # todo/tbd: since text matcher did not do any moved block analysis,
        # if isUncertain, try to match range anywhere else in a and b
        # e.g. find longest match in b that contains a[aBegin:aEnd]
        
        # todo/tbd: do match extrapolation into unmatched regions?

        # todo/tbd: do char based difflib match in unmatched regions to place start/end of range?
        
        # additional utf8-char based analysis:
        # if the mapped range can be found uniquely in the uncertainty range,
        # consider it a perfect match, even though it does not match on word boundaries;
        # if the mapped range can be found uniquely when ignoring spaces in text,
        # expand core match, but keep uncertainty range.
        # motivation: sometimes EPUB-MOBI conversion may add or loose a space; then we want to match partial words too
        if isUncertain and bEndUncertain and aEnd and aBegin >= 0 and aEnd < len(a):
            bEndUncX = utf8.adjustR(b, bEndUncertain+1)
            aEndX = utf8.adjustR(a, aEnd+1)
            aText = a[aBegin:aEndX]
            bBeginAlt = b.find(aText, bBeginUncertain, bEndUncX)
            if bBeginAlt >= 0 and b.find(aText, bBeginAlt+1, bEndUncX) == -1:
                # perfect match within uncertainty region
                bBegin = bBeginAlt
                bEnd = utf8.adjustL(b, bBegin + len(aText) - 1)
                uncertain = None # consider it a certain match
            else:
                start, size = findUniqueIgnoreSpace(aText, b[bBeginUncertain:bEndUncX])
                if size:
                    bBegin = bBeginUncertain+start
                    bEnd = utf8.adjustL(b, bBegin + size - 1)

        return (bBegin, bEnd, uncertain)

# search for a unique match of string a in string b, ignoring any ' ';
# return start index in b and length of match
def findUniqueIgnoreSpace(a, b):
    a = a.replace(' ','')
    bWithSpace = b
    b = b.replace(' ','')
    found = b.find(a)
    if not a or not b or found < 0 or b.find(a, found+1) >= 0:
        return -1, 0
    start = 0
    while found > 0:
        start += found
        found = bWithSpace.count(' ', start-found, start)
    while bWithSpace[start] == ' ':
        start += 1
    stop = start
    found = len(a)
    while found > 0:
        stop += found
        found = bWithSpace.count(' ', stop-found, stop)
    return start, stop-start

# implements just a sub-set of difflib.SequenceMatcher; use as
#   matcher = TextMatcher.TextMatcher(a=text1, b=text2)
#   match = matcher.get_matching_blocks()
#   quality = matcher.ratio()
#   if match is None:
#       print "matching failed because texts are too dissimilar; you may try another matcher"
# matching is word-based, where words are defined to be separated by ' ',
# i.e. words may contain any other character than ' ', including punctuation, quotes, hyphens and newline.
# Matches always include leading and trailing separator ' ', so matches may overlap by one char.
# For simplicity matched texts are required to start and end with a ' '.
class TextMatcher:
    def __init__(self, a = " ", b = " "):
        self.set_seqs(a, b)

    def set_seqs(self, a, b):
        assert a[0] == a[-1] == b[0] == b[-1] == ' '
        self.a = a
        self.b = b
        self.matching_blocks = None
        self._ratio = None

    def ratio(self):
        if self._ratio is None:
            self._ratio = _calculateRatio(self.get_matching_blocks(), len(self.a), len(self.b))
        return self._ratio

    def get_matching_blocks(self):
        if self.matching_blocks is not None:
            return self.matching_blocks
            
        if not self.a or not self.b:
            return None
        if not isinstance(self.a, bytes) or not isinstance(self.b, bytes):
            raise TypeError()
        a = self.a
        b = self.b
        
        # the relevant cases are, in order of expactation and relevance:
        # 1) a and b are the same
        # 2) a and b are the same except for some relatively short prefix/postfix
        # 3) a and b are the same but were edited at some places
        # 4) a and b are the same but were edited at a lot of places
        # 5) a and b are not the same
        #
        # we try to be fast for the relevant cases 1 to 4;
        # case 5 may be treated as an error or take a long time
        #
        # performance note:
        # difflib.SequenceMatcher().get_matching_blocks() takes very long on book texts;
        # minutes have been observed for 500.000 character texts, even if texts are identical.
        # Working on words instead of characters improves matching speed significantly.
        
        # optimization for case 1
        if a == b:
            self.matching_blocks = [
                Match(0, 0, len(a)),
                Match(len(a), len(b), 0) ]
            return self.matching_blocks

        # optimization for cases 2 and 3
        # use seed matcher to find some very long, obviously matching byte sequences fast,
        # then fill in the gaps with difflib.SequenceMatcher working on words
        matcher = SeedMatcher( 30, 41 ) # note: parameters were tuned using 500KB english text with random edits
        matches = matcher.seedMatches( a, b )
        self._seedMatchesPostprocess( matches )

        # fallback for cases 4 and 5 and to improve on cases 2 and 3
        # try to fill the gaps using difflib.SequenceMatcher on words
        matches = self._insertDifflibWordMatches(matches)

        matches.append( Match(len(a), len(b), 0) )
        self.matching_blocks = matches
        return self.matching_blocks

    # postprocess matches from SeedMatcher:
    # - for consistency with final difflib matching, adjust matches to word boundaries
    # - matches may be overlapping; disambiguate by splitting at last space in overlap;
    #   space between words is included in both matches;
    #   if no space in overlap, exclude overlap from both matches
    def _seedMatchesPostprocess(self, matches):
        a = self.a

        for i in xrange(1, len(matches)):
            m0 = matches[i-1]
            m1 = matches[i]
            # split ambiguous overlap if possible
            overlap = max( m0.a+m0.size - m1.a, m0.b+m0.size - m1.b )
            if overlap > 0:
                # take off from m1 everything up to the last space in overlap
                off1 = a.rfind(' ', m1.a, m1.a+overlap) - m1.a
                # no space in overlap => no action needed (word boundary adjustment will exclude from both m0 and m1)
                if off1 >= 0:
                    off0 = overlap - off1 - 1 # space maps to both m0 and m1
                    matches[i-1] = Match(m0.a, m0.b, m0.size-off0)
                    matches[i] = Match(m1.a+off1, m1.b+off1, m1.size-off1)

        i = 0
        while i < len(matches):
            m = matches[i]
            # adjust left and right border of match to word boundary (i.e. to be a space)
            start = a.find(' ', m.a, m.a+m.size) if m.a > 0 else 0
            stop = a.rfind(' ', m.a, m.a+m.size) if m.a+m.size < len(a) else len(a)-1
            size = stop-start+1 # +1 to include space at end
            if size >= 3: # minimum valid size is 3 (one character and two spaces)
                matches[i] = Match(start, m.b+start-m.a, size)
                i += 1
            else:
                del matches[i]

    def _insertDifflibWordMatches(self, matches):
        if not matches:
            return self._getDifflibWordMatches(0, len(self.a), 0, len(self.b))

        matches2 = []
        m0 = Match(1,1,0)
        for m1 in matches:
            matches2.extend( self._getDifflibWordMatches(m0.a+m0.size-1, m1.a+1, m0.b+m0.size-1, m1.b+1) )
            matches2.append( m1 )
            m0 = m1
        matches2.extend( self._getDifflibWordMatches(m0.a+m0.size-1, len(self.a), m0.b+m0.size-1, len(self.b)) )
        return matches2
        
    def _getDifflibWordMatches(self, a0, a1, b0, b1):
        assert self.a[a0] == ' '
        assert self.b[b0] == ' '
        assert self.a[a1-1] == ' '
        assert self.b[b1-1] == ' '
        if a1 - a0 < 2 or b1 - b0 < 2:
            return []
        # split into words
        aWords = self.a[a0+1:a1-1].split(' ')
        bWords = self.b[b0+1:b1-1].split(' ')
        # get difflib matches
        # note: autojunk = False gives better and more stable matches
        #       autojunk = True is a little faster, but which words are considered junk varies wildly
        matcher = _difflibMatcher(a=aWords, b=bWords, autojunk=False)
        matches = matcher.get_matching_blocks()
        matches.pop() # discard dummy match Match(len(aWords),len(bWords),0) at the end
        # convert word matches to string matches
        aWord = 0
        bWord = 0
        aChars = 0
        bChars = 0
        for i in xrange(len(matches)):
            m = matches[i]
            # sum up words before match to get start in characters
            if aWord < m.a:
                aChars += len( ' '.join(aWords[aWord:m.a]) ) + 1 # include joining space before
            if bWord < m.b:
                bChars += len( ' '.join(bWords[bWord:m.b]) ) + 1 # include joining space before
            # sum up words in match to get length in characters
            lenChars = len( ' '.join(aWords[m.a:m.a+m.size]) ) + 1 # include joining space after
            # transform match
            matches[i] = Match(a0+aChars, b0+bChars, lenChars+1) # include surrounding spaces
            assert self.a[matches[i].a:matches[i].a+matches[i].size] == self.b[matches[i].b:matches[i].b+matches[i].size]
            # update positions and sums
            aWord = m.a+m.size
            bWord = m.b+m.size
            aChars += lenChars
            bChars += lenChars
        return matches

class SeedMatcher:
    def __init__(self,
                 seedLen, # length of a seed
                 maxTries, # number of tries to find a valid seed existing in both strings
                 minMatchLen = 0): # minimum length of a match
        self.seedLen = max( int(seedLen), 1 )
        self.minMatchLen = max( int(minMatchLen), self.seedLen )
        self.maxTries = max( int(maxTries), 0 )
        
    def _init(self, a, b):
        self.a = a
        self.b = b
        self._cachedMatches = {}

    def seedMatches(self, a, b):
        self._init(a, b)
        return self._seedMatches(0, len(a), 0, len(b))
        
    def _seedMatches(self, a0, a1, b0, b1):
        #print "_seedMatches in", "'%.9s'" % self.a[a0:a1], "'%.9s'" % self.b[b0:b1], a0, b0, b1-b0
        match = self._seedMatch(a0, a1, b0, b1)
        if not match:
            return []

        matches = self._seedMatches(a0, match.a, b0, match.b)
        matches.append( match )
        matches.extend( self._seedMatches(match.a+match.size, a1, match.b+match.size, b1) )
            
        return matches
            
    def seedMatch(self, a, b):
        self._init(a, b)
        return self._seedMatch(0, len(a), 0, len(b))
        
    def _seedMatch(self, a0, a1, b0, b1):
        #print "_seedMatch in", "'%.9s'" % self.a[a0:a1], "'%.9s'" % self.b[b0:b1], a0, b0, b1-b0
        lenB = b1-b0
        minLen = min(a1-a0, lenB)
        seedLen = self.seedLen
        maxTries = self.maxTries
        if seedLen <= 0 or minLen < seedLen or maxTries <= 0:
            return None
        oddTry = maxTries%2
        tryStep = int((lenB-seedLen+1)/(maxTries-oddTry)) if maxTries > 1 else 1
        if tryStep < 1:
            tryStep = 1
            maxTries = lenB-seedLen+1
            oddTry = maxTries%2
        #print "_findSeed maxTries", maxTries
        bestMatch = None
        bestMatchLen = 0
        for seedIdx in xrange(-oddTry, maxTries-oddTry):
            #print "_findSeed seedIdx", seedIdx
            if seedIdx < 0: # in the middle
                seedStart = b0 + int((lenB-seedLen)/2)
            else:
                off = int((maxTries-oddTry-1 - seedIdx)/2)
                #print "_findSeed off", off
                if seedIdx % 2: # to the right, ending at end
                    seedStart = b1 - seedLen - off*tryStep
                else: # to the left, ending at start
                    seedStart = b0 + off*tryStep
            seedEnd = seedStart + seedLen
            seed = self.b[seedStart:seedEnd]
            #print "_findSeed seed >>>", seed if len(seed)<9 else seed[:4]+"..."+seed[-4:], seedStart, len(seed)
            assert len(seed) == seedLen
            seedStartA = self.a.find(seed, a0, a1)
            if seedStartA < 0:
                continue # seed not found in a
            if self.b.find(seed, seedStart+1, b1) > 0:
                continue # seed not unique in b
            if self.a.find(seed, seedStartA+1, a1) > 0:
                continue # seed not unique in a
            match = self._expandSeed(seedStartA, seedStart, len(seed))
            if match.size >= self.minMatchLen:
                if 2*match.size >= minLen:
                    return match
                elif bestMatchLen < match.size:
                    bestMatch = match
                    bestMatchLen = match.size
        return bestMatch
    
    def _expandSeed(self, a0S, b0S, lenS):
        # expand to the right
        #print "_expandSeed in", seed
        # check cache of previous results
        plane = a0S-b0S
        #cmatch = None
        if self._cachedMatches.has_key(plane):
            for match in self._cachedMatches[plane]:
                if match.a <= a0S < match.a+match.size:
                    return match # found cached match for this seed
                    #cmatch = match
                    #break
        #
        # notes about optimization:
        # - idea is to reduce number of compare operations and using fast native operator '==' on large string slices
        # - optimization is definitely worth it for large texts: seconds become milliseconds
        # - the larger initial step the better (measured for step in [16, 256, 1024, 1<<15] for 2*570KB of text)
        # - using memoryview instead of strings is about twice as fast for large step (1024 or higher),
        #   but only half as fast for small initial step (16 or lower)
        #
        a = memoryview(self.a)
        b = memoryview(self.b)
        match = expandSeed(a, b, a0S, b0S, lenS)
        # write match to cache
        #if cmatch:
        #    if not cmatch == match:
        #        assert cmatch == match
        #else:
        if not self._cachedMatches.has_key(plane):
            self._cachedMatches[plane] = [match]
        else:
            self._cachedMatches[plane].append(match)
        return match

def expandSeed(a, b, a0S, b0S, lenS, step_log2=15):
    a1S = a0S+lenS
    b1S = b0S+lenS
    assert a[a0S:a1S] == b[b0S:b1S]
    # expand to the right
    aStop = len(a) if len(a)-a0S <= len(b)-b0S else a0S + len(b)-b0S
    step = 1<<step_log2
    while step > 0:
        a1Ss = a1S+step
        b1Ss = b1S+step
        while a1Ss <= aStop and a[a1S:a1Ss] == b[b1S:b1Ss]:
            a1S = a1Ss
            b1S = b1Ss
            a1Ss = a1S+step
            b1Ss = b1S+step
        step >>= 1
    assert a[a0S:a1S] == b[b0S:b1S]
    # expand to the left
    aStop = 0 if a0S <= b0S else a0S-b0S
    step = 1<<step_log2
    while step > 0:
        a0Ss = a0S-step
        b0Ss = b0S-step
        while a0Ss >= aStop and a[a0Ss:a0S] == b[b0Ss:b0S]:
            a0S = a0Ss
            b0S = b0Ss
            a0Ss = a0S-step
            b0Ss = b0S-step
        step >>= 1
    match = Match(a0S, b0S, b1S-b0S)
    #print "_expandSeed out", match
    assert a[match.a:match.a+match.size] == b[match.b:match.b+match.size]
    return match

def expandSeedUtf8(a, b, a0S, b0S, lenS, step_log2=15):#(a, b, matches, step_log2=6):
    m = expandSeed(a, b, a0S, b0S, lenS, step_log2)
    ma, adjust = utf8.adjustR2(a, m.a)
    mb = m.b + adjust
    ms = m.size - adjust
    if ma+ms < len(a):
        __, adjust = utf8.adjustL2(a, ma+ms)
        ms -= adjust
    return Match(ma, mb, ms)

# some tests for developer analysis
if __debug__ and __name__ == '__main__':

    doPrint = False

    def printStartTest(*args):
        import inspect
        stack = inspect.stack()
        print "TEST", " ".join(map(lambda i: stack[i][3], range(len(stack)-2, 0, -1))) + str(args)
        
    def printEndTest():
        print "OK"

    def test_findUniqueIgnoreSpace():
        printStartTest()
        
        def test(a, b, matchstrExpected):
            start, size = findUniqueIgnoreSpace(a, b)
            matchstr = b[start:start+size]
            assert matchstr == matchstrExpected, "%r" % matchstr
            
        test("", "", "")
        test(" ", " ", "")
        test("abc", "abc", "abc")
        test("abc", "abcde", "abc")
        test("cde", "abcde", "cde")
        test("abc", "abcabc", "")
        test("cdef g", "ab cde fgh", "cde fg")
        test("  c d e f g  ", "ab cde fgh", "cde fg")
        test("cdefg", "ab cde fgh", "cde fg")
        test("cdefg", "abcdefgh", "cdefg")
        test("cdefg", "  abcdefgh", "cdefg")
        test("cdefg", "a  bcdefgh", "cdefg")
        test("cdefg", "ab  cdefgh", "cdefg")
        test("cdefg", "abc  defgh", "c  defg")
        test("cdefg", "abcd  efgh", "cd  efg")
        test("cdefg", "abcde  fgh", "cde  fg")
        test("cdefg", "abcdef  gh", "cdef  g")
        test("cdefg", "abcdefg  h", "cdefg")
        test("cdefg", "abcdefgh  ", "cdefg")
        
        printEndTest()
        
    def test_seedMatch():
        printStartTest()

        match = SeedMatcher(1, 1).seedMatch("", "")
        if doPrint:
            print "match", match
        assert match is None
        
        match = SeedMatcher(1, 9).seedMatch("abcd", "abcd")
        if doPrint:
            print "match", match
        assert match == Match(0,0,4)

        match = SeedMatcher(1, 9).seedMatch("a", "a")
        if doPrint:
            print "match", match
        assert match == Match(0,0,1)

        match = SeedMatcher(1, 9).seedMatch("aa", "aa")
        if doPrint:
            print "match", match
        assert match is None

        match = SeedMatcher(2, 1).seedMatch("aa", "aa")
        if doPrint:
            print "match", match
        assert match == Match(0,0,2)

        match = SeedMatcher(1, 1).seedMatch("ab", "ab")
        if doPrint:
            print "match", match
        assert match == Match(0,0,2)

        match = SeedMatcher(1, 3).seedMatch("aaa", "aaa")
        if doPrint:
            print "match", match
        assert match is None

        match = SeedMatcher(1, 3).seedMatch("baa", "baa")
        if doPrint:
            print "match", match
        assert match == Match(0,0,3)

        match = SeedMatcher(1, 3).seedMatch("aba", "aba")
        if doPrint:
            print "match", match
        assert match == Match(0,0,3)

        match = SeedMatcher(1, 3).seedMatch("aab", "aab")
        if doPrint:
            print "match", match
        assert match == Match(0,0,3)

        match = SeedMatcher(1, 9).seedMatch("aaaa", "aaaa")
        if doPrint:
            print "match", match
        assert match is None
        
        match = SeedMatcher(1, 4).seedMatch("baaa", "baaa")
        if doPrint:
            print "match", match
        assert match == Match(0,0,4)
        
        match = SeedMatcher(1, 4).seedMatch("abaa", "abaa")
        if doPrint:
            print "match", match
        assert match == Match(0,0,4)
        
        match = SeedMatcher(1, 4).seedMatch("aaba", "aaba")
        if doPrint:
            print "match", match
        assert match == Match(0,0,4)
        
        match = SeedMatcher(1, 4).seedMatch("aaab", "aaab")
        if doPrint:
            print "match", match
        assert match == Match(0,0,4)

        match = SeedMatcher(1, 9).seedMatch("aaaaa", "aaaaa")
        if doPrint:
            print "match", match
        assert match is None
        
        match = SeedMatcher(1, 6).seedMatch("cbaaaa", "baaaac")
        if doPrint:
            print "match", match
        assert match == Match(1,0,5)
        
        match = SeedMatcher(1, 6).seedMatch("cabaaa", "abaaac")
        if doPrint:
            print "match", match
        assert match == Match(1,0,5)
        
        match = SeedMatcher(1, 6).seedMatch("caabaa", "aabaac")
        if doPrint:
            print "match", match
        assert match == Match(1,0,5)
        
        match = SeedMatcher(1, 6).seedMatch("caaaba", "aaabac")
        if doPrint:
            print "match", match
        assert match == Match(1,0,5)
        
        match = SeedMatcher(1, 6).seedMatch("caaaab", "aaaabc")
        if doPrint:
            print "match", match
        assert match == Match(1,0,5)
        
        match = SeedMatcher(2, 6).seedMatch("ababab", "ababab")
        if doPrint:
            print "match", match
        assert match is None
        
        core = "12345" * 30
        str1 = "abc" * 100 + core + "abc" * 500
        str2 = "def" * 500 + core + "def" * 100
        match = SeedMatcher(150,len(str2)).seedMatch(str1, str2)
        if doPrint:
            print "match", match
        assert match == Match(300,1500,150)

        core = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
        str1 = "123" * 100 + core + "123" * 500
        str2 = "789" * 500 + core + "789" * 100
        match = SeedMatcher(5, 51).seedMatch(str1, str2)
        if doPrint:
            print "match", match
        assert match == Match(300,1500,52)
        
        printEndTest()

    def test_seedMatches():
        printStartTest()

        match = SeedMatcher(1,10).seedMatches("", "")
        if doPrint:
            print "match", match
        assert match == []
        
        match = SeedMatcher(1, 9).seedMatches("abcd", "abcd")
        if doPrint:
            print "match", match
        assert match == [Match(0,0,4)]

        match = SeedMatcher(1, 6).seedMatches("uvabwcdewf", "abxcdyefz")
        if doPrint:
            print "match", match
        assert match == [(2,0,2), (5,3,2), (7,6,1), (9,7,1)]
        
        printEndTest()
        
    def test_seedMatchesPostprocess():
        printStartTest()
        
        def checkMatches(matcher, matches):
            a = matcher.a
            b = matcher.b
            for m in matches:
                assert m.a >= 0 and m.a+m.size <= len(a)
                assert m.b >= 0 and m.b+m.size <= len(b)
                assert a[m.a:m.a+m.size] == b[m.b:m.b+m.size]
                assert m.a == 0 or m.b == 0 or a[m.a-1] != b[m.b-1]
                assert m.a+m.size == len(a) or m.b+m.size == len(b) or a[m.a+m.size] != b[m.b+m.size]
                
        def checkResults(matcher, matches, expectedMatches, expectedRatio):
            ratio = _calculateRatio(matches, len(matcher.a), len(matcher.b))
            if doPrint:
                print "ratio =", ratio, "matches =", matches
            assert matches == expectedMatches
            assert abs(ratio - expectedRatio) < 1e-9

        matcher = TextMatcher()
        
        matcher.set_seqs(" ", " ")
        matches = [Match(0,0,1)]
        checkMatches(matcher, matches)
        matcher._seedMatchesPostprocess(matches)
        checkResults(matcher, matches, [], 0.0)
        
        matcher.set_seqs(" a ", " a ")
        matches = [Match(0,0,3)]
        checkMatches(matcher, matches)
        matcher._seedMatchesPostprocess(matches)
        checkResults(matcher, matches, [Match(0,0,3)], 1.0)
        
        matcher.set_seqs(" a ", " b ")
        matches = [Match(0,0,1), Match(2,2,1)]
        checkMatches(matcher, matches)
        matcher._seedMatchesPostprocess(matches)
        checkResults(matcher, matches, [], 0.0)
        
        matcher.set_seqs(" aaa ",
                         " aa ")
        #                 012345678901234567890
        #                 0         1         2
        matches = [Match(0,0,3), Match(2,1,3)]
        checkMatches(matcher, matches)
        matcher._seedMatchesPostprocess(matches)
        checkResults(matcher, matches, [], 0.0)
        
        matcher.set_seqs(" word1 word2 ",
                         " word1 word3 ")
        #                 012345678901234567890
        #                 0         1         2
        matches = [Match(0,0,11), Match(12,12,1)]
        checkMatches(matcher, matches)
        matcher._seedMatchesPostprocess(matches)
        checkResults(matcher, matches, [Match(0,0,7)], 7/13.0)
        
        matcher.set_seqs(" x wordword y ",
                         " u word v ")
        #                 012345678901234567890
        #                 0         1         2
        matches = [Match(0,0,1), Match(2,2,5), Match(7,3,5)]
        checkMatches(matcher, matches)
        matcher._seedMatchesPostprocess(matches)
        checkResults(matcher, matches, [], 0.0)
        
        matcher.set_seqs(" x wordword y ",
                         " x word y ")
        #                 012345678901234567890
        #                 0         1         2
        matches = [Match(0,0,7), Match(7,3,7)]
        checkMatches(matcher, matches)
        matcher._seedMatchesPostprocess(matches)
        checkResults(matcher, matches, [Match(0,0,3), Match(11,7,3)], 6/12.0)
        
        matcher.set_seqs(" x word word y ",
                         " x word y ")
        #                 012345678901234567890
        #                 0         1         2
        matches = [Match(0,0,8), Match(7,2,8)]
        checkMatches(matcher, matches)
        matcher._seedMatchesPostprocess(matches)
        checkResults(matcher, matches, [Match(0,0,8), Match(12,7,3)], 10/12.5)
        
        matcher.set_seqs(" word word word ",
                         " word word ")
        #                 012345678901234567890
        #                 0         1         2
        matches = [Match(0,0,11), Match(5,0,11)]
        checkMatches(matcher, matches)
        matcher._seedMatchesPostprocess(matches)
        checkResults(matcher, matches, [Match(0,0,11)], 11/13.5)
        
        matcher.set_seqs(" x word word ",
                         " x wordword word ")
        #                 012345678901234567890
        #                 0         1         2
        matches = [Match(0,0,7), Match(3,7,10)]
        checkMatches(matcher, matches)
        matcher._seedMatchesPostprocess(matches)
        checkResults(matcher, matches, [Match(0,0,3), Match(7,11,6)], 9/15.0)
        
        printEndTest()
        
    def test_getDifflibWordMatches():
        printStartTest()
        
        matcher = TextMatcher(" ", " ")
        matches = matcher._getDifflibWordMatches(0, len(matcher.a), 0, len(matcher.b))
        if doPrint:
            print "matches =", matches
        assert matches == []
        
        matcher = TextMatcher("  ", "  ")
        matches = matcher._getDifflibWordMatches(0, len(matcher.a), 0, len(matcher.b))
        if doPrint:
            print "matches =", matches
        assert matches == [Match(0,0,2)]
        
        matcher = TextMatcher("   ", "   ")
        matches = matcher._getDifflibWordMatches(0, len(matcher.a), 0, len(matcher.b))
        if doPrint:
            print "matches =", matches
        assert matches == [Match(0,0,3)]
        
        matcher = TextMatcher("    ", "    ")
        matches = matcher._getDifflibWordMatches(0, len(matcher.a), 0, len(matcher.b))
        if doPrint:
            print "matches =", matches
        assert matches == [Match(0,0,4)]
        
        matcher = TextMatcher("    ", "    ")
        matches = matcher._getDifflibWordMatches(1, len(matcher.a), 1, len(matcher.b))
        if doPrint:
            print "matches =", matches
        assert matches == [Match(1,1,3)]
        
        matcher = TextMatcher(" word1 ", " word1 ")
        matches = matcher._getDifflibWordMatches(0, len(matcher.a), 0, len(matcher.b))
        if doPrint:
            print "matches =", matches
        assert matches == [Match(0,0,7)]
        
        matcher = TextMatcher(" word1 ", " word2 ")
        matches = matcher._getDifflibWordMatches(0, len(matcher.a), 0, len(matcher.b))
        if doPrint:
            print "matches =", matches
        assert matches == []
        
        matcher = TextMatcher(" a b ", " x b ")
        matches = matcher._getDifflibWordMatches(0, len(matcher.a), 0, len(matcher.b))
        if doPrint:
            print "matches =", matches
        assert matches == [Match(2,2,3)]

        matcher = TextMatcher(" a b c ", " a x c ")
        matches = matcher._getDifflibWordMatches(0, len(matcher.a), 0, len(matcher.b))
        if doPrint:
            print "matches =", matches
        assert matches == [Match(0,0,3), Match(4,4,3)]
        
        matcher = TextMatcher(" a b c ",
                              " a x b y c z ")
        #                      012345678901234567890
        #                      0         1         2
        matches = matcher._getDifflibWordMatches(0, len(matcher.a), 0, len(matcher.b))
        if doPrint:
            print "matches =", matches
        assert matches == [Match(0,0,3), Match(2,4,3), Match(4,8,3)]
        
        matcher = TextMatcher(" aa bbb ",
                              " aa xxxx yyyyy bbb ")
        #                      012345678901234567890
        #                      0         1         2
        matches = matcher._getDifflibWordMatches(0, len(matcher.a), 0, len(matcher.b))
        if doPrint:
            print "matches =", matches
        assert matches == [Match(0,0,4), Match(3,14,5)]

        printEndTest()
        
    def test_insertDifflibWordMatches():
        printStartTest()
        
        matcher = TextMatcher(" a b c ",
                              " a x c ")
        #                      012345678901234567890
        #                      0         1         2
        matches = matcher._insertDifflibWordMatches( [] )
        if doPrint:
            print "matches =", matches
        assert matches == [Match(0,0,3), Match(4,4,3)]
        
        matcher = TextMatcher(" a b c ",
                              " a x c ")
        #                      012345678901234567890
        #                      0         1         2
        matches = matcher._insertDifflibWordMatches( [Match(0,0,3)] )
        if doPrint:
            print "matches =", matches
        assert matches == [Match(0,0,3), Match(4,4,3)]
        
        matcher = TextMatcher(" a b c ",
                              " a x c ")
        #                      012345678901234567890
        #                      0         1         2
        matches = matcher._insertDifflibWordMatches( [Match(4,4,3)] )
        if doPrint:
            print "matches =", matches
        assert matches == [Match(0,0,3), Match(4,4,3)]
        
        matcher = TextMatcher(" a b c d e ",
                              " a x c y e ")
        #                      012345678901234567890
        #                      0         1         2
        matches = matcher._insertDifflibWordMatches( [] )
        if doPrint:
            print "matches =", matches
        assert matches == [Match(0,0,3), Match(4,4,3), Match(8,8,3)]
        
        matcher = TextMatcher(" a b c d e ",
                              " a x c y e ")
        #                      012345678901234567890
        #                      0         1         2
        matches = matcher._insertDifflibWordMatches( [Match(0,0,3)] )
        if doPrint:
            print "matches =", matches
        assert matches == [Match(0,0,3), Match(4,4,3), Match(8,8,3)]
        
        matcher = TextMatcher(" a b c d e ",
                              " a x c y e ")
        #                      012345678901234567890
        #                      0         1         2
        matches = matcher._insertDifflibWordMatches( [Match(4,4,3)] )
        if doPrint:
            print "matches =", matches
        assert matches == [Match(0,0,3), Match(4,4,3), Match(8,8,3)]
        
        matcher = TextMatcher(" a b c d e ",
                              " a x c y e ")
        #                      012345678901234567890
        #                      0         1         2
        matches = matcher._insertDifflibWordMatches( [Match(8,8,3)] )
        if doPrint:
            print "matches =", matches
        assert matches == [Match(0,0,3), Match(4,4,3), Match(8,8,3)]
        
        matcher = TextMatcher(" a b c d e ",
                              " a x c y e ")
        #                      012345678901234567890
        #                      0         1         2
        matches = matcher._insertDifflibWordMatches( [Match(4,4,3), Match(8,8,3)] )
        if doPrint:
            print "matches =", matches
        assert matches == [Match(0,0,3), Match(4,4,3), Match(8,8,3)]
        
        matcher = TextMatcher(" a b c d e ",
                              " a x c y e ")
        #                      012345678901234567890
        #                      0         1         2
        matches = matcher._insertDifflibWordMatches( [Match(0,0,3), Match(8,8,3)] )
        if doPrint:
            print "matches =", matches
        assert matches == [Match(0,0,3), Match(4,4,3), Match(8,8,3)]
        
        matcher = TextMatcher(" a b c d e ",
                              " a x c y e ")
        #                      012345678901234567890
        #                      0         1         2
        matches = matcher._insertDifflibWordMatches( [Match(0,0,3), Match(4,4,3)] )
        if doPrint:
            print "matches =", matches
        assert matches == [Match(0,0,3), Match(4,4,3), Match(8,8,3)]
        
        matcher = TextMatcher(" a b c d e ",
                              " a c e ")
        #                      012345678901234567890
        #                      0         1         2
        matches = matcher._insertDifflibWordMatches( [] )
        if doPrint:
            print "matches =", matches
        assert matches == [Match(0,0,3), Match(4,2,3), Match(8,4,3)]
        
        matcher = TextMatcher(" a b c d e ",
                              " a c e ")
        #                      012345678901234567890
        #                      0         1         2
        matches = matcher._insertDifflibWordMatches( [Match(0,0,3)] )
        if doPrint:
            print "matches =", matches
        assert matches == [Match(0,0,3), Match(4,2,3), Match(8,4,3)]
        
        matcher = TextMatcher(" a b c d e ",
                              " a c e ")
        #                      012345678901234567890
        #                      0         1         2
        matches = matcher._insertDifflibWordMatches( [Match(4,2,3)] )
        if doPrint:
            print "matches =", matches
        assert matches == [Match(0,0,3), Match(4,2,3), Match(8,4,3)]
        
        matcher = TextMatcher(" a b c d e ",
                              " a c e ")
        #                      012345678901234567890
        #                      0         1         2
        matches = matcher._insertDifflibWordMatches( [Match(8,4,3)] )
        if doPrint:
            print "matches =", matches
        assert matches == [Match(0,0,3), Match(4,2,3), Match(8,4,3)]
        
        matcher = TextMatcher(" a b c d e ",
                              " a c e ")
        #                      012345678901234567890
        #                      0         1         2
        matches = matcher._insertDifflibWordMatches( [Match(0,0,3), Match(4,2,3), Match(8,4,3)] )
        if doPrint:
            print "matches =", matches
        assert matches == [Match(0,0,3), Match(4,2,3), Match(8,4,3)]
        

        matcher = TextMatcher(" a b c d e ",
                              " a c e ")
        #                      012345678901234567890
        #                      0         1         2
        matches = matcher._insertDifflibWordMatches( [Match(4,2,3), Match(8,4,3)] )
        if doPrint:
            print "matches =", matches
        assert matches == [Match(0,0,3), Match(4,2,3), Match(8,4,3)]
        
        matcher = TextMatcher(" a b c d e ",
                              " a c e ")
        #                      012345678901234567890
        #                      0         1         2
        matches = matcher._insertDifflibWordMatches( [Match(0,0,3), Match(8,4,3)] )
        if doPrint:
            print "matches =", matches
        assert matches == [Match(0,0,3), Match(4,2,3), Match(8,4,3)]
        
        matcher = TextMatcher(" a b c d e ",
                              " a c e ")
        #                      012345678901234567890
        #                      0         1         2
        matches = matcher._insertDifflibWordMatches( [Match(0,0,3), Match(4,2,3)] )
        if doPrint:
            print "matches =", matches
        assert matches == [Match(0,0,3), Match(4,2,3), Match(8,4,3)]
        
        printEndTest()
        
    def test_MatchMapper():
        printStartTest()
        
        def applyMatch(match, text, adjustUtf8 = False):
            
            if adjustUtf8:
                def adjust(text, loc):
                    if loc is None or loc < 0 or loc >= len(text):
                        return loc
                    return utf8.adjustL(text, loc)
                    
                unc = match[2]
                if unc:
                    unc = (adjust(text, unc[0]),
                           adjust(text, unc[1]))
                match = (adjust(text, match[0]),
                         adjust(text, match[1]),
                         unc)
            
            def applyChanges(changesList, oldHtml):
                newHtml = []
                locCur = 0
                for loc, off, ka, text in changesList:
                    assert off in (0, 1)
                    if loc < locCur:
                        loc = locCur
                    elif off:
                        loc = utf8.adjustR(oldHtml, loc+1)
                    newHtml.append(oldHtml[locCur:loc])
                    newHtml.append(text)
                    locCur = loc
                newHtml.append(oldHtml[locCur:])
                return ''.join(newHtml)

            changesList = []
            ka = None
            if match[1] is not None:
                changesList.append([match[0], 0, ka, '['])
                changesList.append([match[1], 1, ka, ']'])
            else:
                changesList.append([match[0], 0, ka, '[]'])
            if match[2]:
                changesList.insert(0, [match[2][0], 0, ka, '{'])
                if match[2][1] is not None:
                    changesList.append([match[2][1], 1, ka, '}'])
                else:
                    changesList[-1][3] += '}'
            newText = applyChanges(changesList, text)
            try:
                newText = newText.decode('utf-8')
            except:
                print "ERROR: invalid utf-8 in", newText.encode('string_escape')
                raise
            return newText

        expectedResults = {(' \xc3\x84\xc3\x84 \xc3\x96\xc3\x96 \xc3\x9c\xc3\x9c ',
            ' \xc3\x84\xc3\x84 \xc3\x9c\xc3\x9c '):{(6,9):(5,None,(5,None)),(7,12):(6,6,(6,6)),(4,None):(3,None,
            None),(11,11):(6,6,None),(-1,14):(1,8,(1,8)),(0,4):(1,3,None),(1,1):(1,1,None),(4,10):(3,3,(3,3)),(2,
            6):(1,3,(1,3)),(5,11):(6,6,(6,6)),(4,5):(3,5,None),(2,2):(1,1,None),(4,16):(3,8,(3,8)),(6,None):(5,
            None,(5,None)),(8,12):(6,6,(6,6)),(9,9):(5,None,(5,None)),(10,14):(5,8,None),(11,15):(6,9,None),(15,
            16):(10,None,(10,None)),(11,16):(6,8,(6,8)),(13,13):(8,8,None),(0,14):(1,8,(1,8)),(3,11):(3,6,(3,6)),
            (1,15):(1,9,(1,9)),(4,12):(3,6,(3,6)),(2,12):(1,6,(1,6)),(6,14):(6,8,(6,8)),(7,15):(6,9,(6,9)),(0,6):
            (1,3,(1,3)),(-1,2):(1,1,(1,1)),(0,None):(1,None,None),(0,5):(1,5,None),(0,8):(1,3,(1,3)),(4,11):(3,6,
            (3,6)),(3,5):(3,5,None),(2,7):(1,3,(1,3)),(-1,None):(1,None,(1,None)),(5,10):(5,None,(5,None)),(4,6):
            (3,3,(3,3)),(5,7):(5,None,(5,None)),(8,13):(6,8,(6,8)),(10,15):(5,9,None),(2,None):(1,None,None),(-1,
            5):(1,5,(1,5)),(0,15):(1,9,(1,9)),(3,10):(3,3,(3,3)),(1,14):(1,8,(1,8)),(4,13):(3,8,(3,8)),(5,6):(5,
            None,(5,None)),(2,13):(1,8,(1,8)),(3,16):(3,8,(3,8)),(6,15):(6,9,(6,9)),(7,14):(6,8,(6,8)),(-1,12):
            (1,6,(1,6)),(14,16):(8,8,(8,8)),(-1,3):(1,3,(1,3)),(10,16):(5,8,(5,8)),(14,None):(8,None,None),(1,7):
            (1,3,(1,3)),(0,9):(1,3,(1,3)),(3,4):(3,3,None),(2,4):(1,3,None),(5,9):(5,None,(5,None)),(4,7):(3,3,
            (3,3)),(6,6):(5,None,(5,None)),(7,None):(5,None,(5,None)),(12,None):(6,None,None),(7,7):(5,None,(5,
            None)),(7,16):(6,8,(6,8)),(8,14):(6,8,(6,8)),(9,15):(6,9,(6,9)),(10,12):(5,6,None),(-1,10):(1,3,(1,
            3)),(14,14):(8,8,None),(15,15):(10,None,None),(2,16):(1,8,(1,8)),(0,0):(1,None,None),(3,13):(3,8,(3,
            8)),(1,13):(1,8,(1,8)),(4,14):(3,8,(3,8)),(2,10):(1,3,(1,3)),(5,15):(6,9,(6,9)),(6,12):(6,6,(6,6)),
            (7,9):(5,None,(5,None)),(3,None):(3,None,None),(8,None):(5,None,(5,None)),(8,8):(5,None,(5,None)),
            (-1,13):(1,8,(1,8)),(-1,0):(1,None,(1,None)),(12,12):(6,6,None),(0,7):(1,3,(1,3)),(1,6):(1,3,(1,3)),
            (0,10):(1,3,(1,3)),(3,7):(3,3,(3,3)),(2,5):(1,5,None),(1,11):(1,6,(1,6)),(5,8):(5,None,(5,None)),(6,
            7):(5,None,(5,None)),(5,5):(5,5,None),(5,None):(5,None,None),(6,10):(5,None,(5,None)),(10,None):(5,
            None,None),(8,15):(6,9,(6,9)),(9,14):(6,8,(6,8)),(10,13):(5,8,None),(-1,11):(1,6,(1,6)),(9,16):(6,8,
            (6,8)),(14,15):(8,9,None),(0,1):(1,1,None),(3,12):(3,6,(3,6)),(1,12):(1,6,(1,6)),(4,15):(3,9,(3,9)),
            (2,11):(1,6,(1,6)),(5,14):(6,8,(6,8)),(17,None):(10,None,(10,None)),(7,8):(5,None,(5,None)),(6,16):
            (6,8,(6,8)),(16,16):(10,None,(10,None)),(11,14):(6,8,None),(8,9):(5,None,(5,None)),(15,None):(10,
            None,None),(-1,1):(1,1,(1,1)),(12,13):(6,8,None),(1,16):(1,8,(1,8)),(13,16):(8,8,(8,8)),(1,5):(1,5,
            None),(0,11):(1,6,(1,6)),(3,6):(3,3,(3,3)),(-1,-1):(1,None,(1,None)),(1,10):(1,3,(1,3)),(6,13):(6,8,
            (6,8)),(6,11):(6,6,(6,6)),(-1,4):(1,3,(1,3)),(9,13):(6,8,(6,8)),(1,None):(1,None,None),(10,10):(5,5,
            None),(-1,8):(1,3,(1,3)),(0,2):(1,1,None),(3,15):(3,9,(3,9)),(1,3):(1,3,None),(4,8):(3,3,(3,3)),(2,
            8):(1,3,(1,3)),(5,13):(6,8,(6,8)),(7,11):(6,6,(6,6)),(-1,7):(1,3,(1,3)),(8,10):(5,None,(5,None)),(9,
            11):(6,6,(6,6)),(8,16):(6,8,(6,8)),(11,13):(6,8,None),(12,14):(6,8,None),(13,15):(8,9,None),(1,4):(1,
            3,None),(0,12):(1,6,(1,6)),(3,9):(3,3,(3,3)),(2,3):(1,3,None),(1,9):(1,3,(1,3)),(2,14):(1,8,(1,8)),
            (11,None):(6,None,None),(16,None):(10,None,(10,None)),(6,8):(5,None,(5,None)),(5,16):(6,8,(6,8)),(7,
            13):(6,8,(6,8)),(-1,6):(1,3,(1,3)),(9,12):(6,6,(6,6)),(10,11):(5,6,None),(-1,9):(1,3,(1,3)),(12,16):
            (6,8,(6,8)),(0,3):(1,3,None),(3,14):(3,8,(3,8)),(1,2):(1,1,None),(13,None):(8,None,None),(4,9):(3,3,
            (3,3)),(3,3):(3,3,None),(2,9):(1,3,(1,3)),(5,12):(6,6,(6,6)),(4,4):(3,3,None),(7,10):(5,None,(5,
            None)),(8,11):(6,6,(6,6)),(9,10):(5,None,(5,None)),(11,12):(6,6,None),(12,15):(6,9,None),(13,14):(8,
            8,None),(0,13):(1,8,(1,8)),(3,8):(3,3,(3,3)),(1,8):(1,3,(1,3)),(2,15):(1,9,(1,9)),(9,None):(5,None,
            (5,None))},(' a ',' a b '):{(0,1):(1,1,None),(-1,None):(1,None,(1,None)),(-1,1):(1,1,(1,1)),(0,0):(1,
            None,None),(3,None):(4,None,(3,3)),(3,3):(4,None,(3,3)),(1,None):(1,None,None),(1,2):(1,2,None),(2,
            2):(2,2,None),(0,None):(1,None,None),(1,1):(1,1,None),(4,None):(4,None,(3,3)),(2,3):(4,None,(3,3)),
            (2,None):(2,None,None),(-1,-1):(1,None,(1,None)),(1,3):(1,1,(1,3)),(0,2):(1,2,None),(-1,0):(1,None,
            (1,None))},(' a bb c ',' a xy c '):{(1,3):(1,1,(1,4)),(-1,0):(1,None,(1,None)),(4,8):(6,6,(3,6)),(2,
            8):(6,6,(3,6)),(0,7):(1,6,(1,6)),(4,None):(5,None,(3,4)),(1,6):(1,6,(1,6)),(2,None):(2,None,None),
            (-1,5):(1,1,(1,4)),(3,7):(6,6,(3,6)),(2,5):(5,None,(3,4)),(5,8):(5,6,(5,6)),(6,7):(6,6,None),(5,5):
            (5,5,None),(5,None):(5,None,None),(7,None):(7,None,None),(0,4):(1,1,(1,4)),(1,1):(1,1,None),(-1,6):
            (1,6,(1,6)),(2,6):(6,6,(3,6)),(4,5):(5,None,(3,4)),(-1,-1):(1,None,(1,None)),(1,4):(1,1,(1,4)),(2,3):
            (5,None,(3,4)),(6,None):(6,None,None),(0,1):(1,1,None),(6,8):(6,6,(6,6)),(-1,3):(1,1,(1,4)),(0,6):(1,
            6,(1,6)),(1,7):(1,6,(1,6)),(-1,4):(1,1,(1,4)),(7,8):(7,None,(7,None)),(2,4):(5,None,(3,4)),(4,7):(6,
            6,(3,6)),(6,6):(6,6,None),(5,6):(5,6,None),(7,7):(7,None,None),(0,3):(1,1,(1,4)),(1,2):(1,2,None),
            (-1,1):(1,1,(1,1)),(3,3):(5,None,(3,4)),(4,4):(5,None,(3,4)),(1,5):(1,1,(1,4)),(3,6):(6,6,(3,6)),(2,
            2):(2,2,None),(0,0):(1,None,None),(-1,2):(1,2,(1,2)),(0,None):(1,None,None),(0,5):(1,1,(1,4)),(3,5):
            (5,None,(3,4)),(2,7):(6,6,(3,6)),(-1,None):(1,None,(1,None)),(3,None):(5,None,(3,4)),(4,6):(6,6,(3,
            6)),(1,None):(1,None,None),(8,None):(7,None,(7,None)),(3,4):(5,None,(3,4)),(5,7):(5,6,None),(3,8):(6,
            6,(3,6)),(1,8):(1,6,(1,6)),(8,8):(7,None,(7,None)),(0,2):(1,2,None),(9,None):(7,None,(7,None))},
            (' a c ',' a bbb c '):{(1,2):(1,2,None),(1,3):(1,7,(1,7)),(-1,0):(1,None,(1,None)),(4,5):(8,None,(8,
            None)),(-1,2):(1,2,(1,2)),(1,1):(1,1,None),(1,4):(1,7,(1,7)),(0,None):(1,None,None),(4,None):(8,None,
            None),(2,3):(6,7,None),(2,None):(2,None,None),(6,None):(8,None,(8,None)),(0,3):(1,7,(1,7)),(2,5):(6,
            7,(6,7)),(3,5):(7,7,(7,7)),(0,1):(1,1,None),(-1,None):(1,None,(1,None)),(-1,1):(1,1,(1,1)),(3,None):
            (7,None,None),(5,5):(8,None,(8,None)),(1,None):(1,None,None),(-1,3):(1,7,(1,7)),(5,None):(8,None,(8,
            None)),(4,4):(8,None,None),(2,2):(2,2,None),(0,0):(1,None,None),(2,4):(6,7,None),(1,5):(1,7,(1,7)),
            (-1,-1):(1,None,(1,None)),(0,4):(1,7,(1,7)),(3,3):(7,7,None),(3,4):(7,7,None),(0,2):(1,2,None)},
            (' A C ',' A b b C '):{(0,1):(1,1,None),(1,3):(1,7,(1,7)),(-1,0):(1,None,(1,None)),(4,5):(8,None,(8,
            None)),(-1,2):(1,2,(1,2)),(1,4):(1,7,(1,7)),(0,None):(1,None,None),(4,None):(8,None,None),(2,3):(6,7,
            None),(2,None):(2,None,None),(0,2):(1,2,None),(6,None):(8,None,(8,None)),(0,3):(1,7,(1,7)),(2,5):(6,
            7,(6,7)),(3,5):(7,7,(7,7)),(1,2):(1,2,None),(-1,None):(1,None,(1,None)),(-1,1):(1,1,(1,1)),(3,None):
            (7,None,None),(5,5):(8,None,(8,None)),(1,None):(1,None,None),(-1,3):(1,7,(1,7)),(5,None):(8,None,(8,
            None)),(4,4):(8,None,None),(-1,-1):(1,None,(1,None)),(0,0):(1,None,None),(2,4):(6,7,None),(1,5):(1,7,
            (1,7)),(2,2):(2,2,None),(0,4):(1,7,(1,7)),(3,3):(7,7,None),(3,4):(7,7,None),(1,1):(1,1,None)},(' b ',
            ' a b '):{(0,1):(2,3,None),(-1,None):(2,None,(1,1)),(-1,1):(3,3,(1,3)),(0,0):(2,2,None),(3,None):(4,
            None,(4,None)),(3,3):(4,None,(4,None)),(1,None):(3,None,None),(1,2):(3,3,None),(2,2):(4,None,None),
            (0,None):(2,None,None),(1,1):(3,3,None),(4,None):(4,None,(4,None)),(2,3):(4,None,(4,None)),(2,None):
            (4,None,None),(-1,-1):(2,None,(1,1)),(1,3):(3,3,(3,3)),(0,2):(2,3,None),(-1,0):(2,None,(1,1))},
            (' abc ',' abc '):{(1,2):(1,2,None),(1,3):(1,3,None),(-1,0):(1,None,(1,None)),(4,5):(4,None,(4,
            None)),(-1,2):(1,2,(1,2)),(1,1):(1,1,None),(1,4):(1,3,None),(0,None):(1,None,None),(4,None):(4,None,
            None),(2,3):(2,3,None),(2,None):(2,None,None),(6,None):(4,None,(4,None)),(0,3):(1,3,None),(2,5):(2,3,
            (2,3)),(3,5):(3,3,(3,3)),(0,1):(1,1,None),(-1,None):(1,None,(1,None)),(-1,1):(1,1,(1,1)),(3,None):(3,
            None,None),(5,5):(4,None,(4,None)),(1,None):(1,None,None),(-1,3):(1,3,(1,3)),(5,None):(4,None,(4,
            None)),(4,4):(4,None,None),(2,2):(2,2,None),(0,0):(1,None,None),(2,4):(2,3,None),(1,5):(1,3,(1,3)),
            (-1,-1):(1,None,(1,None)),(0,4):(1,3,None),(3,3):(3,3,None),(3,4):(3,3,None),(0,2):(1,2,None)},
            (' a b ',' b '):{(1,2):(1,None,(1,None)),(1,3):(1,1,(1,1)),(-1,0):(1,None,(1,None)),(4,5):(2,None,(2,
            None)),(-1,2):(1,None,(1,None)),(1,1):(1,None,(1,None)),(1,4):(1,1,(1,1)),(0,None):(1,None,(1,None)),
            (4,None):(2,None,None),(2,3):(1,1,None),(2,None):(1,None,None),(6,None):(2,None,(2,None)),(0,3):(1,1,
            (1,1)),(2,5):(1,1,(1,1)),(3,5):(1,1,(1,1)),(0,1):(1,None,(1,None)),(-1,None):(1,None,(1,None)),(-1,
            1):(1,None,(1,None)),(3,None):(1,None,None),(5,5):(2,None,(2,None)),(1,None):(1,None,(1,None)),(-1,
            3):(1,1,(1,1)),(5,None):(2,None,(2,None)),(4,4):(2,None,None),(2,2):(1,None,None),(0,0):(1,None,(1,
            None)),(2,4):(1,1,None),(1,5):(1,1,(1,1)),(-1,-1):(1,None,(1,None)),(0,4):(1,1,(1,1)),(3,3):(1,1,
            None),(3,4):(1,1,None),(0,2):(1,None,(1,None))},(' foo * ',' foo* '):{(4,7):(5,None,(1,4)),(1,3):(1,
            3,None),(-1,0):(5,None,(1,4)),(6,6):(5,None,(1,4)),(7,None):(5,None,(1,4)),(7,7):(5,None,(1,4)),(4,
            None):(5,None,(1,4)),(1,6):(1,4,(1,4)),(2,None):(5,None,(1,4)),(-1,2):(5,None,(1,4)),(3,7):(5,None,
            (1,4)),(0,3):(1,3,(1,4)),(2,5):(2,4,(1,4)),(1,2):(1,2,None),(-1,1):(5,None,(1,4)),(6,7):(5,None,(1,
            4)),(3,3):(5,None,(1,4)),(1,5):(1,4,(1,4)),(5,None):(5,None,(1,4)),(4,4):(5,None,(1,4)),(5,6):(4,4,
            (1,4)),(3,6):(3,4,(1,4)),(2,2):(5,None,(1,4)),(0,4):(1,3,(1,4)),(3,5):(3,4,(1,4)),(1,1):(1,1,None),
            (2,7):(5,None,(1,4)),(2,6):(2,4,(1,4)),(0,0):(5,None,(1,4)),(4,5):(4,4,(1,4)),(-1,-1):(5,None,(1,4)),
            (1,4):(1,3,(1,4)),(0,None):(5,None,(1,4)),(2,3):(2,3,None),(6,None):(5,None,(1,4)),(5,5):(4,4,None),
            (0,1):(1,1,(1,4)),(-1,None):(5,None,(1,4)),(-1,5):(5,None,(1,4)),(3,None):(5,None,(1,4)),(4,6):(4,4,
            (1,4)),(1,None):(5,None,(1,4)),(8,None):(5,None,(1,4)),(-1,3):(5,None,(1,4)),(5,7):(5,None,(1,4)),(0,
            5):(1,4,(1,4)),(2,4):(2,3,(1,4)),(0,6):(1,4,(1,4)),(1,7):(5,None,(1,4)),(-1,4):(5,None,(1,4)),(3,4):
            (5,None,(1,4)),(0,2):(1,2,(1,4))},(' \xc3\xa4 \xc3\xb6\xc3\xb6 \xc3\xbc ',
            ' \xc3\xa4 \xc3\x9f\xc3\x9f \xc3\xbc '):{(6,9):(9,9,(4,9)),(11,11):(11,None,None),(7,12):(9,9,(4,9)),
            (12,12):(11,None,(11,None)),(0,7):(1,1,(1,6)),(4,None):(8,None,(4,6)),(1,6):(1,1,(1,6)),(0,10):(1,9,
            (1,9)),(3,7):(8,None,(4,6)),(2,5):(1,1,(1,6)),(1,11):(1,10,(1,10)),(5,8):(8,None,(4,6)),(-1,0):(1,
            None,(1,None)),(6,7):(8,None,(4,6)),(5,5):(8,None,(4,6)),(5,None):(8,None,(4,6)),(6,10):(9,9,(4,9)),
            (10,None):(9,None,None),(0,4):(1,1,(1,6)),(1,1):(1,1,None),(4,10):(9,9,(4,9)),(2,6):(1,1,(1,6)),(5,
            11):(9,10,(4,10)),(4,5):(8,None,(4,6)),(-1,-1):(1,None,(1,None)),(6,None):(8,None,(4,6)),(0,1):(1,1,
            None),(3,12):(9,9,(4,9)),(1,12):(1,9,(1,9)),(8,12):(8,9,(8,9)),(2,11):(1,10,(1,10)),(9,9):(9,9,None),
            (-1,4):(1,1,(1,6)),(7,8):(8,None,(4,6)),(3,11):(9,10,(4,10)),(8,9):(8,9,None),(4,12):(9,9,(4,9)),(2,
            12):(1,9,(1,9)),(-1,1):(1,1,(1,1)),(1,5):(1,1,(1,6)),(0,11):(1,10,(1,10)),(3,6):(8,None,(4,6)),(2,2):
            (1,1,None),(1,10):(1,9,(1,9)),(6,11):(9,10,(4,10)),(-1,2):(1,1,(1,1)),(0,None):(1,None,None),(0,5):
            (1,1,(1,6)),(0,8):(1,1,(1,6)),(4,11):(9,10,(4,10)),(3,5):(8,None,(4,6)),(2,7):(1,1,(1,6)),(-1,None):
            (1,None,(1,None)),(5,10):(9,9,(4,9)),(4,6):(8,None,(4,6)),(1,None):(1,None,None),(10,10):(9,9,None),
            (5,7):(8,None,(4,6)),(-1,8):(1,1,(1,6)),(0,2):(1,1,None),(1,3):(1,3,None),(4,8):(8,None,(4,6)),(2,8):
            (1,1,(1,6)),(2,None):(1,None,None),(-1,5):(1,1,(1,6)),(7,11):(9,10,(4,10)),(-1,7):(1,1,(1,6)),(3,10):
            (9,9,(4,9)),(8,10):(8,9,None),(5,6):(8,None,(4,6)),(9,11):(9,10,None),(-1,6):(1,1,(1,6)),(1,4):(1,1,
            (1,6)),(3,9):(9,9,(4,9)),(2,3):(1,3,None),(1,9):(1,9,(1,9)),(11,None):(11,None,None),(6,8):(8,None,
            (4,6)),(-1,3):(1,3,(1,3)),(0,6):(1,1,(1,6)),(1,7):(1,1,(1,6)),(0,9):(1,9,(1,9)),(3,4):(8,None,(4,6)),
            (2,4):(1,1,(1,6)),(9,12):(9,9,(9,9)),(5,9):(9,9,(4,9)),(4,7):(8,None,(4,6)),(10,11):(9,10,None),(6,
            6):(8,None,(4,6)),(7,None):(8,None,(4,6)),(-1,9):(1,9,(1,9)),(12,None):(11,None,(11,None)),(7,7):(8,
            None,(4,6)),(0,3):(1,3,None),(1,2):(1,1,None),(13,None):(11,None,(11,None)),(4,9):(9,9,(4,9)),(3,3):
            (3,3,None),(2,9):(1,9,(1,9)),(5,12):(9,9,(4,9)),(4,4):(8,None,(4,6)),(10,12):(9,9,(9,9)),(-1,10):(1,
            9,(1,9)),(7,10):(9,9,(4,9)),(0,0):(1,None,None),(8,11):(8,10,None),(2,10):(1,9,(1,9)),(9,10):(9,9,
            None),(6,12):(9,9,(4,9)),(11,12):(11,None,(11,None)),(7,9):(9,9,(4,9)),(3,None):(3,None,None),(8,
            None):(8,None,None),(3,8):(8,None,(4,6)),(1,8):(1,1,(1,6)),(8,8):(8,8,None),(9,None):(9,None,None)},
            (' foo* ',' foo * '):{(1,3):(1,3,None),(-1,0):(6,None,(1,5)),(6,6):(6,None,(1,5)),(7,None):(6,None,
            (1,5)),(4,None):(6,None,(1,5)),(1,6):(6,None,(1,5)),(2,None):(6,None,(1,5)),(-1,2):(6,None,(1,5)),(0,
            3):(1,3,(1,5)),(2,5):(2,5,(1,5)),(1,2):(1,2,None),(-1,1):(6,None,(1,5)),(3,3):(6,None,(1,5)),(1,5):
            (1,5,(1,5)),(5,None):(6,None,(1,5)),(4,4):(5,5,None),(5,6):(6,None,(1,5)),(3,6):(6,None,(1,5)),(2,2):
            (6,None,(1,5)),(0,4):(1,5,(1,5)),(3,5):(3,5,(1,5)),(1,1):(1,1,None),(2,6):(6,None,(1,5)),(0,0):(6,
            None,(1,5)),(4,5):(5,5,(1,5)),(-1,-1):(6,None,(1,5)),(1,4):(1,5,(1,5)),(0,None):(6,None,(1,5)),(2,3):
            (2,3,None),(6,None):(6,None,(1,5)),(5,5):(4,4,None),(0,1):(1,1,(1,5)),(-1,None):(6,None,(1,5)),(3,
            None):(6,None,(1,5)),(4,6):(6,None,(1,5)),(1,None):(6,None,(1,5)),(-1,3):(6,None,(1,5)),(0,5):(1,5,
            (1,5)),(2,4):(2,5,(1,5)),(-1,4):(6,None,(1,5)),(3,4):(3,5,(1,5)),(0,2):(1,2,(1,5))},(' XX YY ZZ ',
            ' XX ZZ '):{(6,9):(3,5,None),(-1,0):(1,None,(1,None)),(0,7):(1,4,(1,4)),(4,None):(3,None,(3,None)),
            (1,6):(1,2,(1,2)),(3,7):(4,4,(4,4)),(2,5):(2,2,(2,2)),(5,8):(4,5,(4,5)),(6,7):(3,4,None),(5,5):(3,
            None,(3,None)),(5,None):(3,None,(3,None)),(6,10):(3,5,(3,5)),(10,None):(6,None,(6,None)),(0,4):(1,2,
            (1,2)),(1,1):(1,1,None),(4,10):(4,5,(4,5)),(2,6):(2,2,(2,2)),(4,5):(3,None,(3,None)),(-1,-1):(1,None,
            (1,None)),(6,None):(3,None,None),(0,1):(1,1,None),(9,9):(6,None,None),(-1,4):(1,2,(1,2)),(7,8):(4,5,
            None),(8,9):(5,5,None),(-1,1):(1,1,(1,1)),(1,5):(1,2,(1,2)),(3,6):(3,None,(3,None)),(2,2):(2,2,None),
            (1,10):(1,5,(1,5)),(-1,2):(1,2,(1,2)),(0,None):(1,None,None),(0,5):(1,2,(1,2)),(0,8):(1,5,(1,5)),(3,
            5):(3,None,(3,None)),(2,7):(2,4,(2,4)),(-1,None):(1,None,(1,None)),(5,10):(4,5,(4,5)),(4,6):(3,None,
            (3,None)),(1,None):(1,None,None),(10,10):(6,None,(6,None)),(5,7):(4,4,(4,4)),(-1,8):(1,5,(1,5)),(0,
            2):(1,2,None),(1,3):(1,3,None),(4,8):(4,5,(4,5)),(2,8):(2,5,(2,5)),(2,None):(2,None,None),(-1,5):(1,
            2,(1,2)),(3,10):(4,5,(4,5)),(8,10):(5,5,(5,5)),(5,6):(3,None,(3,None)),(-1,6):(1,2,(1,2)),(1,4):(1,2,
            (1,2)),(3,9):(4,5,(4,5)),(2,3):(2,3,None),(1,9):(1,5,(1,5)),(11,None):(6,None,(6,None)),(6,8):(3,5,
            None),(-1,3):(1,3,(1,3)),(0,6):(1,2,(1,2)),(1,7):(1,4,(1,4)),(0,9):(1,5,(1,5)),(3,4):(3,None,(3,
            None)),(2,4):(2,2,(2,2)),(5,9):(4,5,(4,5)),(4,7):(4,4,(4,4)),(6,6):(3,3,None),(7,None):(4,None,None),
            (7,7):(4,4,None),(0,3):(1,3,None),(1,2):(1,2,None),(4,9):(4,5,(4,5)),(3,3):(3,3,None),(2,9):(2,5,(2,
            5)),(4,4):(3,None,(3,None)),(7,10):(4,5,(4,5)),(0,0):(1,None,None),(2,10):(2,5,(2,5)),(9,10):(6,None,
            (6,None)),(-1,7):(1,4,(1,4)),(7,9):(4,5,None),(3,None):(3,None,None),(8,None):(5,None,None),(3,8):(4,
            5,(4,5)),(1,8):(1,5,(1,5)),(8,8):(5,5,None),(9,None):(6,None,None)},('  ','  '):{(0,1):(1,None,None),
            (0,0):(1,None,None),(-1,0):(1,None,(1,None)),(2,2):(1,None,(1,None)),(0,None):(1,None,None),(2,None):
            (1,None,(1,None)),(1,2):(1,None,(1,None)),(-1,None):(1,None,(1,None)),(3,None):(1,None,(1,None)),(1,
            None):(1,None,None),(-1,-1):(1,None,(1,None)),(1,1):(1,None,None)},(' a b ',' a '):{(1,2):(1,1,None),
            (1,3):(1,1,(1,1)),(-1,0):(1,None,(1,None)),(4,5):(2,None,(2,None)),(-1,2):(1,1,(1,1)),(1,1):(1,1,
            None),(1,4):(1,1,(1,1)),(0,None):(1,None,None),(4,None):(2,None,(2,None)),(2,3):(2,None,(2,None)),(2,
            None):(2,None,None),(6,None):(2,None,(2,None)),(0,3):(1,1,(1,1)),(2,5):(2,None,(2,None)),(3,5):(2,
            None,(2,None)),(0,1):(1,1,None),(-1,None):(1,None,(1,None)),(-1,1):(1,1,(1,1)),(3,None):(2,None,(2,
            None)),(5,5):(2,None,(2,None)),(1,None):(1,None,None),(-1,3):(1,1,(1,1)),(5,None):(2,None,(2,None)),
            (4,4):(2,None,(2,None)),(2,2):(2,None,None),(0,0):(1,None,None),(2,4):(2,None,(2,None)),(1,5):(1,1,
            (1,1)),(-1,-1):(1,None,(1,None)),(0,4):(1,1,(1,1)),(3,3):(2,None,(2,None)),(3,4):(2,None,(2,None)),
            (0,2):(1,1,None)}}
            
        results = {}
        for text1, text2 in (
            ("  ", # empty text
             "  "),
             #012345678901234567890
             #0         1         2
            (" abc ",
             " abc "),
             #012345678901234567890
             #0         1         2
            (" a b ",
             " a "),
             #012345678901234567890
             #0         1         2
            (" a b ",
             " b "),
             #012345678901234567890
             #0         1         2
            (" a ",
             " a b "),
             #012345678901234567890
             #0         1         2
            (" b ",
             " a b "),
             #012345678901234567890
             #0         1         2
            (" a bb c ",
             " a xy c "),
             #012345678901234567890
             #0         1         2
            (" ä öö ü ",
             " ä ßß ü "),
             #012345678901234567890
             #0         1         2
            (" ÄÄ ÖÖ ÜÜ ",
             " ÄÄ ÜÜ "),
             #012345678901234567890
             #0         1         2
            (" a c ",
             " a bbb c "),
             #012345678901234567890
             #0         1         2
            (" A C ",
             " A b b C "),
             #012345678901234567890
             #0         1         2
            (" XX YY ZZ ",
             " XX ZZ "),
             #012345678901234567890
             #0         1         2
            (" foo* ",
             " foo * "),
             #012345678901234567890
             #0         1         2
            (" foo * ",
             " foo* "),
             #012345678901234567890
             #0         1         2
        ):
            resDict = {}
            results[(text1,text2)] = resDict
            
            matcher = TextMatcher(text1, text2)
            if doPrint:
                print "text1 = '%s'" % text1.decode('utf-8')
                print "text2 = '%s'" % text2.decode('utf-8')
                print "matches =", matcher.get_matching_blocks()
            mapper = MatchMapper(matcher)
            for len_ in xrange(len(text1)+1):
                for begin in xrange(-1, len(text1)-len_+2):
                    end = begin+len_-1
                    if end < begin:
                        end = None
                    match = mapper.mapBeginEnd(begin,end)#(2, None)#
                    resDict[(begin, end)] = match
                    
                    try:
                        expectedMatch = expectedResults[(text1, text2)][(begin, end)]
                        if expectedMatch == match:
                            continue
                    except KeyError:
                        expectedMatch = ""
                    
                    print " map:", (begin, end), "=>", match, "expected: %s" % (expectedMatch,) if expectedMatch else ''

                    if text1 == text2 and begin >= 1 and begin+len_-1 < len(text1)-1:
                        assert match == (begin, end, None)
                        continue
                        
                    print "     '%s'" % applyMatch((begin, end, None), text1, True)
                    print "     '%s'" % applyMatch(match, text2)
                    assert match[1] is None or match[0] <= match[1]
                    if match[2]:
                        assert match[2][1] is None or match[2][0] <= match[2][1]
                        assert match[2][0] <= match[0]
                        if match[1] is None:
                            assert match[2][1] is None or utf8.adjustL(text2, match[0]-1) <= match[2][1]
                        else:
                            assert match[2][1] is None or match[1] <= match[2][1]
                    #assert not expectedMatch
                    if expectedMatch:
                        print " was:'%s'" % applyMatch(expectedMatch, text2)
                        print "ERROR: ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ result changed\n"
        if doPrint:
            print ""
            print "Take over following results as expectedResults if differences are ok"
            #print "results =", results
            for key, dic in results.iteritems():
                if not expectedResults.has_key(key):
                    expectedResults[key] = dic
                else:
                    for aRange, bRange in dic.iteritems():
                        expectedResults[key][aRange] = bRange
            import re
            print re.sub(r"(.{,100}[,:}])", r"\1\n"+' '*12, "results = " + re.sub("([:,]) ", r"\1", str(expectedResults)))

        printEndTest()
        
    def generateLargeTestText():
#        with open(r"C:\tmp\mobi.txt", "rb") as file:
#            text = file.read().replace("\n", " ")
#        return text
        
        print "Generating large test text..."
        import random
        
        def generateHistogram(text):
            histo = {}
            for c in text:
                key = " " if c.isspace() else c.encode('utf-8')
                if histo.has_key(key):
                    histo[key] += 1
                else:
                    histo[key] = 1
            # add/modify some interesting characters for unicode test
            extra = { 'a': u'äáàæ', 'e': u'ëéè', 'i': u'ïíìî', 'o': u'öóòôøõō', 'u': u'üúùûụů' }
            total = sum(histo.values())
            for orig, mods in extra.items():
                if histo.has_key(orig):
                    count = max(1, int(histo[orig] / (2*len(mods))))
                    histo[orig] -= len(mods)*count
                    if histo[orig] <= 0:
                        histo[orig] = count
                else:
                    count = total/30
                for mod in mods:
                    histo[mod.encode('utf-8')] = count
            return sorted(histo.items(), key=lambda i: i[1], reverse=True)

        def generateWordHistogram(text):
            import re
            histo = {}
            ofInterest = set()
            word = []
            assert u"é".isalnum()
            assert re.search("[\x80-\xff]", u"café")
            for c in text:
                if c.isalnum():
                    word.append(c)
                    continue
                if not word:
                    continue
                key = ''.join(word).encode('utf-8')
                word = []
                if re.search("[\x80-\xff]", key):
                    ofInterest.add(key)
                if histo.has_key(key):
                    histo[key] += 1
                else:
                    histo[key] = 1
            print "ofInterest =", ofInterest
            return sorted(histo.items(), key=lambda i: i[1], reverse=True)

        #with open(r"C:\tmp\mobi.txt", "rb") as utf8File:
        #    histogram = generateHistogram(utf8File.read().decode('utf-8'))
        #print ',\n('.join(re.findall(r"(.{,100}\))(?:, \(|\]$)", str(histogram))).replace("[","histogram = [\n",1).replace('\n', '\n'+' '*12)+"]"
        #return
        histogram = [
            (' ', 106167), ('t', 42262), ('n', 29539), ('e', 28093), ('s', 27357), ('r', 26499), ('h', 25559),
            ('d', 21360), ('l', 18928), ('a', 18729), ('o', 17065), ('i', 14371), ('c', 11885), ('g', 10663),
            ('m', 10637), ('w', 9888), ('y', 9486), ('\xc3\xa8', 9364), ('\xc3\xab', 9364), ('\xc3\xa9', 9364),
            ('f', 9004), ('p', 8351), ('"', 7801), ('.', 7500), ('u', 7047), (',', 6828), ('b', 5897), ('k', 4906),
            ('\xc3\xa0', 4681), ('\xc3\xa4', 4681), ('\xc3\xa1', 4681), ('\xc3\xa6', 4681), ('I', 4095),
            ('v', 3798), ('\xc3\xac', 3591), ('\xc3\xaf', 3591), ('\xc3\xae', 3591), ('\xc3\xad', 3591),
            ("'", 2779), ('\xc3\xb4', 2436), ('\xc3\xb8', 2436), ('\xc3\xb3', 2436), ('\xc3\xb2', 2436),
            ('\xc3\xb6', 2436), ('\xc3\xb5', 2436), ('\xc5\x8d', 2436), ('T', 1502), ('\xe1\xbb\xa5', 1174),
            ('\xc3\xbb', 1174), ('\xc3\xba', 1174), ('\xc3\xb9', 1174), ('\xc3\xbc', 1174), ('\xc5\xaf', 1174),
            ('B', 1053), ('?', 984), ('A', 979), ('S', 880), ('H', 875), ('M', 866), ('-', 740), ('W', 735),
            ('x', 690), ('j', 493), ('Y', 478), ('F', 407), ('C', 387), ('N', 383), ('q', 358), ('z', 345),
            ('O', 303), ('J', 274), ('E', 269), ('L', 262), ('P', 238), ('R', 224), ('Q', 214), ('D', 170),
            ('K', 166), ('G', 144), ('U', 101), (';', 96), (':', 85), ('V', 51), ('*', 39), ('/', 33), ('!', 30),
            ('1', 22), ('6', 19), ('^', 18), ('5', 18), ('3', 16), ('0', 15), ('4', 12), ('2', 10), ('9', 10),
            ('7', 8), ('|', 7), ('\xc2\xbb', 6), ('\xc2\xa6', 5), ('\xc2\xab', 5), ('\xe2\x80\xa2', 5), ('8', 4),
            ('<', 4), ('Z', 4), ('~', 4), ('X', 2), ('\\', 2), ('#', 2), ('[', 2), ('>', 2), ('%', 2), ('_', 1),
            ('{', 1), ('&', 1), ('\xc2\xa7', 1), ('\xc2\xae', 1), ('\xc2\xa9', 1)]

        def generatePseudoRandomText(seed, length, histogram):
            total = reduce(lambda sum, item: sum + item[1], histogram, 0)
            rand = random.Random(seed)
            chars = []
            while len(chars) < length:
                randPos = rand.randint(0,total-1)
                sum = 0
                for char, count in histogram:
                    sum += count
                    if randPos < sum:
                        break
                chars.append(char)
            return ''.join(chars)
        
        text = " %s " % generatePseudoRandomText(4711, 500000, histogram)

        #print "text1 ='%s'" % text1.encode('string_escape')
        print "... finished generating large test text."
        return text
        
    def makeTwoTextsByRandomInsert(text, nrInserts):
        import random
        inserts = {}
        rand = random.Random(987654)
        for _ in xrange(nrInserts):
            inserts[ rand.randint(1, len(text))-2 ] = ''.join(rand.choice("UVWXYZ") for i in range(rand.randint(1,4)))
        inserts = sorted(inserts.iteritems())
        def applyInserts(text1, inserts):
            text2 = []
            cur = 0
            for pos, txt in inserts:
                text2.append(text1[cur:pos])
                text2.append(txt)
                cur = pos
            text2.append(text1[cur:])
            off = 0
            for i in xrange(len(inserts)):
                inserts[i] = (inserts[i][0] + off, inserts[i][1])            
                off += len(inserts[i][1])
            return ''.join(text2)
        text1 = applyInserts(text, inserts[::2])
        text2 = applyInserts(text, inserts[1::2])
        #text1 = text
        #text2 = applyInserts(text, inserts)

        #with open(r"C:\tmp\text1.txt", "wb") as file:
        #    file.write(text1.replace(" ", "\n"))
        #    
        #with open(r"C:\tmp\text2.txt", "wb") as file:
        #    file.write(text2.replace(" ", "\n"))
        
        return text1, text2

    def saveTextDiff(matches, text1, text2, filename):
        with open(filename, "wb") as file:
            aCur = 0
            bCur = 0
            for m in matches:
                file.write("< "+text1[aCur:m.a]+"\n")
                file.write("> "+text2[bCur:m.b]+"\n")
                file.write("= "+text1[m.a:m.a+m.size]+"\n")
                aCur = m.a+m.size
                bCur = m.b+m.size
            if matches:
                m = matches[-1]
                file.write("< "+text1[m.a:]+"\n")
                file.write("> "+text2[m.b:]+"\n")

    def test_SeedMatcher_large(text, seedLen, maxTries, minMatchLen = 0):
        import time
        printStartTest(seedLen, maxTries, minMatchLen)

        perfs = []
        for nrInserts in (0, 1, 2, 3, 4, 6, 8, 10, 15, 20, 25, 50, 100, 500, 1000, 5000, 10000, 50000):
            if doPrint:
                print "nrInserts =", nrInserts
            else:
                import sys
                sys.stdout.write('.')

            text1, text2 = makeTwoTextsByRandomInsert(text, nrInserts)
    
            # now let's see how SeedMatcher does
            if doPrint:
                print "BEGIN matching..."
            tstart = time.clock()
            matcher = SeedMatcher(seedLen, maxTries, minMatchLen)
            matches = matcher.seedMatches(text1, text2)
            tstop = time.clock()
            if doPrint:
                print "END matching!"
                print matches
            cachestatus = ""
            try:
                cache = matcher._cachedMatches
                count = reduce(lambda sum, list: sum + len(list), cache.itervalues(), 0)
                total = reduce(lambda sum, list: sum + reduce(lambda sum, item: sum + item[-1], list, 0), cache.itervalues(), 0)
                cachestatus = str((len(cache), count, total/count if count>0 else 0))
            except:
                raise
                
            ratio = _calculateRatio(matches,len(text1),len(text2))
            
            TextMatcher(text1, text2)._seedMatchesPostprocess(matches)
            ratio2 = _calculateRatio(matches,len(text1),len(text2))

            perfs.append((nrInserts, tstop-tstart, ratio, ratio2, cachestatus))
                
        if not doPrint:
            print ""
        for perf in perfs:
            print " % 6d %.6f %f > %f %s" % perf
        
        print "END"
        
    def test_TextMatcher_large(text):
        import time
        printStartTest()

        perfs = []
        for nrInserts in (0, 1, 2, 3, 4, 6, 8, 10, 15, 20, 25, 50, 100, 500, 1000, 5000, 10000, 50000):
            if doPrint:
                print "nrInserts =", nrInserts
            else:
                import sys
                sys.stdout.write('.')

            text1, text2 = makeTwoTextsByRandomInsert(text, nrInserts)
    
            # now let's see how TextMatcher does
            if doPrint:
                print "BEGIN matching..."
            tstart = time.clock()
            matcher = TextMatcher(text1, text2)
            matches = matcher.get_matching_blocks()
            tstop = time.clock()
            if doPrint:
                print "END matching!"
                print matches
                
            ratio = _calculateRatio(matches,len(text1),len(text2))

            #saveTextDiff(matches, text1, text2, r"C:\tmp\diff%07d.txt" % nrInserts)

            perfs.append((nrInserts, tstop-tstart, ratio))
                
        if not doPrint:
            print ""
        for perf in perfs:
            print " % 6d %.6f %f" % perf
        
        print "END"

if __debug__ and __name__ == '__main__':
    # run tests
    test_findUniqueIgnoreSpace()
    test_seedMatch()
    test_seedMatches()
    test_seedMatchesPostprocess()
    test_getDifflibWordMatches()
    test_insertDifflibWordMatches()
    test_MatchMapper()
    
    # performance tests for manual evaluation
    #text = generateLargeTestText()
    #test_SeedMatcher_large(text, 50, 31, 1000)
    #test_SeedMatcher_large(text, 30, 41)
    #test_SeedMatcher_large(text, 20, 51)
    #test_TextMatcher_large(text)

