# -*- coding: utf-8 -*-
assert '…' == '\xe2\x80\xa6', "file encoding error"

# Copyright 2013 Axel Walthelm
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

from __builtin__ import ord as _ord
import re

# what to recognize as a space (sequences of which will be turned into a single ' ' in text)
# basically we use anything unicode regular expressions considers a whitespace
WHITESPACE = frozenset([
    '\t', # u'\t' Cc CHARACTER TABULATION
    '\n', # u'\n' Cc LINE FEED (LF)
    '\x0b', # u'\x0b' Cc LINE TABULATION
    '\x0c', # u'\x0c' Cc FORM FEED (FF)
    '\r', # u'\r' Cc CARRIAGE RETURN (CR)
    '\x1c', # u'\x1c' Cc 
    '\x1d', # u'\x1d' Cc 
    '\x1e', # u'\x1e' Cc 
    '\x1f', # u'\x1f' Cc 
    ' ', # u' ' Zs SPACE
    '\xc2\x85', # u'\x85' Cc NEXT LINE
    '\xc2\xa0', # u'\xa0' Zs NO-BREAK SPACE
    # '\xe1\x9a\x80', # u'\u1680' Zs OGHAM SPACE MARK # usually not really a space but a dash
    '\xe2\x80\x80', # u'\u2000' Zs EN QUAD
    '\xe2\x80\x81', # u'\u2001' Zs EM QUAD
    '\xe2\x80\x82', # u'\u2002' Zs EN SPACE
    '\xe2\x80\x83', # u'\u2003' Zs EM SPACE
    '\xe2\x80\x84', # u'\u2004' Zs THREE-PER-EM SPACE
    '\xe2\x80\x85', # u'\u2005' Zs FOUR-PER-EM SPACE
    '\xe2\x80\x86', # u'\u2006' Zs SIX-PER-EM SPACE
    '\xe2\x80\x87', # u'\u2007' Zs FIGURE SPACE
    '\xe2\x80\x88', # u'\u2008' Zs PUNCTUATION SPACE
    '\xe2\x80\x89', # u'\u2009' Zs THIN SPACE
    '\xe2\x80\x8a', # u'\u200a' Zs HAIR SPACE
    '\xe2\x80\xa8', # u'\u2028' Zl LINE SEPARATOR
    '\xe2\x80\xa9', # u'\u2029' Zp PARAGRAPH SEPARATOR
    '\xe2\x80\xaf', # u'\u202f' Zs NARROW NO-BREAK SPACE
    '\xe2\x81\x9f', # u'\u205f' Zs MEDIUM MATHEMATICAL SPACE
    '\xe3\x80\x80', # u'\u3000' Zs IDEOGRAPHIC SPACE
    ])

RE_WHITESPACES = re.compile(br"([\t\n\x0b\x0c\r\x1c-\x1f ]|\xc2[\x85\xa0]|\xe2\x80[\x80-\x8a\xa8\xa9\xaf]|\xe2\x81\x9f|\xe3\x80\x80)+")
    
# what to recognize as a zero width space (to be removed from text)
ZERO_WIDTH_SPACE = frozenset([
    '\xc2\xad', # u'\xad' Soft Hyphen # html: &shy;
    '\xe1\xa0\x8e', # u'\u180e' Zs MONGOLIAN VOWEL SEPARATOR
    '\xe2\x80\x8b', # u'\u200b' Zero Width Space (ZWSP)
    '\xe2\x80\x8c', # u'\u200c' Zero Width Non Joiner (ZWNJ)
    '\xe2\x80\x8d', # u'\u200d' Zero Width Joiner (ZWJ)
    '\xe2\x81\xa0', # u'\u2060' Word Joiner (WJ)
    '\xef\xbb\xbf', # u'\ufeff' ZERO WIDTH NO-BREAK SPACE # mostly used as BOM
    ])
    
RE_ZERO_WIDTH_SPACE = re.compile(br"\xc2\xad|\xe1\xa0\x8e|\xe2\x80[\x8b-\x8d]|\xe2\x81\xa0|\xef\xbb\xbf")

# return adjusted value loc so that it points to a valid utf-8 character start byte;
# if necessary move backward to the left to find this characters start location
def adjustL( data, loc ):
    while 0x80 <= _ord(data[loc]) < 0xc0: # while is continuation byte
        loc -= 1
    return loc

# like adjustL, but also returns the correction done as second return argument
def adjustL2( data, loc ):
    adjusted = 0
    while 0x80 <= _ord(data[loc]) < 0xc0: # while is continuation byte
        loc -= 1
        adjusted += 1
    return loc, adjusted
    
# return adjusted value loc so that it points to a valid utf-8 character start byte;
# if necessary move forward to the right to find next characters start location
def adjustR( data, loc ):
    while loc < len(data) and 0x80 <= _ord(data[loc]) < 0xc0: # while is continuation byte
        loc += 1
    return loc

# like adjustR, but also returns the correction done as second return argument
def adjustR2( data, loc ):
    adjusted = 0
    while loc < len(data) and 0x80 <= _ord(data[loc]) < 0xc0: # while is continuation byte
        loc += 1
        adjusted += 1
    return loc, adjusted

# check if given position loc references a utf-8 continuation byte in data
def isCont( data, loc ):
    return 0x80 <= _ord(data[loc]) < 0xc0
    
# like builtin unichar(), but returns the utf-8 byte string
def chr( i ):
    try:
        return unichr(i).encode('utf-8')
    except ValueError: # unichr() arg not in range(0x10000) (narrow Python build)
        # workaround using long unicode constant \U and UTF-16 surrogate pair
        return (r'\U%0.8x' % i).decode('unicode-escape').encode('utf8')
        
# like builtin ord(), but expects a utf-8 byte string as input
def ord( c ):
    try:
        return _ord(c.decode('utf-8'))
    except TypeError: # ord() expected a character, but string of length 2 found
        # workaround using long unicode constant \U for UTF-16 surrogate pair
        return int( c.decode('utf-8').encode('unicode-escape')[2:], 16 )
        
# return text as utf-8 bytestring; convert it if it is unicode
def str(text):
    return text.encode('utf-8') if isinstance(text, unicode) else text

# return the given range of utf-8 text in a guaranteed to be printable form
def forPrint(text, i = 0, j = None):
    if not text:
        return text
    try:
        i = adjustL(text, i)
        j = adjustR(text, j if j is not None else len(text))
        return text[i:j].decode('utf-8', 'ignore').encode('unicode_escape')
    except:
        assert isinstance(text, bytes)
        return "\uERROR{%d:%d|%s|%s|%s|%d}" % (i, j,
            text[max(0,i-4):i].encode('string_escape'),
            text[i:j].encode('string_escape'),
            text[j:min(len(text),j+4)].encode('string_escape'),
            len(text))
    
# return the given character of utf-8 text in a guaranteed to be printable form
def charForPrint(text, i):
    return forPrint(text, i, i+1)

if __debug__ and __name__ == '__main__':
    def cpForPrint(cp):
        assert isinstance(cp, int)
        try:
            return "%d 0x%x '%s' \"%s\"" % (cp, cp, chr(cp).encode('string_escape'), forPrint(chr(cp)))
        except:
            return "%d 0x%x" % (cp, cp)
            
    def _runTests():
        print "Test ord vs. chr"
        errorCount = 0
        import sys
        for cp in range(0x10ffff+1):
            if cp % 10000 == 0:
                sys.stdout.write('.')
            if ord(chr(cp)) != cp:
                print
                print "ord() or chr() failed for code point "+cpForPrint(cp)
                errorCount += 1
        print
        print "OK" if not errorCount else "FAILED %d" % errorCount
        assert not errorCount
    
        print "Test RE_ZERO_WIDTH_SPACE vs. ZERO_WIDTH_SPACE"
        import sys
        errorCount = 0
        for cp in range(0x10ffff+1):
            if cp % 10000 == 0:
                sys.stdout.write('.')
            c = chr(cp)
            match = re.match(RE_ZERO_WIDTH_SPACE, c)
            if c in ZERO_WIDTH_SPACE:
                if not match:
                    print
                    print "code point %s is matched by RE_ZERO_WIDTH_SPACE but not in ZERO_WIDTH_SPACE" % cpForPrint(cp)
                    errorCount += 1
            else:
                if match:
                    print
                    print "code point %s is not matched by RE_ZERO_WIDTH_SPACE but in ZERO_WIDTH_SPACE" % cpForPrint(cp)
                    errorCount += 1
        print
        print "OK" if not errorCount else "FAILED %d" % errorCount
        assert not errorCount
        
        print "Test RE_WHITESPACES vs. WHITESPACE"
        import sys
        errorCount = 0
        for cp in range(0x10ffff+1):
            if cp % 10000 == 0:
                sys.stdout.write('.')
            c = chr(cp)
            match = re.match(RE_WHITESPACES, c)
            if c in WHITESPACE:
                if not match:
                    print
                    print "code point %s is matched by RE_WHITESPACES but not in WHITESPACE" % cpForPrint(cp)
                    errorCount += 1
            else:
                if match:
                    print
                    print "code point %s is not matched by RE_WHITESPACES but in WHITESPACE" % cpForPrint(cp)
                    errorCount += 1
        print
        print "OK" if not errorCount else "FAILED %d" % errorCount
        assert not errorCount
    
        print r"Test RE_WHITESPACES vs. builtin regular expression (must not match; for informational use)"
        #import re
        #import unicodedata
        #for i in xrange(sys.maxunicode + 1):
        #    u = unichr(i)
        #    if re.match(u"\s", u, re.UNICODE):
        #        try:
        #            name = unicodedata.name(u)
        #        except ValueError:
        #            name = ""
        #        print("'{}', # u'{}' {} {}".format(u.encode('utf8').encode('string_escape'), u.encode('unicode_escape'), unicodedata.category(u), name))
        errorCount = 0
        knownExceptions1 = [
        ]
        knownExceptions2 = [
            0x1680, # OGHAM SPACE MARK # usually not really a space but a dash
            0x180e, # MONGOLIAN VOWEL SEPARATOR # is a zero width space
        ]    
        for cp in range(0x10ffff+1):
            if cp % 10000 == 0:
                sys.stdout.write('.')
            try:
                cUtf8 = chr(cp)
                cUnic = cUtf8.decode('utf-8')
            except:
                print
                print "can't test code point %s" % cpForPrint(cp)
                continue # ignore code points that python can't convert to unicode string
            assert isinstance(cUtf8, bytes)
            assert isinstance(cUnic, unicode)
            match1 = re.match(RE_WHITESPACES, cUtf8)
            match2 = re.match(r"\s", cUnic, re.UNICODE)
            if (bool(match1) != bool(match2)):
                print
                if match1:
                    print r'code point %s is matched by RE_WHITESPACES but not by r"\s"' % cpForPrint(cp)
                    if not cp in knownExceptions1:
                        errorCount += 1
                else:
                    print r'code point %s is not matched by RE_WHITESPACES but by r"\s"' % cpForPrint(cp)
                    if not cp in knownExceptions2:
                        errorCount += 1
        print
        print "OK" if not errorCount else "WARNINGS %d" % errorCount
        assert not errorCount

if __debug__ and __name__ == '__main__':
    _runTests()
