#!/usr/bin/env python
# vim:fileencoding=UTF-8:ts=4:sw=4:sta:et:sts=4:ai

from __future__ import (unicode_literals, division, absolute_import, print_function)
calibre_translate = _
from calibre_plugins.ki_notes.localize import (_, gettext)

__license__   = 'GPL v3'
__copyright__ = '2013, Axel Walthelm'
__docformat__ = 'restructuredtext en'

pluginIcon = None
pluginName = 'KiNotes'

import sys, os, re, threading

# import our stuff
from calibre_plugins.ki_notes.lib import FileTools
from calibre_plugins.ki_notes.config import prefs

if False:
    # This is here to keep my python error checker from complaining about
    # the builtin functions that will be defined by the plugin loading system
    get_icons = get_resources = None

# The class that all interface action plugins must inherit from
from calibre.gui2.actions import InterfaceAction
from calibre_plugins.ki_notes.main import KiNotesDialog
from calibre.gui2.device import device_signals
from calibre.gui2.dialogs.message_box import MessageBox
from calibre.gui2 import (error_dialog, question_dialog, warning_dialog, info_dialog, choose_dir, open_local_file, gprefs)
from PyQt4.Qt import (pyqtSignal, Qt, QApplication, QIcon, QMenu, QPixmap, QTimer, QTextCursor)

# The first icon is the plugin icon, referenced by position.
# The rest of the icons are referenced by name
_PLUGIN_ICONS = ('images/icon.png',)

_MAIN_ACTION_DESCRIPTION = _('Put Kindle annotations into EPUB')

# Calibre has some nice popup dialogs based on class MessageBox,
# but no designated interface to indicate which plugin is showing the message.
# To work around this, we hook MessageBox.__init__ to add some customization;
# hooking is a rarely used technique, which has the problem that
# unhooking is generally not safe when multiple parties hook the same target;
# therefore we use "hook late, never unhook" as the safest hooking strategy.
global _MessageBox__init__orig
global _MessageBox__init__hooked
_MessageBox__init__orig = None
_MessageBox__init__hooked = False

def _MessageBox__init__hook(*args,**keyargs):
    _MessageBox__init__orig(*args,**keyargs)
    global _MessageBox__init__hooked
    if not _MessageBox__init__hooked:
        return
    _MessageBox__init__hooked = False # only the first MessageBox is customized
    print("KiNotes: _MessageBox__init__hook(%r, %r)" % (args, keyargs))
    try:
        if pluginName:
            title = unicode(args[0].windowTitle())
            prefix = pluginName + u' '
            if not (title.startswith(prefix) or title == pluginName):
                args[0].setWindowTitle(prefix + title.strip())
        if pluginIcon:
            args[0].setWindowIcon(pluginIcon)
    except:
        import traceback
        traceback.print_exc()
    print("KiNotes: _MessageBox__init__hook end")
    
class KiNoteMessageBox():    
    def __enter__ (self):
        global _MessageBox__init__orig
        if _MessageBox__init__orig is None:
            _MessageBox__init__orig = MessageBox.__init__
            MessageBox.__init__ = _MessageBox__init__hook
        global _MessageBox__init__hooked
        _MessageBox__init__hooked = True
    def __exit__ (self, type, value, traceback):
        global _MessageBox__init__hooked
        _MessageBox__init__hooked = False

def uri2LocalEpubFile(url):
    path = re.sub(ur"\.[^.]+$", ur".epub", FileTools.localFilePathOfUri(url))
    if not os.path.isfile(path):
        return None # file does not exist
    return path

class KiNotesInterfaceAction(InterfaceAction):

    name = pluginName

    # Declare the main action associated with this plugin
    # The keyboard shortcut can be None if you dont want to use a keyboard
    # shortcut. Remember that currently calibre has no central management for
    # keyboard shortcuts, so try to use an unusual/unused shortcut.
    action_spec = (pluginName, None, _MAIN_ACTION_DESCRIPTION, None)

    #: Set of locations to which this action must not be added.
    #: See :attr:`all_locations` for a list of possible locations
    #dont_add_to = frozenset([])

    #: Set of locations from which this action must not be removed.
    #: See :attr:`all_locations` for a list of possible locations
    #dont_remove_from = frozenset(['toolbar', 'toolbar-device'])
    
    #all_locations = frozenset(['toolbar', 'toolbar-device', 'context-menu',
    #    'context-menu-device', 'toolbar-child', 'menubar', 'menubar-device',
    #    'context-menu-cover-browser'])

    #: a suggestion into which toolbars/menus to install
    #: not (yet) part of the plugin interface
    default_add_to = frozenset(['toolbar', 'toolbar-device', 'context-menu', 'context-menu-device'])
    
    #: suggestions where to place action in toolbar/menu
    #: after given entry if second value is True,
    #: before given entry if second value is False.
    #: first match wins; if no match is found, action is appended as usual
    #: not (yet) part of the plugin interface
    default_add_after = (("Send To Device", True), ("View", True), ("Save To Disk", False))

    #: If True, then this InterfaceAction will have the opportunity to interact
    #: with drag and drop events. See the methods, :meth:`accept_enter_event`,
    #: :meth`:accept_drag_move_event`, :meth:`drop_event` for details.
    accepts_drops = True
    
    nr_jobs_running = 0
    interactive_job = None
    
    def genesis(self):
        print("KiNotes: genesis")
        self.menus_lock = threading.RLock()
        self.connected_device = None
        self.connected_device_metadata_available = False
        
        # This method is called once per plugin, do initial setup here

        # Set the icon for this interface action
        # The get_icons function is a builtin function defined for all your
        # plugin code. It loads icons from the plugin zip file. It returns
        # QIcon objects, if you want the actual data, use the analogous
        # get_resources builtin function.
        #
        # Note that if you are loading more than one icon, for performance, you
        # should pass a list of names to get_icons. In this case, get_icons
        # will return a dictionary mapping names to QIcons. Names that
        # are not found in the zip file will result in null QIcons.
        #icon = get_icons('images/icon.png')
        # Read the plugin icons and store for potential sharing with the config widget
        icons = get_icons(_PLUGIN_ICONS)
        self.icons = icons if isinstance(icons, dict) else { _PLUGIN_ICONS[0]: icons }
        print("KiNotes: self.icons=", self.icons)
        global pluginIcon
        pluginIcon = icons[_PLUGIN_ICONS[0]] if isinstance(icons, dict) else icons

        # The qaction is automatically created from the action_spec defined
        # above
        self.menu = QMenu(self.gui)
        self.qaction.setMenu(self.menu)
        self.qaction.setIcon(pluginIcon)
        self.qaction.triggered.connect(self.main_button_pressed)
        self.menu.aboutToShow.connect(self.about_to_show_menu)
        
        # on first launch, adjust location of action in toolbars/menus
        # and try to ensure the user sees the action in the UI
        if not prefs['_is_installed']:
            try:
                self.install_action_to_locations()
            except:
                import traceback
                traceback.print_exc()

    def install_action_to_locations(self):
        print("KiNotes: install_action_to_locations")
        for forced_locations in ([], self.default_add_to):
            is_installed = False
            for location in self.all_locations:
                actions = gprefs.get('action-layout-'+location, [])
                if self.name in actions or location in forced_locations:
                    is_installed = True
                    new_actions = [a for a in actions if a != self.name]
                    for action, add_after in self.default_add_after:
                        if action in new_actions:
                            pos = new_actions.index(action)
                            if add_after:
                                pos += 1
                            new_actions.insert(pos, self.name)
                            break
                    if not self.name in new_actions:
                        new_actions.append(self.name)
                    if list(actions) != new_actions:
                        print("KiNotes: install or move in", location)
                        gprefs['action-layout-'+location] = tuple(new_actions)
            if is_installed:
                break

    def get_icon(self, name):
        try:
            key = 'images/'+name
            if self.icons.has_key(key):
                return self.icons[key]
            return I(name)
        except:
            return None
    
    # subclass override
    def initialization_complete(self):
        print("KiNotes: initialization_complete")
        prefs['_is_installed'] = True
        self.rebuild_menus()
        # Subscribe to device connection events
        device_signals.device_connection_changed.connect(self.on_device_connection_changed)
        device_signals.device_metadata_available.connect(self.on_device_metadata_available)

    def on_device_connection_changed(self, is_connected):
        print("KiNotes: on_device_connection_changed")
        if not is_connected:
            self.connected_device = None
            self.connected_device_metadata_available = False
        else:
            self.connected_device = self.gui.device_manager.device
            print("KiNotes: device info %r" % (self.connected_device.get_device_information(),))
            print("KiNotes: device name %r" % (self.connected_device.name,))
            print("KiNotes: device gui_name %r" % (self.connected_device.gui_name,))

        self.rebuild_menus()
        # Inventory USB reader apps
        #USBReader.get_usb_reader_classes()

    def on_device_metadata_available(self):
        print("KiNotes: on_device_metadata_available")
        self.connected_device_metadata_available = True
        
    def get_kindle_path_interactive(self):
        # check for a connected Kindle
        kindleConnected = False
        kindleRoot = None
        if self.connected_device:
            devName = self.connected_device.get_gui_name()
            devInfo = self.connected_device.driveinfo
            print("KiNotes: device = %r %r" % (devName, devInfo))
            if devName == 'Amazon Kindle':
                kindleRoot = devInfo[u'main'][u'prefix']
                kindleConnected = True
        # check if Calibre has read all book information from Kindle
        if kindleConnected:
            if not self.connected_device_metadata_available:
                with KiNoteMessageBox(): error_dialog(self.gui, '',
                    _('Calibre is still reading Kindle information.\n'
                      'Please try again in a moment.'), show=True)
                return # abort
        else:
            with KiNoteMessageBox(): answer = question_dialog(self.gui, pluginName,
                _("Calibre is not aware of a connected Kindle. "
                  "Usually you will want to (re-)connect your Kindle to the computer Calibre runs on using your USB cable. "
                  "If you have a copy of the Kindle folder, you may try to use it instead.\n"
                  "\n"
                  "Would you like to select the Kindle root folder manually?"),
                _("The selected folder needs to contain a 'documents' and 'system' folder. "
                  "The kindle system folder is hidden, so some tools may not show or copy it.\n"
                  "\n"
                  "Here is an example of what a Kindle root folder may contain:\n") +
                  "\n"
                  "documents/"+_("Carroll, Lewis/Alice in Wonderland")+".mobi\n"
                  "documents/My Clippings.txt\n"
                  "system/userannotlog\n"
                  "system/version.txt"
                  .replace('/', os.sep),
                skip_dialog_name='KiNotes_OfflineUsageInfo')
            print("KiNotes: answer = %r" % answer)
            if not answer is True:
                return
            kindleRoot = choose_dir(self.gui,
                pluginName + '_kindlePathChooser',
                _('Please select Kindle root folder for KiNotes; Calibre is not aware of a connected Kindle'))
            if not kindleRoot:
                return # user cancelled
            for expectedFolder in ('documents', 'system'):
                if kindleRoot and not os.path.exists( os.path.join(kindleRoot, expectedFolder) ):
                    with KiNoteMessageBox(): error_dialog(self.gui, '',
                        _("Selected folder is not a Kindle root folder because it does not contain a '%s' folder") % expectedFolder,
                        '', show=True)
                    return # abort
        print("KiNotes: kindleRoot = %r%s" % (kindleRoot, " connected" if kindleConnected else ""))
        return kindleRoot, kindleConnected

    def main_button_pressed(self, dropped_files=None):
        # check for a connected Kindle
        answer = self.get_kindle_path_interactive()
        if not answer:
            return
        kindleRoot, kindleConnected = answer

        # map ids to paths
        def get_device_path_from_id(book_id):
            for x in ('memory', 'card_a', 'card_b'):
                x = getattr(self.gui, x+'_view').model()
                paths = x.paths_for_db_ids(set([book_id]), as_map=True)[book_id]
                if paths:
                    return paths[0].path
        get_kindle_path = get_device_path_from_id
        get_calibre_path = lambda book_id: self.gui.current_db.format_abspath(book_id, 'EPUB', index_is_id=True)
        
        if dropped_files:
            # to find given file in Calibre db, we use that the path contains the book_id,
            # which is a decimal integer number, surrounded by non-digit characters;
            id2path = {}
            for nr, dropped_file in enumerate(dropped_files):
                book_id = -nr-1 # use negative numbers as fake book ids for files not in Calibre db
                for number in re.findall(r"[0-9]+", dropped_file):
                    path = get_calibre_path(int(number))
                    if FileTools.samefile(path, dropped_file):
                        book_id = int(number)
                        dropped_file = path # prefer path generated by Calibre
                        break
                id2path[book_id] = dropped_file
            book_ids = id2path.keys()
            get_calibre_path = lambda book_id: id2path[book_id]
        else:
            # Get currently selected rows
            rows = self.gui.current_view().selectionModel().selectedRows()
            if not rows or len(rows) == 0:
                with KiNoteMessageBox(): error_dialog(self.gui, '',
                    _('No books selected'), show=True)
                return
                
            # Get currently selected book ids and a way to get_kindle_path from id
            if self.gui.current_view() is self.gui.library_view:
                # user is currently viewing the list of books in Calibre library
                book_ids = list(map(self.gui.library_view.model().id, rows))
            else:
                # user is currently viewing the list of books on Kindle
                id2path = {}
                model = self.gui.current_view().model()
                for nr, idx in enumerate(model.indices(rows)):
                    book = model.db[idx]
                    book_id = book.application_id
                    if book_id is None:
                        book_id = -nr-1 # use negative numbers as fake book ids for books not in Calibre db
                    id2path[book_id] = book.path
                book_ids = id2path.keys()
                get_kindle_path = lambda book_id: id2path[book_id]                
        print("KiNotes: book ids =", book_ids)

        # Get pairs of Calibre-EPUB and Kindle book paths
        jobs = []
        errors = []
        def existing(path):
            if path and os.path.isfile(path):
                return path
            print("KiNotes: path not found: %r" % (path,))
        for book_id in sorted(book_ids):
            # Get the epub file for the book
            cal_path = existing(get_calibre_path(book_id))
            dev_path = existing(get_kindle_path(book_id))
            print("KiNotes: Calibre path =", cal_path)
            print("KiNotes: Kindle path =", dev_path)
            if cal_path and (dev_path or not kindleConnected):
                jobs.append([cal_path, dev_path, kindleRoot])
            else:
                title = self.gui.current_db.title(book_id, index_is_id=True) if book_id >= 0 else None
                authors = self.gui.current_db.authors(book_id, index_is_id=True) if book_id >= 0 else None
                if authors:
                    authors = u" & ".join([v.replace('|', ',') for v in authors.split(',')])
                if title:
                    error = u"%s (%s)\n" % (title, authors) if authors else title
                elif cal_path or dev_path:
                    error = _("File %s") % (cal_path or dev_path) + u"\n"
                else:
                    error = _("Book #%s") % book_id + u"\n"
                if not cal_path:
                    error += u"  • " + _("has no EPUB format in Calibre library") + "\n"
                if not dev_path and kindleConnected:
                    error += u"  • " + _("was not found on Kindle") + "\n"
                errors.append(error)
        if not jobs:
            with KiNoteMessageBox(): error_dialog(self.gui, '',
                 gettext.ungettext(
                     'Dropped book is not suitable for KiNotes processing',
                     'No dropped book is suitable for KiNotes processing',
                     len(book_ids))
                 if dropped_files else
                 gettext.ungettext(
                     'Selected book is not suitable for KiNotes processing',
                     'No selected book is suitable for KiNotes processing',
                     len(book_ids)),
                u'\n'.join(errors), show=True)
            return

        icon, title = None, pluginName
        if errors:
            # use warning icon and title to generate a warning+question dialog
            wd = warning_dialog(self.gui, '', '')
            icon, title = wd.windowIcon(), wd.windowTitle()
        with KiNoteMessageBox(): answer = question_dialog(self.gui, title,
            ('' if not errors else (
                _('Some of the dropped books are not suitable for KiNotes processing.') if dropped_files else
                _('Some of the selected books are not suitable for KiNotes processing.')) + '\n\n') +
            gettext.ungettext("Ok to start job?", "Ok to start %d jobs?" % len(jobs), len(jobs)), u'\n'.join(errors),
            skip_dialog_name='KiNotes_StartJobs_'+str(len(str(len(jobs))) if len(jobs)>1 else '0')+('_err' if errors else ''),
            override_icon=icon)
        print("KiNotes: answer = %r" % answer)
        if not answer is True:
            return
        print("KiNotes: starting %d jobs..." % len(jobs))
        self.interactive_job = jobs[0] if len(jobs)==1 and self.nr_jobs_running == 0 else None
        self.nr_jobs_running += len(jobs)
        #from calibre.utils.ipc import worker
        job_options = {
            'style': prefs['style'],
            'style_highlighting': prefs['style_highlighting'],
            'style_anchors': prefs['style_anchors'],
            'style_tabstops': prefs['style_tabstops'],
            'style_id_timestamp': prefs['style_id_timestamp'],
            'postfix': prefs['save_postfix'],
            'backup': prefs['save_backups']
            }
        for job_args in jobs:
            job_args.append(job_options)
            self.gui.job_manager.run_job(self.Dispatcher(self.job_done), 'arbitrary_n',
                args=('calibre_plugins.ki_notes.KiNotes', 'run_calibre_job',
                      tuple(job_args)),
                      description='KiNotes processing')

#        return None
#        
#        # The base plugin object defined in __init__.py
#        base_plugin_object = self.interface_action_base_plugin
#        # Show the config dialog
#        # The config dialog can also be shown from within
#        # Preferences->Plugins, which is why the do_user_config
#        # method is defined on the base plugin class
#        do_user_config = base_plugin_object.do_user_config
#
#        # self.gui is the main calibre GUI. It acts as the gateway to access
#        # all the elements of the calibre user interface, it should also be the
#        # parent of the dialog
#        d = KiNotesDialog(self.gui, self.qaction.icon(), do_user_config)
#        d.show()
    
    def job_done(self, job):
        print("KiNotes: job_done result = %r failed = %r" % (job.result, job.failed))
        self.nr_jobs_running -= 1
        if self.nr_jobs_running == 0 and self.interactive_job:
            if job.failed:
                with KiNoteMessageBox(): error_dialog(self.gui, '',
                    _("Single job failed. See job details for more information."), show=True)
            else:
                with KiNoteMessageBox(): answer = question_dialog(self.gui, pluginName,
                    _('Single job finished successfully. '
                      'Resulting EPUB file with inserted KiNotes was written to the folder the Calibre EPUB is located in. '
                      'Would you like to open this folder?'), '',
                    skip_dialog_name='KiNotes_ShowSingleJobResultFolder',
                    skip_dialog_skipped_value=False)
                if answer is True:
                    open_local_file(os.path.split(self.interactive_job[0])[0])
        
#        if job.result:
#            # Problems during processing
#            # jobs.results is a list
#            # - the first entry is the intended title for the dialog
#            # - the second entry is a tuple indicating errors and warnings
#            # Subsequent strings are messages
#            dialog_title = job.result[0]
#            nr_error, nr_warning = job.result[1]
#            messages = '\n'.join(job.result[2:])
#            if nr_errors == 0:
#                msg = _("KiNotes job complete, with warnings.")
#                with KiNoteMessageBox(): warning_dialog(self.gui, dialog_title, msg, det_msg='\n'.join(job.result), show=True)
#            else:
#                msg = _("KiNotes job )
#                job.result.append("KiNotes job terminated.")
#                with KiNoteMessageBox(): error_dialog(self.gui, dialog_title,'\n'.join(job.result),show=True)
#                return
#
#        if job.failed:
#            return self.gui.job_exception(job)

        if self.nr_jobs_running == 0:
            if self.interactive_job:
                self.interactive_job = None
                self.gui.status_bar.show_message(_('KiNotes job done.'), 3000)
            else:
                self.gui.status_bar.show_message(_('All KiNotes jobs done.'), 3000)
        
    def apply_settings(self):
        from calibre_plugins.ki_notes.config import prefs
        # In an actual non trivial plugin, you would probably need to
        # do something based on the settings in prefs
        prefs
        
    def get_drop_args(self, event, md, store=False):
        if False:
            for fmt in md.formats():
                print("KiNotes: drop format %r:" % fmt)
                try:
                    print("KiNotes:    %r" % md.data(fmt))
                except Exception, e:
                    print("KiNotes:    raised exception: %s" % e)
        if not md.hasFormat("text/uri-list"):
            return False
        urls = re.sub(r"[\r\n]+", "\n", str(md.data("text/uri-list")).strip()).split('\n')
        files = filter(None, map(uri2LocalEpubFile, urls))
        if not files:
            return False
        if store:
            self.dropped_files = files
        return True

    # subclass override
    def accept_enter_event(self, event, md):
        return self.get_drop_args(event, md)

    # subclass override
    def accept_drag_move_event(self, event, md):
        return self.get_drop_args(event, md)

    # subclass override
    def drop_event(self, event, md):
        if not self.get_drop_args(event, md, True):
            return False
        QTimer.singleShot(1, self.do_drop_event)
        return True

    def do_drop_event(self):
        # Allow the accepted event to process
        QApplication.processEvents()
        print("KiNotes: do_drop_event with %s files", self.dropped_files and len(self.dropped_files))
        if self.dropped_files and len(self.dropped_files) < 5:
            print("KiNotes: dropped_files =", self.dropped_files)
        self.main_button_pressed(self.dropped_files)

    def about_to_show_menu(self):
        self.rebuild_menus()
        
    def rebuild_menus(self):
        print("KiNotes: rebuild_menus", self.menus_lock)
        with self.menus_lock:
            m = self.menu
            m.clear()

            ac = self.create_menu_item(m, _MAIN_ACTION_DESCRIPTION + '…', image=self.get_icon("icon.png") )
            ac.triggered.connect(self.main_button_pressed)
            m.addSeparator()

            # Add 'Customize plugin…'
            ac = self.create_menu_item(m, _('KiNotes preferences') + '…', image=self.get_icon("config.png"))
            ac.triggered.connect(self.show_configuration)

            # Add 'About…' / 'Help…'
            ac = self.create_menu_item(m, _('About KiNotes') + '…', image=self.get_icon("help.png"))
            ac.triggered.connect(self.show_about)
            
    def create_menu_item(self, m, menu_text, image=None, tooltip=None, shortcut=None):
        ac = self.create_action(spec=(menu_text, None, tooltip, shortcut), attr=menu_text)
        if image:
            ac.setIcon(QIcon(image))
        m.addAction(ac)
        return ac

    def show_configuration(self):
        self.interface_action_base_plugin.do_user_config(self.gui)

    def show_about(self):
        version = self.interface_action_base_plugin.version
        title = "%s %d.%d.%d" % (self.name, version[0], version[1], version[2])
        msg = _('<u>The <b>KiNotes</b> Calibre Plugin</u><br><br>'
              'Created {copyright} by {author}<br><br>'
              'To learn more about this plugin, visit the <a href="{forum_url}">KiNotes plugin thread</a> '
              'at MobileRead\'s Calibre forum.<br>').format(
              copyright = '© 2013',
              author = '<a href="http://www.walthelm.net/">Axel Walthelm</a>',
              forum_url = 'http://www.mobileread.com/forums/showthread.php?t=XXX')
        html = get_resources('about.txt').decode('utf-8')
        text = re.sub(ur"<[^>]*>", u"", html).replace(u"&nbsp;", u" ")
        html = re.sub(ur"\r\n|\r|\n", u"<br>", html)
        d = MessageBox(MessageBox.INFO, title, msg, det_msg=text, q_icon=pluginIcon)
        try:
            d.det_msg.clear()
            d.det_msg.appendHtml(html)
            d.det_msg.moveCursor( QTextCursor.Start, QTextCursor.MoveAnchor )
        except:
            pass
        d.exec_()
